; a simple merlin debugger


;;; DEBUGGER class

(define-class debugger
  (instance-variables
    processor
    mpe-debuggers))
    
(define-method (debugger 'initialize p)
  (let ((mpe-count (p 'mpe-count)))
    (set! processor p)
    (set! mpe-debuggers (make-vector mpe-count))
    (dotimes (i mpe-count)
      (vector-set! mpe-debuggers i (mpe-debugger 'new (p 'mpe i))))
    self))

(define-method (debugger 'processor)
  processor)

(define-method (debugger 'mpe-debugger n)
  (vector-ref mpe-debuggers n))


;;; MPE-DEBUGGER class

(define-class mpe-debugger
  (instance-variables
    processor
    start-address
    run-state
    watch-forms
    symbols))

(define-method (mpe-debugger 'initialize p)
  (set! processor p)
  (set! start-address *instruction-rom-base*)
  (set! run-state 'stop)
  (set! symbols '())
  self)

(define-method (mpe-debugger 'processor)
  processor)

(define-method (mpe-debugger 'run-state)
  run-state)

(define-method (mpe-debugger 'set-run-state! state)
  (set! run-state state)
  self)

(define-method (mpe-debugger 'start-address)
  start-address)

(define-method (mpe-debugger 'set-start-address! a)
  (set! start-address a)
  self)

(define-method (mpe-debugger 'set-symbols! s)
  (set! symbols s)
  self)

(define-method (mpe-debugger 'find-symbol sym)
  (and (bound? sym symbols) (symbol-value sym symbols)))

(define-method (mpe-debugger 'map-over-symbols fun)
  (let ((wrapper (lambda (s) (fun (car s) (cdr s)))))
    (map wrapper (environment-bindings symbols))))

(define-method (mpe-debugger 'runtime-eval expr)
  (eval expr symbols))

(define-method (mpe-debugger 'watch var args)
  (let ((form (assoc var watch-forms)))
    (if form
      (set-cdr! form args)
      (set! watch-forms (acons var args watch-forms))))
  self)

; setup a watch expression
(define-macro (watch var &rest args)
  `(*default-mpe-debugger* 'watch ',var ',args))

(define-method (mpe-debugger 'unwatch var)
  (set! watch-forms (remove var watch-forms :key car))
  self)

; remove a watch expression
(define-macro (unwatch var)
  `(*default-mpe-debugger* 'unwatch ',var))

(define-method (mpe-debugger 'display-state)
  (processor 'display)
  (when watch-forms
    (newline)
    (self 'display-watch-forms))
  self)

(define-method (mpe-debugger 'display-watch-forms &optional (s *standard-output*))
  (let loop ((forms (reverse watch-forms)))
    (when forms
      (let* ((form (car forms))
             (sym (car form))
             (reg (self 'find-symbol sym))
             (args (cdr form)))
        (when reg
          (let ((reg-name (reg 'name)))
            (if (string=? (symbol-name sym) reg-name)
              (format s "~%~S:" sym)
              (format s "~%~S (~A):" sym reg-name))
            (apply self 'display-watch-form s reg args)))
        (loop (cdr forms))))))

(define-method (mpe-debugger 'display-watch-form s reg &key (fracbits #f)
                                                            (bits #t)
                                                            (real #t))
  (if fracbits
    (unless (number? fracbits)
      (set! fracbits (self 'runtime-eval fracbits)))
    (set! fracbits (reg 'fracbits)))
  (when bits
    (reg 'display-bits fracbits s))
  (when real
    (let ((real-value (reg 'value->real :fracbits fracbits)))
      (write-char #\Space s)
      (when bits
        (write-char #\( s))
      (format s "~S" real-value)
      (when bits
        (write-char #\) s))))
  (values))

(define-method (mpe 'write-image-to-file addr x-size y-size filename)
  (let ((p (self 'mmp)))
    (write-image-to-file filename (p 'memory) addr x-size y-size)))

; debugger globals
(define *default-mmp* nil)
(define *default-mpe* nil)
(define *default-debugger* nil)
(define *default-mpe-debugger* nil)

; called by emulator on halt conditions
(define (halt-hook memory code addr)
  (let ((p (*default-mmp* 'find-mpe-from-memory-handle memory)))
    (when p
      (if (eq? code hh-halt) 
	(*halt-hook* p addr)
        (if (eq? code hh-break)
	  (*break-hook* p addr)
          (if (eq? code hh-error)
            (*error-hook* p addr)
            (if (eq? code hh-badload)
              (*badload-hook* p addr)
              (if (eq? code hh-badmultiply)
                (*badmultiply-hook* p addr)
	        (debug-halt p addr "unknown halt code")))))))))

(define (debug-halt p addr reason)
  (format #t "~%~A ~A at: ~A" p reason addr)
  #f)

(define (*halt-hook* p addr)
  (debug-halt p addr "halted"))

(define (*break-hook* p addr)
  (p 'single-step addr))

(define (*error-hook* p addr)
  (debug-halt p addr "error"))

(define (*badload-hook* p addr)
  (debug-halt p addr "bad reference after load"))

(define (*badmultiply-hook* p addr)
  (debug-halt p addr "bad reference after multiply"))

; setup debugger environment
(define (setup-debugger &key (mpe-count 4))
  (set! *default-mmp* (mmp 'new :mpe-count mpe-count))
  (set! *default-debugger* (debugger 'new *default-mmp*))
  (mmp-enable-halt-hook! (*default-mmp* 'memory) halt-hook)
  (select-processor 0))

(define (select-processor i &optional (d *default-debugger*))
  (when d
    (set! *default-debugger* d)
        (set! *default-mmp* (d 'processor))
    (set! *default-mpe* (*default-mmp* 'mpe i))
    (set! *default-mpe-debugger* (d 'mpe-debugger i))))

; setup default environment
(setup-debugger :mpe-count 4)

; load merlin source code
(define (mload file &key ((:debugger d) *default-mpe-debugger*))
  (multiple-value-bind (status text file line offset)
                       (mparse file "mload.out")
    (if (= status 0)
      (begin
        (set! &d d)
        (set! &m (d 'processor))
        (load "mload.out")
        (d 'start))
      (begin
        (if *gui-debug*
          (begin
            (when file
              (gui-display-source file line 1))
            (message-box (or text "Unknown error")))
          (begin
            (format #t "~%Error: ~A~%" text)
            (when file
              (display-lines file line 1))))
        #f))))

; run the program
(define-method (mpe-debugger 'run)
  (self 'start)
  (let ((m (self 'processor)))
    (let loop ()
      (when (m 'fetch-instruction (m 'next))
        (m 'step)
        (loop)))
    self))

(define-method (mpe-debugger 'start)
  ((self 'processor) 'set-pc! (self 'start-address))
  self)

; a simple single stepper
(define-method (mpe-debugger 'single-step)
  (let ((m (self 'processor)))
    (let loop ()
      (let ((i (m 'fetch-instruction (m 'next))))
        (when i
          (self 'display-state)
          (format #t "~%~%")
          (display-source (i 'source))
          (when (self 'execute-step-command)
            (loop)))))
    (self 'display-state)
    self))

(define-method (mpe-debugger 'execute-step-command)
  (let loop ((return? #f) (return-value #f))
    (if return?
      return-value
      (begin
        (format t "~%[s to step]")
        (let ((cmd (read)))
          ; handle commands that exit the command loop
          (case cmd
            (s ((self 'processor) 'step)
               (loop #t #t))
            (q (set! return? #t)
               (loop #t #f))
            ; handle commands that don't exit the command loop
            (else (case cmd
                    (e (let ((expr (read)))
                         (fresh-line)
                         (write (self 'runtime-eval expr))))
                    (else (format #t "~%Unknown step command ~S" cmd)))
                  (loop #f #f))))))))

(define (display-source ref)
  (let ((file (first ref))
        (start (second ref))
        (count (third ref)))
    (display-lines file start count)))


; handy functions

(define (run &optional (d *default-mpe-debugger*))
  (d 'run))

(define (step &optional (d *default-mpe-debugger*))
  (d 'start)
  (d 'single-step))

(define (continue &optional (d *default-mpe-debugger*))
  (d 'single-step))

(define (write-image name x-size y-size &key (base *external-ram-base*)
                                             ((:mpe p) *default-mpe*))
  (p 'write-image-to-file base x-size y-size name))

(define (runtime-eval expr &optional (d *default-mpe-debugger*))
  (d 'runtime-eval expr))

(define (elapsed-ticks &optional (p *default-mmp*))
  (mmp-elapsed-ticks (p 'memory)))

(define-method (mmp 'find-mpe-from-memory-handle mpe-memory)
  (let loop ((i 0))
    (if (< i (self 'mpe-count))
      (let ((p (self 'mpe i)))
        (if (foreign-pointer-eq? mpe-memory (p 'memory))
	  p
	  (loop (1+ i))))
      #f)))

(define-method (debugger 'find-mpe-debugger-from-mpe p)
  (let loop ((i 0))
    (if (< i (vector-length mpe-debuggers))
      (let ((d (self 'mpe-debugger i)))
        (if (eq? (d 'processor) p)
	  d
	  (loop (1+ i))))
      #f)))

(define (mparse-warning text &optional file line offset)
  (format #t "~%Warning: ~A" text)
  (when file
    (format #t "~%  File: ~A, Line: ~A, Offset: ~A" file line offset)))
	    
