(set! *backtrace* 10)

(define *register-window* nil)
(define *watch-window* nil)
(define *graphics-window* nil)
(define *gui-initialized* #f)

; do nothing now that the clocking is done by the c code
(define (gui-idle) 0)

; display dimensions
(define *display-width* 352)
(define *display-height* 288)

(define (gui-setup-debugger)
  (set! *register-window* (make-register-window   4 600 680 280))
  (set! *watch-window*    (make-watch-window    688 600 560 280))
  (set! *graphics-window* (make-graphics-window 888 307 *display-width* *display-height*))
  (gui-start-mmp *default-mmp*)
  #t)

; make the register window
(define (make-register-window x y w h)
  (make-debugger-window "Register" x y w h
                        (lambda (dc)
                          (let ((s (make-string-output-stream)))
                            (*default-mpe* 'display s)
                            (win-stream-out dc s)))))

; make the watch window
(define (make-watch-window x y w h)
  (make-debugger-window "Watch" x y w h
                        (lambda (dc)
                          (when *gui-initialized*
                            (let ((s (make-string-output-stream)))
                              (*default-mpe-debugger* 'display-watch-forms s)
                              (win-stream-out dc s))))))
  
; make the graphics window
(define (make-graphics-window x y w h)
  (make-debugger-window 
    "Graphics" x y w h
    (lambda (dc)
      (display-image dc (*default-mmp* 'memory) *external-ram-base* 352 288))))

(define (update-graphics-window addr value)
  (let* ((offset (ash (- addr *external-ram-base*) -2))
         (x (remainder offset *display-width*))
         (y (quotient offset *display-width*)))
    (display-pixel *graphics-window* x y (ycrcb->rgb value))))

(define (gui-start-mmp p)
  (mmp-start (p 'memory))
  (mmp-start-stat (p 'memory))
  (p 'set-display-update-function! *external-ram-base*
     (* *display-width* *display-height* 4)
     update-graphics-window))

(define (display-pixel w x y c)
  (win-call/dc w (lambda (dc) (win-set-pixel dc x y c))))

; make a debugger output window
(define (make-debugger-window title x y w h fcn)
  (let ((win (make-output-window title)))
    (move-window win x y w h)
    (set-output-function win fcn)
    win))

(define *source-windows* nil)

(define (gui-update-display)
  (when *register-window* (update-window *register-window*))
  (when *watch-window* (update-window *watch-window*))
  (gui-display-current-instruction))

(define (gui-display-current-instruction)
  (let* ((i (*default-mpe* 'fetch-instruction (*default-mpe* 'next))))
    (when i
      (let* ((source (i 'source))
         (file (first source))
             (start (second source))
         (count (third source)))
    (gui-display-source file start count)))))

(define (gui-display-source file start count)
  (let ((binding (assoc file *source-windows*)))
    (if binding
      (gui-display-source-lines (cdr binding) start count)
      (let ((win (make-document-window file)))
        (when win
          (set! *source-windows* (acons file win *source-windows*))
      (gui-display-source-lines win start count))))))

(define (gui-display-source-lines win start count)
  (let ((start (line-index win (-1+ start)))
        (end (line-index win (+ start count -1))))
    (set-selection win start end)
    (win-bring-window-to-top win)))

(define *source-filter* "Source Files (*.a)|*.a|All Files (*.*)|*.*||")

(define *last-loader*)
(define *last-path*)

; called by the "Merlin/Load Source" menu item
(define (gui-load)
  (let ((path (open-file-dialog *source-filter*)))
    (and path (load-source-file path))))

(define (load-source-file path)
  (when (mload path)
    (set! *last-loader* load-source-file)
    (set! *last-path* path)
    (gui-update-display)
    #t))

(define *debug-filter* "Debug Files (*.d)|*.d|All Files (*.*)|*.*||")

; called by the "Merlin/Load Debug File" menu item
(define (gui-load-debug)
  (let ((path (open-file-dialog *debug-filter*)))
    (when path
	  (make-document-window path)
	  (load-debug-file path))))

(define (load-debug-file path)
  (when (load path)
    (set! *last-loader* load-debug-file)
    (set! *last-path* path)
    (gui-update-display)
    #t))

(define-class watch-dialog
  (instance-variables
    debugger))

(define-method (watch-dialog 'initialize &optional (d *default-mpe-debugger*))
  (set! debugger d)
  self)

(define-method (watch-dialog 'initialize-variables combo-box)
  (debugger 'map-over-symbols
    (lambda (s v) (add-string combo-box (symbol-name s))))
  (set-current-selection combo-box 0)
  self)

; called by the "Merlin/Watch" menu item
(define (gui-watch)
  (let ((dialog (watch-dialog 'new)))
    (display-watch-dialog dialog)))

; called by the "Merlin/Step" menu item
(define (gui-step)
  (when (*default-mpe* 'fetch-instruction (*default-mpe* 'next))
    (*default-mpe* 'step)
    (gui-update-display)))

; called by the "Merlin/Run" menu item
(define (gui-run)
  (when *default-mpe*
    (*default-mpe* 'run)))

(define (gui-proc-running?)
  (and *default-mpe* (if (eq? (*default-mpe* 'run-state) 'run) 1 0)))

; called by the "Merlin/Run All" menu item
(define (gui-run-all)
  (when *default-mmp*
    (dotimes (i (*default-mmp* 'mpe-count))
      ((*default-mmp* 'mpe i) 'run))))

; called by the "Merlin/Stop" menu item
(define (gui-stop)
  (when *default-mpe*
    (*default-mpe* 'stop)
    (gui-update-display)))

; called by the "Merlin/Stop All" menu item
(define (gui-stop-all)
  (when *default-mmp*
    (dotimes (i (*default-mmp* 'mpe-count))
      ((*default-mmp* 'mpe i) 'stop))
    (gui-update-display)))

; called by the "Merlin/Restart" menu item
(define (gui-restart)
  (when *last-loader*
    (mmp-stop) ; stop the idle clocking
    (setup-debugger) ; sledge hammer!
    (*last-loader* *last-path*)
    (gui-start-mmp *default-mmp*)
    (gui-update-display)
    (when *graphics-window* (update-window *graphics-window*))))

; called by the "Processor/Processor n" menu item
(define (gui-set-proc! n)
  (when *default-mmp*
    (select-processor n)
    (gui-update-display)))

; called by the "Processor/Processor n" menu item
(define (gui-get-proc)
  (and *default-mpe* (*default-mpe* 'unit-number)))

; called by the "Merlin/User 1" menu item
(define (gui-user-1))

; called by the "Merlin/User 2" menu item
(define (gui-user-2))

; called by the "Merlin/User 3" menu item
(define (gui-user-3))

; called by the "Merlin/User 4" menu item
(define (gui-user-4))

; handlers
(define (*halt-hook* p addr)
  (message-box "MPE halted")
  (gui-update-display))

(define (*error-hook* p addr)
  (message-box "Error")
  (gui-update-display))

(define (*badload-hook* p addr)
  (message-box "Bad reference after load")
  (gui-update-display))

(define (*badmultiply-hook* p addr)
  (message-box "Bad reference after multiply")
  (gui-update-display))

; load any user definitions of the above functions
(load "gui-user.lsp")

; setup the debugger windows
(gui-setup-debugger)

; done initializing
(set! *gui-initialized* #t)


