(format #t
"Merlin emulator, March 13, 1996, Copyright (c) 1996, by VM Labs, Inc.")

(define *default-fracbits* 16)

; bytes per instruction
(define *bytes-per-instruction* 2) ; 16 bits

; local memory map
(define *shared-ram-base*       #x00200000)
(define *shared-ram-size*       1024)
(define *local-rom-base*        #x20100000)
(define *local-rom-size*        2048)
(define *local-ram-base*        #x20200000)
(define *local-ram-size*        (+ 1024 512))

; instruction memory map
(define *instruction-rom-base*  #x20300000)
(define *max-rom-instructions*  2048)
(define *instruction-rom-size*  (* *max-rom-instructions*
                                   *bytes-per-instruction*))
(define *instruction-rom-top*   (+ *instruction-rom-base*
                                   *instruction-rom-size*))
(define *instruction-ram-base*  #x20400000)
(define *max-ram-instructions*  2048)
(define *instruction-ram-size*  (* *max-ram-instructions*
                                   *bytes-per-instruction*))
(define *instruction-ram-top*   (+ *instruction-ram-base*
                                   *instruction-ram-size*))

; external memory map
(define *external-ram-base*     #x40000000)
(define *external-ram-size*     (* 1024 1024))

; condition codes
(define cc-z   #x00000001)
(define cc-c   #x00000002)
(define cc-n   #x00000004)
(define cc-v   #x00000008)
(define cc-mv  #x00000010)
(define cc-c0z #x00000020)
(define cc-c1z #x00000040)

; halt hook codes
(define hh-halt         0)
(define hh-break        1)
(define hh-error        2)
(define hh-badload      3)
(define hh-badmultiply  4)


;;; MMP class

(define-class mmp
  (instance-variables
    mpe-units
    memory))

(define-method (mmp 'initialize &key (mpe-count 4))
  (set! memory (make-mmp mpe-count
                         :local-ram-size *local-ram-size*
                         :local-rom-size *local-rom-size*
                         :instruction-ram-size *instruction-ram-size*
                         :instruction-rom-size *instruction-rom-size*
                         :shared-ram-size *shared-ram-size*
                         :external-ram-size *external-ram-size*))
  (set! mpe-units (make-vector mpe-count))
  (dotimes (i mpe-count)
    (vector-set! mpe-units i (mpe 'new :mmp self :unit i)))
  self)

(define-method (mmp 'mpe-count)
  (vector-length mpe-units))

(define-method (mmp 'memory)
  memory)

(define-method (mmp 'mpe i)
  (vector-ref mpe-units i))

(define-method (mmp 'mpe-memory i)
  (get-mpe memory i))

(define-method (mmp 'set-display-update-function! base size fun)
  (set-update-function! memory base size fun)
  self)

(define-method (mmp 'clock)
  (mmp-clock memory)
  self)


;;; MPE class

(define-class mpe
  (instance-variables
    mmp-unit
    unit-number
    registers
    vector-registers
    ac
    rx ry
    ru rv
    xy xy-width xy-itype xy-base xy-mipmap xy-tile
    uv uv-width uv-itype uv-base uv-mipmap uv-tile
    pc rz sp cc rc0 rc1 sv-shift ac-shift
    dma-eaddr dma-flags dma-length dma-iaddr dma-1
    dma-x dma-y dma-2
    memory
    instruction-ram
    instruction-rom))

(define-method (mpe 'initialize &key ((:mmp immp)) (unit 0))
  
  ; save the associated mmp and unit number
  (set! mmp-unit immp)
  (set! unit-number unit)

  ; setup the memory
  (set! memory (mmp-unit 'mpe-memory unit-number))
  (set! instruction-ram (make-vector *instruction-ram-size*))
  (set! instruction-rom (make-vector *instruction-rom-size*))

  ; create the scalar registers
  (set! registers (make-vector 32))
  (dotimes (i 32)
    (let ((name (format nil "R~S" i)))
      (vector-set! registers i
                   (register 'new :name name :mpe self))))
  
  ; create the vector registers
  (set! vector-registers (make-vector 8))
  (dotimes (i 8)
    (let ((name (format nil "V~S" i))
          (n (* i 4)))
      (vector-set! vector-registers i
                   (vector-register 'new n :name name :mpe self))))
  
  ; create the accumulator
  (set! ac (accumulator 'new :name "AC" :mpe self))
  
  ; setup the x and y index registers
  (set! rx         (register 'new :name "RX"        :mpe self))
  (set! ry         (register 'new :name "RY"        :mpe self))
  (set! xy-width   (register 'new :name "XY_WIDTH"  :mpe self))
  (set! xy-itype   (register 'new :name "XY_ITYPE"  :mpe self))
  (set! xy-base    (register 'new :name "XY_BASE"   :mpe self))
  (set! xy-mipmap  (register 'new :name "XY_MIPMAP" :mpe self))
  (set! xy-tile    (register 'new :name "XY_TILE"   :mpe self))
  (set! xy (bilinear-index-register 'new
                                    :name "XY"
                                    :x rx
                                    :y ry
                                    :width xy-width
                                    :itype xy-itype
                                    :base xy-base
                                    :mipmap xy-mipmap
                                    :tile xy-tile))
  
  ; setup the u and v index registers
  (set! ru         (register 'new :name "RU"        :mpe self))
  (set! rv         (register 'new :name "RV"        :mpe self))
  (set! uv-width   (register 'new :name "UV_WIDTH"  :mpe self))
  (set! uv-itype   (register 'new :name "UV_ITYPE"  :mpe self))
  (set! uv-base    (register 'new :name "UV_BASE"   :mpe self))
  (set! uv-mipmap  (register 'new :name "UV_MIPMAP" :mpe self))
  (set! uv-tile    (register 'new :name "UV_TILE"   :mpe self))
  (set! uv (bilinear-index-register 'new
                                    :name "UV"
                                    :x ru
                                    :y rv
                                    :width uv-width
                                    :itype uv-itype
                                    :base uv-base
                                    :mipmap uv-mipmap
                                    :tile uv-tile))
  
  ; setup the special registers
  (set! pc         (register 'new :name "PC"        :mpe self))
  (set! rz         (register 'new :name "RZ"        :mpe self))
  (set! cc         (register 'new :name "CC"        :mpe self))
  (set! rc0        (register 'new :name "RC0"       :mpe self))
  (set! rc1        (register 'new :name "RC1"       :mpe self))
  (set! sv-shift   (register 'new :name "SV_SHIFT"  :mpe self))
  (set! ac-shift   (register 'new :name "AC_SHIFT"  :mpe self))
  (set! sp         (register 'new :name "SP"        :mpe self))

  ; setup the dma registers
  (set! dma-eaddr  (register 'new :name "DMA_EADDR" :mpe self))
  (set! dma-flags  (register 'new :name "DMA_FLAGS" :mpe self))
  (set! dma-length (register 'new :name "DMA_LENGTH":mpe self))
  (set! dma-iaddr  (register 'new :name "DMA_IADDR" :mpe self))
  (set! dma-1      (vector-io-register 'new dma-eaddr
                                            dma-flags
                                            dma-length
                                            dma-iaddr
                                            :name "DMA_1" :mpe self))
  (set! dma-x      (register 'new :name "DMA_X"     :mpe self))
  (set! dma-y      (register 'new :name "DMA_Y"     :mpe self))
  (set! dma-2      (vector-io-register 'new dma-x
                                            dma-y
                                            dma-length
                                            dma-iaddr
                                            :name "DMA_2":mpe self))
   (set! dma-stat  (register 'new :name "DMA_STAT"  :mpe self))

  ; setup the initial program counter
  (self 'set-pc! *instruction-rom-base*)

  ; return the new mpe
  self)

(define-method (mpe 'mmp)
  mmp-unit)

(define-method (mpe 'unit-number)
  unit-number)

(define-method (mpe 'memory)
  memory)

(define-method (mpe 'register n)
  (vector-ref registers n))

(define-method (mpe 'vector-register n)
  (vector-ref vector-registers n))

(define-method (mpe 'find-register-by-name name)
  (or (self 'find-scalar-register-by-name name)
      (self 'find-vector-register-by-name name)
      (case name
        (p0 (self 'vector-register 0))
        (p1 (self 'vector-register 1))
        (p2 (self 'vector-register 2))
        (p3 (self 'vector-register 3))
        (p4 (self 'vector-register 4))
        (p5 (self 'vector-register 5))
        (p6 (self 'vector-register 6))
        (p7 (self 'vector-register 7))
        (ac ac)
        (rx rx)
        (ry ry)
        (xy_width xy-width)
        (xy_itype xy-itype)
        (xy_base xy-base)
        (xy_mipmap xy-mipmap)
        (xy_tile xy-tile)
        (xy xy)
        (ru ru)
        (rv rv)
        (uv_width uv-width)
        (uv_itype uv-itype)
        (uv_base uv-base)
        (uv_mipmap uv-mipmap)
        (uv_tile uv-tile)
        (uv uv)
        (pc pc)
        (rz rz)
        (cc cc)
        (rc0 rc0)
        (rc1 rc1)
        (sv_shift sv-shift)
        (ac_shift ac-shift)
        (dma_1 dma-1)
        (dma_eaddr dma-eaddr)
        (dma_flags dma-flags)
        (dma_length dma-length)
        (dma_iaddr dma-iaddr)
        (dma_2 dma-2)
        (dma_stat dma-stat)
        (dma_x dma-x)
        (dma_y dma-y)
        (sp sp)
        (else #f))))

(define-method (mpe 'find-scalar-register-by-name name)
  (let ((pname (symbol-name name)))
    (let loop ((i 0))
      (when (< i 32)
        (let ((r (self 'register i)))
          (if (equal? pname (r 'name))
            r
            (loop (1+ i))))))))

(define-method (mpe 'find-vector-register-by-name name)
  (let ((pname (symbol-name name)))
    (let loop ((i 0))
      (when (< i 8)
        (let ((r (self 'vector-register i)))
          (if (equal? pname (r 'name))
            r
            (loop (1+ i))))))))

(define-method (mpe 'fetch-scalar addr)
  (let ((value (mpe-fetch-scalar memory addr)))
    (unless value
      (self 'illegal-memory-access addr))
    value))

(define-method (mpe 'store-scalar! addr value)
  (unless (mpe-store-scalar! memory addr value)
    (self 'illegal-memory-access addr))
  #t)

(define-method (mpe 'init-scalar! addr value)
  (unless (if (string? value)
            (mpe-load-binary! memory addr value)
			(mpe-init-scalar! memory addr value))
    (self 'illegal-memory-access addr))
  #t)

(define-method (mpe 'illegal-memory-access addr)
  (self 'stop)
  (error "Illegal memory access ~S" addr)
  self)

(define-method (mpe 'map-instruction-address addr)
  (if (and (>= addr *instruction-ram-base*)
           (< addr *instruction-ram-top*))
    (let ((offset (quotient (- addr *instruction-ram-base*)
                            *bytes-per-instruction*)))
      (if (< offset *max-ram-instructions*)
        (values instruction-ram offset)
        (self 'illegal-instruction-access addr)))
    (if (and (>= addr *instruction-rom-base*)
             (< addr *instruction-rom-top*))
      (let ((offset (quotient (- addr *instruction-rom-base*)
                              *bytes-per-instruction*)))
        (if (< offset *max-rom-instructions*)
          (values instruction-rom offset)
          (self 'illegal-instruction-access addr)))
      (self 'illegal-instruction-access addr))))

(define-method (mpe 'fetch-instruction addr)
  (multiple-value-bind (data off)
                       (self 'map-instruction-address addr)
    (vector-ref data off)))

(define-method (mpe 'store-instruction! addr &key debugger
                                                  source
                                                  code
                                                  before
                                                  after)
  (let ((i (instruction 'new self :debugger debugger
				  :source source
				  :before before
				  :after after))
	(flags (first code))
	(code (second code)))
    (unless (mpe-store-instruction! (self 'memory) addr flags code)
      (error "error storing instruction at ~S" addr))
    (multiple-value-bind (data off)
                         (self 'map-instruction-address addr)
      (vector-set! data off i))
    self))

(define-method (mpe 'illegal-instruction-access addr)
  (self 'stop)
  (error "Illegal instruction access ~S" addr)
  self)

(define-method (mpe 'next)
  (mpe-register-value memory "next"))

(define-method (mpe 'pc)
  (pc 'value))

(define-method (mpe 'set-pc! ipc)
  (pc 'set-value! ipc)
  self)

(define-method (mpe 'run-state)
  (if (mpe-running? memory) 'run 'stop))

(define-method (mpe 'run)
  (mpe-run memory)
  self)
  
(define-method (mpe 'stop)
  (mpe-stop memory)
  self)

(define-method (mpe 'clock)
  (mpe-clock memory)
  self)
              
(define-method (mpe 'pause)
  (if (mpe-running? memory)
    (mpe-stop memory)
    (mpe-run memory))
  self)

(define-method (mpe 'step)
  (mpe-single-step memory))

(define-method (mpe 'single-step ea)
  (let ((inst (self 'fetch-instruction ea)))
    (if inst
      (let ((before (inst 'before-method))
            (after (inst 'after-method)))
        (when before
          ((inst 'debugger) 'runtime-eval before))
        (when (mpe-running? memory)
	      (let ((breakpoint? (mpe-breakpoint? memory ea)))
            (when breakpoint?
              (mpe-clear-breakpoint! memory ea))
            (mpe-clock memory)
            (when breakpoint?
              (mpe-set-breakpoint! memory ea)))
          (when after
            ((inst 'debugger) 'runtime-eval after))))
      (self 'stop))
    self))

(define-method (mpe 'display &optional (s *standard-output*))
  
  ; display the vector registers
  (dotimes (i 8)
    (fresh-line s)
    ((self 'vector-register i) 'display s))
  
  ; display the x, y, u and v index registers
  (format s "~%     ")
  (rx 'display s)
  (format s "  ")
  (ry 'display s)
  (format s "  ")
  (ru 'display s)
  (format s "  ")
  (rv 'display s)

  ; display the xy and uv i/o registers
  (newline s)
  (xy 'display s)
  (newline s)
  (uv 'display s)

  ; display the special registers
  (format s "~%     ")
  (pc 'display s)
  (format s "  ")
  (rz 'display s)
  (format s "  ")
  (rc0 'display s)
  (format s "  ")
  (rc1 'display s)
  (format s "  ")
  (sp 'display s)

  ; display the dma registers
  (format s "~%DMA: EAD:")
  (write-hex (dma-eaddr 'value) s)
  (format s "  FLG:")
  (write-hex (dma-flags 'value) s)
  (format s "  LEN:")
  (write-hex (dma-length 'value) s)
  (format s "  IAD:")
  (write-hex (dma-iaddr 'value) s)
  (format s "~%DMA: X:  ")
  (write-hex (dma-x 'value) s)
  (format s "  Y:  ")
  (write-hex (dma-y 'value) s)
  (format s "  STA:")
  (write-hex (dma-stat 'value) s)

  ; display the shift count registers and the accumulator
  (format s "~%     ")
  (sv-shift 'display s)
  (format s "  ")
  (ac-shift 'display s)
  (format s "  ")
  (ac 'display s)
  
  ; display the alu flags
  (let ((cc-value (cc 'value)))
    (format s "~%     C: ~S  Z: ~S  N: ~S  V: ~S  MV: ~S  C0Z: ~S  C1Z: ~S"
            (if (zero? (logand cc-value cc-c)) 0 1)
            (if (zero? (logand cc-value cc-z)) 0 1)
            (if (zero? (logand cc-value cc-n)) 0 1)
            (if (zero? (logand cc-value cc-v)) 0 1)
            (if (zero? (logand cc-value cc-mv)) 0 1)
            (if (zero? (logand cc-value cc-c0z)) 0 1)
            (if (zero? (logand cc-value cc-c1z)) 0 1)))
  (values))


;;; NAMED-REGISTER class

(define-class named-register
  (instance-variables
    name
    mpe))

(define-method (named-register 'initialize &key ((:name iname))
                                                ((:mpe impe)))
  (set! name (or iname "Register"))
  (set! mpe impe)
  self)
  
(define-method (named-register 'name)
  name)

(define-method (named-register 'mpe)
  mpe)

(define-method (named-register 'available?)
  #t)

(define-method (named-register 'print &optional (s *standard-output*))
  (format s "#<~A>" name)
  (values))


;;; BASIC-REGISTER class

(define-class basic-register
  (super-class named-register)
  (instance-variables
    fracbits
    new-fracbits
    real-value
    new-real-value))

(define-method (basic-register 'initialize &key ((:name iname))
                                                ((:mpe impe)))
  (super 'initialize :name iname :mpe impe)
  (set! real-value 0)
  (set! fracbits 0)
  self)
  
(define-method (basic-register 'fracbits)
  fracbits)
  
(define-method (basic-register 'set-fracbits! ifracbits)
  (set! new-fracbits ifracbits)
  self)

(define-method (basic-register 'load-fracbits! ifracbits)
  (let ((delta (- fracbits ifracbits)))
    (set! real-value (* real-value (expt 2 delta)))
    (set! fracbits ifracbits))
  self)

(define-method (basic-register 'real-value)
  real-value)
  
(define-method (basic-register 'set-real-value! rvalue)
  (self 'internal-set-real-value! rvalue))

(define-method (basic-register 'internal-set-real-value! rvalue)
  (set! new-real-value rvalue)
  self)

(define-method (basic-register 'load-real-value! rvalue)
  (set! real-value rvalue)
  self)
  

;;; REGISTER class

(define-class register
  (super-class basic-register))

(define-method (register 'initialize &key ((:name iname) "Register")
                                          ((:mpe impe)))
  (super 'initialize :name iname :mpe impe)
  self)

(define-method (register 'value)
  (mpe-register-value (mpe 'memory) name))

(define-method (register 'set-value! ivalue &key ((:fracbits ifracbits))
                                                 ((:real-value rvalue)))
  (self 'internal-set-value! ivalue :fracbits ifracbits
                                    :real-value rvalue))

(define-method (register 'internal-set-value! ivalue
                                              &key ((:fracbits ifracbits))
                                                   ((:real-value rvalue)))
  (mpe-set-register-value! (mpe 'memory) name ivalue)
  (set! new-fracbits (or ifracbits fracbits))
  (if rvalue
    (self 'set-real-value! rvalue)
    (self 'set-real-value! (32bits->real ivalue :fracbits new-fracbits)))
  self)

(define-method (register 'load! ivalue &key ((:fracbits ifracbits)))
  (when ifracbits
    (set! fracbits ifracbits))
  (mpe-set-register-value! (mpe 'memory) name ivalue)
  (set! real-value (32bits->real (self 'value) fracbits))
  self)
  
(define-method (register 'load-from-real! rvalue &key ((:fracbits ifracbits)))
  (when ifracbits
    (set! fracbits ifracbits))
  (mpe-set-register-value! (mpe 'memory) name (real->32bits rvalue :fracbits fracbits))
  (set! real-value rvalue)
  self)

(define-method (register 'load-real-value! rvalue)
  (set! real-value rvalue)
  self)
  
(define-method (register 'value->real &key ((:fracbits ifracbits)))
  (32bits->real (self 'value) :fracbits (or ifracbits fracbits)))
  
(define-method (register 'display &optional (s *standard-output*))
  (format s "~A:" name)
  (let ((spaces (- 3 (string-length name))))
    (dotimes (i spaces)
      (write-char #\Space s)))
  (write-hex (self 'value) s)
  (values))

(define-method (register 'display-bits fracbits
                       &optional (s *standard-output*))
  (write-bits (self 'value) fracbits s)
  (values))


;;; ACCUMULATOR class

(define-class accumulator
  (super-class basic-register))

(define-method (accumulator 'initialize &key ((:name iname) "Accumulator")
                                             ((:mpe impe)))
  (super 'initialize :name iname :mpe impe)
  (set! value-high 0)
  (set! value-low 0)
  self)
  
(define-method (accumulator 'value-high)
  (multiple-value-bind (high low)
                       (mpe-accumulator-value (mpe 'memory))
    high))
  
(define-method (accumulator 'value-low)
  (multiple-value-bind (high low)
                       (mpe-accumulator-value (mpe 'memory))
    low))
      
(define-method (accumulator 'set-value! high low &key ((:fracbits ifracbits)))
  (when ifracbits
    (set! new-fracbits ifracbits))
  (mpe-set-accumulator-value! (mpe 'memory) high low)
  self)
  
(define-method (accumulator 'value->real &key ((:fracbits ifracbits)))
  (multiple-value-bind (high low)
                       (mpe-accumulator-value (mpe 'memory))
    (64bits->real high low :fracbits (or ifracbits fracbits))))

(define-method (accumulator 'display &optional (s *standard-output*))
  (multiple-value-bind (high low)
                       (mpe-accumulator-value (mpe 'memory))
    (format s "~A: " name)
    (write-hex high s)
    (write-hex low s)
    (values)))

(define-method (accumulator 'display-bits fracbits
                      &optional (s *standard-output*))
  (multiple-value-bind (high low)
                       (mpe-accumulator-value (mpe 'memory))
    (write-bits high (- fracbits 32) s)
    (write-bits low fracbits s)
    (values)))


;;; BASIC-VECTOR-REGISTER class

(define-class basic-vector-register
  (super-class named-register)
  (instance-variables
    name
    registers))
    
(define-method (basic-vector-register 'initialize &key ((:name iname) "Vector-Register")
                                                       ((:mpe impe)))
  (super 'initialize :name iname :mpe impe)
  (set! registers (make-vector 4))
  self)

(define-method (basic-vector-register 'register n)
  (vector-ref registers n))
    
(define-method (basic-vector-register 'set-value! v0 v1 v2 v3 &key (fracbits 7))
  ((self 'register 0) 'set-value! v0 :fracbits fracbits)
  ((self 'register 1) 'set-value! v1 :fracbits fracbits)
  ((self 'register 2) 'set-value! v2 :fracbits fracbits)
  ((self 'register 3) 'set-value! v3 :fracbits fracbits)
  self)
  
(define-method (basic-vector-register 'load! v0 v1 v2 v3 &key (fracbits 7))
  ((self 'register 0) 'load! v0 :fracbits fracbits)
  ((self 'register 1) 'load! v1 :fracbits fracbits)
  ((self 'register 2) 'load! v2 :fracbits fracbits)
  ((self 'register 3) 'load! v3 :fracbits fracbits)
  self)
  
(define-method (basic-vector-register 'load-from-reals! v0 v1 v2 v3 &key (fracbits 7))
  ((self 'register 0) 'load-from-real! v0 :fracbits fracbits)
  ((self 'register 1) 'load-from-real! v1 :fracbits fracbits)
  ((self 'register 2) 'load-from-real! v2 :fracbits fracbits)
  ((self 'register 3) 'load-from-real! v3 :fracbits fracbits)
  self)

(define-method (basic-vector-register 'set-element! n v)
  ((self 'register n) 'set-value! v)
  self)

(define-method (basic-vector-register 'display &optional (s *standard-output*))
  (format s "~A:" name)
  (dotimes (i 4)
    (format s "  ")
    ((vector-ref registers i) 'display s))
  (values))


;;; VECTOR-REGISTER class

(define-class vector-register
  (super-class basic-vector-register))

(define-method (vector-register 'initialize base &key ((:name iname) "Vector-Register")
                                                      ((:mpe impe)))
  (super 'initialize :name iname :mpe impe)
  (dotimes (i 4)
    (vector-set! registers i (mpe 'register (+ base i))))
  self)


;;; VECTOR-IO-REGISTER class

(define-class vector-io-register
  (super-class basic-vector-register))

(define-method (vector-io-register 'initialize r0 r1 r2 r3 &key ((:name iname) "Vector-Register")
                                                                ((:mpe impe)))
  (super 'initialize :name iname :mpe impe)
  (vector-set! registers 0 r0)
  (vector-set! registers 1 r1)
  (vector-set! registers 2 r2)
  (vector-set! registers 3 r3)
  self)


;;; BILINEAR-INDEX-REGISTER	class

(define-class bilinear-index-register
  (super-class named-register)
  (instance-variables
    name
    x
    y
    width
    itype
    base
    mipmap
    tile))
    
(define-method (bilinear-index-register 'initialize
                        &key ((:name iname) "Bilinear-Index-Register")
                             ((:mpe impe))
                             ((:x ix))
                             ((:y iy))
                             ((:width iwidth))
                             ((:itype iitype))
                             ((:base ibase))
                             ((:mipmap imipmap))
                             ((:tile itile)))
  (super 'initialize :name iname :mpe impe)
  (set! x ix)
  (set! y iy)
  (set! width iwidth)
  (set! itype iitype)
  (set! base ibase)
  (set! mipmap imipmap)
  (set! tile itile)
  self)

(define-method (bilinear-index-register 'compute-address scale)
  (let ((image-base (base 'value))
        (row-width (logand (width 'value) #x7ff))
        (x-offset (ash (x 'value) -16))
        (y-offset (ash (y 'value) -16))
        (mask (lsh -1 (- (+ (tile 'value) 16)))))
    (let ((offset (+ (* (logand y-offset mask) row-width)
                     (logand x-offset mask))))
      (+ image-base (* offset scale)))))

(define-method (bilinear-index-register 'display
                     &optional (s *standard-output*))
  (format s "~A:" name)
  (format s "  TYP:")
  (write-hex (itype 'value) s)
  (format s "  BAS:")
  (write-hex (base 'value) s)
  (format s "  WID:")
  (write-hex (width 'value) s)
  (format s "  MIP:")
  (write-hex (mipmap 'value) s)
  (format s "  TIL:")
  (write-hex (tile 'value) s)
  (values))


;;; INSTRUCTION class

(define-class instruction
  (instance-variables
    label
    source
    mpe-unit
    debugger
    before-method
    after-method))

(define-method (instruction 'initialize impe &key ((:source isource))
                                                  ((:debugger idebugger))
                                                  before
                                                  after)
  (set! source isource)
  (set! mpe-unit impe)
  (set! debugger idebugger)
  (set! before-method before)
  (set! after-method after)
  self)

(define-method (instruction 'source)
  source)

(define-method (instruction 'debugger)
  debugger)

(define-method (instruction 'before-method)
  before-method)

(define-method (instruction 'after-method)
  after-method)

(define-method (instruction 'display &optional (s *standard-output*))
  (format s "#<Inst")
  (format s "~% Source: ~S" source)
  (when before-method
    (format s "~% Before method: ~S" before-method))
  (when after-method
    (format s "~% After method: ~S" after-method))
  (format s ">")
  (values))


;;; Utility functions

(define (real->32bits val &key (fracbits *default-fracbits*))
  (let* ((int (floor val))
         (frac (- val int)))
    (+ (ash int fracbits) (floor (* frac (expt 2 fracbits))))))
    
(define (32bits->real val &key (fracbits *default-fracbits*))
  (/ val (expt 2 fracbits)))

(define (64bits->real val-high val-low &key (fracbits 32))
  (real-64 val-high val-low fracbits))
         
(define (64bits->string val-high val-low &key (fracbits 32) (fracdigits 10))
  (if (negative? val-high)
    (multiple-value-bind (high low) (neg-64 val-high val-low)
      (string-append "-" (unsigned-64bits->string high low
                                                  :fracbits fracbits
                                                  :fracdigits fracdigits)))
    (unsigned-64bits->string val-high val-low
                            :fracbits fracbits
                            :fracdigits fracdigits)))
    
(define (unsigned-64bits->string val-high val-low &key (fracbits 32) (fracdigits 10))
  (multiple-value-bind (int-high int-low)
                       (unsigned-shift-64 val-high val-low fracbits)
    (multiple-value-bind (frac-high frac-low)
                         (unsigned-shift-64 val-high val-low (- fracbits 64))
      (let ((int-str (int64->string int-high int-low))
            (frac-str (frac64->string frac-high frac-low fracdigits)))
        (string-append int-str "." frac-str)))))

(define (int64->string int-high int-low)
  (let loop ((str "") (high int-high) (low int-low))
    (if (and (= high 0) (= low 0))
      (if (string=? str "") "0" str)
      (multiple-value-bind (high low digit)
                           (/10 high low)
        (loop (string-append (string (digit->char digit)) str)
              high low)))))
              
(define (frac64->string frac-high frac-low max-digits)
  (let loop ((str "") (high frac-high) (low frac-low) (digits 0))
    (if (or (= digits max-digits) (and (= high 0) (= low 0)))
      (if (string=? str "") "0" str)
      (multiple-value-bind (high low digit)
                           (*10 high low)
        (loop (string-append str (string (digit->char digit)))
              high low
              (1+ digits))))))

(define (write-bits n fracbits &optional (stream *standard-output*))
  (let loop ((n n) (i 32))
    (when (> i 0)
      (if (= i fracbits)
        (format stream "."))
      (if (= (remainder i 4) 0)
        (write-char #\Space stream))
      (write-char (if (zero? (logand #x80000000 n)) #\0 #\1) stream)
      (loop (ash n 1) (-1+ i)))))

(define (write-hex n  &optional (stream *standard-output*))
  (write-hex-16 (lsh n -16) stream)
  (write-hex-16 (logand n #xffff) stream))

(define (write-hex-16 n  &optional (stream *standard-output*))
  (write-char #\Space stream)
  (let loop ((n (lsh n 16)) (i 4))
    (when (> i 0)
      (write-char (digit->char (lsh n -28)) stream)
      (loop (ash n 4) (-1+ i)))))

(define (sign-extend-16 value)
  (let ((value-16 (logand value #xFFFF)))
    (if (= (logand value-16 #x8000) 0)
      value-16
      (logior value-16 #xFFFF0000))))
      
(define (decompose-pixel-64 value-high value-low)
  (values (sign-extend-16 (ash value-high -16))
          (sign-extend-16 value-high)
          (sign-extend-16 (ash value-low -16))
          (sign-extend-16 value-low)))
    
(define (saturate-16 value)
  (if (< value 0) 0
    (if (> value 32767) 32767 value)))

(define (rotate-32 n shift)
  (logior (lsh n (- shift))
          (lsh n (- 32 shift))))

(define (boolean->bit b)
  (if b 1 0))
