/* mdma.c - dma routines */

#define __MERLIN_INTERNAL__
#include "merdll.h"

/* local functions */
static void StartRead(MPE *mpe,MScalar iAddr,int dir);
static int ReadHandler(busRequest *r);
static void StartWrite(MPE *mpe,MScalar iAddr,int dir);
static int WriteHandler(busRequest *r);
static cacheElement *FindCacheWork(MPE *mpe);
static int StartCacheFill(mpeBusRequest *req,cacheElement *ce);
static int StartDMATransfer(mpeBusRequest *req);
static void ComputePixelAddress(MPE *mpe,int size,MScalar *pAddr);
static void FinishCacheFill(MPE *mpe,cacheElement *ce);
static void FinishDMATransfer(MPE *mpe);

/* mpeStartDMA_1 - start dma after writing the dma_eaddr register */
void mpeStartDMA_1(MPE *mpe)
{
    int type = mpe->dma_flags.newValue & DMAFlagFunctionMask;
    switch (type) {
    case DMALongRead:
    case DMALongWrite:
        mpe->dma_stat.value = mpe->dma_stat.newValue = mpe->dma_stat.value | dsPending;
        mpeRequestBus(mpe,DMARequestPriority);
        break;
    }
}

/* mpeStartDMA_2 - start dma after writing the dma_x register */
void mpeStartDMA_2(MPE *mpe)
{
    int type = mpe->dma_flags.value & DMAFlagFunctionMask;
    switch (type) {
    case DMAPixelRead:
    case DMAPixelWrite:
    case DMAPixelWriteZ:
        mpe->dma_stat.value = mpe->dma_stat.newValue = mpe->dma_stat.value | dsPending;
        mpeRequestBus(mpe,DMARequestPriority);
        break;
    }
}

/* mpeRequestBus - request the bus */
void mpeRequestBus(MPE *mpe,int priority)
{
    mpe->busRequest.busRequest.priority = priority;
    mmpRequestBus(mpe->mmp,(busRequest *)(void *)&mpe->busRequest);
}

/* mpeBusStartHandler - bus request handler */
int mpeBusStartHandler(busRequest *r)
{
    mpeBusRequest *req = (mpeBusRequest *)r;
    MPE *mpe = req->mpe;
    cacheElement *ce;

    /* first check for cache fills */
    if ((ce = FindCacheWork(mpe)) != NULL)
        return StartCacheFill(req,ce);

    /* otherwise check for dma work */
    else if (mpe->dma_stat.value & dsPending);
        return StartDMATransfer(req);

    /* nothing to do */
    return FALSE;
}

/* FindCacheWork - find a cache element that needs filling */
static cacheElement *FindCacheWork(MPE *mpe)
{
    int count = mpe->cacheElementActiveCount;
    cacheElement *ce = mpe->cacheElements;
    while (--count >= 0) {
	if (ce->transferState == cePending)
	    return ce;
        ++ce;
    }
    return NULL;
}

/* StartCacheFill - start a cache fill */
static int StartCacheFill(mpeBusRequest *req,cacheElement *ce)
{
    MPE *mpe = req->mpe;
    MMP *mmp = mpe->mmp;
    mmp->dma.flags = DMALongRead;
    mmp->dma.eAddr = ce->eAddr;
    mmp->dma.eInc = 4;
    mmp->dma.iAddr = ce->iAddr;
    mmp->dma.iInc = 4;
    mmp->dma.cnt = cePageSize;
    mpe->busRequest.busRequest.clockHandler = ReadHandler;
    return TRUE;
}

/* StartDMATransfer - start a dma transfer */
static int StartDMATransfer(mpeBusRequest *req)
{
    MPE *mpe = req->mpe;
    MMP *mmp = mpe->mmp;
    int type,dir;
    
    /* get the dma flags */
    mmp->dma.flags = mpe->dma_flags.value;
    type = mmp->dma.flags & DMAFlagFunctionMask;
    dir = (mmp->dma.flags & DMAFlagBackwards) ? -1 : 1;

    /* setup the transfer according to type */
    switch (type) {
    case DMALongRead:
        mmp->dma.eAddr = mpe->dma_eaddr.value;
        mmp->dma.eInc = 4;
        mmp->dma.cnt = mpe->dma_length.value & DMALengthMask;
        if (mmp->dma.cnt == 0)
            mmp->dma.cnt = DMAMaxLength;
        StartRead(mpe,mpe->dma_iaddr.value,dir);
        break;
    case DMALongWrite:
        mmp->dma.eAddr = mpe->dma_eaddr.value;
        mmp->dma.eInc = 4;
        mmp->dma.cnt = mpe->dma_length.value & DMALengthMask;
        if (mmp->dma.cnt == 0)
            mmp->dma.cnt = DMAMaxLength;
        StartWrite(mpe,mpe->dma_iaddr.value,dir);
        break;
    case DMAPixelRead:
        ComputePixelAddress(mpe,4,&mmp->dma.eAddr);
        mmp->dma.eInc = 4;
        mmp->dma.cnt = mpe->dma_length.value & DMALengthMask;
        if (mmp->dma.cnt == 0)
            mmp->dma.cnt = DMAMaxLength;
        StartRead(mpe,mpe->dma_iaddr.value,dir);
        break;
    case DMAPixelWrite:
    case DMAPixelWriteZ:
        ComputePixelAddress(mpe,4,&mmp->dma.eAddr);
        mmp->dma.eInc = 4;
        mmp->dma.cnt = mpe->dma_length.value & DMALengthMask;
        if (mmp->dma.cnt == 0)
            mmp->dma.cnt = DMAMaxLength;
        StartWrite(mpe,mpe->dma_iaddr.value,dir);
        break;
    }

    /* this isn't a cache fill */
    mmp->dma.cacheElement = NULL;
    
    /* clear the pending bit and set the active bit */
    mpe->dma_stat.value = mpe->dma_stat.newValue = (mpe->dma_stat.value & ~dsPending) | dsActive;
    return TRUE;
}

/* ComputePixelAddress - compute the address of a pixel in memory */
static void ComputePixelAddress(MPE *mpe,int size,MScalar *pAddr)
{
    MScalar base = mpe->dma_eaddr.value;
    MScalar width = ((mpe->dma_flags.value >> DMAFlagWidthShift) & DMAFlagWidthMask) * 16;
    MScalar x = (mpe->dma_x.value >> 16) & 0x3ff;
    MScalar y = (mpe->dma_y.value >> 16) & 0x3ff;
    *pAddr = base + (y * width + x) * size;
}

/* StartRead - start a bus read */
static void StartRead(MPE *mpe,MScalar iAddr,int dir)
{
    MMP *mmp = mpe->mmp;
    MScalar longAddr = iAddr >> 2;
    mpe->busRequest.busRequest.clockHandler = ReadHandler;
    mmp->dma.iAddr = iAddr;
    mmp->dma.iInc = 4 * dir;
    mmp->dma.bufferPtr = mmp->dma.buffer;
    mmp->dma.bufferCnt = 0;
    if (dir > 0)
        mmp->dma.bufferMax = ((longAddr + dmaBufferSize) & ~3) - longAddr;
    else
        mmp->dma.bufferMax = longAddr - (longAddr & ~3) + 1;
}

/* ReadHandler - bus request read handler */
static int ReadHandler(busRequest *r)
{
    mpeBusRequest *req = (mpeBusRequest *)r;
    MPE *mpe = req->mpe;
    MMP *mmp = mpe->mmp;
    
    /* fetch the next scalar */
    mpeFetchScalar(mpe,mmp->dma.eAddr,mmp->dma.bufferPtr++);
    mmp->dma.eAddr += mmp->dma.eInc;
    --mmp->dma.cnt;
    
    /* check for the buffer being full or the end of the transfer */
    if (++mmp->dma.bufferCnt >= mmp->dma.bufferMax || mmp->dma.cnt == 0) {
	MScalar *p = mmp->dma.buffer;
	int count = mmp->dma.bufferCnt;
        mmp->dma.bufferPtr = mmp->dma.buffer;
	mmp->dma.bufferCnt = 0;
	while (--count >= 0) {
	    mpeStoreScalar(mpe,mmp->dma.iAddr,*p++);
	    mmp->dma.iAddr += mmp->dma.iInc;
	}
	mmp->dma.bufferMax = dmaBufferSize;
    }
    return mmp->dma.cnt != 0;
}

/* StartWrite - start a bus write */
static void StartWrite(MPE *mpe,MScalar iAddr,int dir)
{
    MMP *mmp = mpe->mmp;
    MScalar longAddr = iAddr >> 2;
    mpe->busRequest.busRequest.clockHandler = WriteHandler;
    mmp->dma.iAddr = iAddr;
    mmp->dma.iInc = 4 * dir;
    mmp->dma.bufferPtr = mmp->dma.buffer;
    mmp->dma.bufferCnt = 0;
    if (dir > 0)
        mmp->dma.bufferMax = ((longAddr + dmaBufferSize) & ~3) - longAddr;
    else
        mmp->dma.bufferMax = longAddr - (longAddr & ~3);
}

/* WriteHandler - bus request write handler */
static int WriteHandler(busRequest *r)
{
    mpeBusRequest *req = (mpeBusRequest *)r;
    MPE *mpe = req->mpe;
    MMP *mmp = mpe->mmp;
    int type = mmp->dma.flags & DMAFlagFunctionMask;
    
    /* check for the buffer being empty */
    if (mmp->dma.bufferCnt == 0) {
	MScalar *p = mmp->dma.buffer;
	int count = min(mmp->dma.bufferMax,mmp->dma.cnt);
        mmp->dma.bufferPtr = mmp->dma.buffer;
	mmp->dma.bufferCnt = count;
	while (--count >= 0) {
	    mpeFetchScalar(mpe,mmp->dma.iAddr,p++);
	    mmp->dma.iAddr += mmp->dma.iInc;
	}
        mmp->dma.bufferMax = dmaBufferSize;
    }
    --mmp->dma.bufferCnt;
    
    /* store the next scalar */
    if (type == DMAPixelWriteZ) {
        MScalar newZ,currentValue,currentZ;
        MScalar newValue = *mmp->dma.bufferPtr++;
        mpeFetchScalar(mpe,mmp->dma.eAddr,&currentValue);
        newZ = (UMScalar)newValue >> 24;
        currentZ = (UMScalar)currentValue >> 24;
        if ((newZ < currentZ && (mmp->dma.flags & DMAZTypeLT))
        ||  (newZ == currentZ && (mmp->dma.flags & DMAZTypeEQ))
        ||  (newZ > currentZ && (mmp->dma.flags & DMAZTypeGT)))
            mpeStoreScalar(mpe,mmp->dma.eAddr,newValue);
    }
    else
        mpeStoreScalar(mpe,mmp->dma.eAddr,*mmp->dma.bufferPtr++);
    mmp->dma.eAddr += mmp->dma.eInc;
    return --mmp->dma.cnt != 0;
}

/* mpeBusStopHandler - bus request handler */
void mpeBusStopHandler(busRequest *r)
{
    mpeBusRequest *req = (mpeBusRequest *)r;
    MPE *mpe = req->mpe;
    MMP *mmp = mpe->mmp;

    /* finish the current transfer */
    if (mmp->dma.cacheElement)
        FinishCacheFill(mpe,mmp->dma.cacheElement);
    else
        FinishDMATransfer(mpe);

    /* check for another pending transfer */
    if (FindCacheWork(mpe) || mpe->dma_stat.value & dsPending)
        mpeRequestBus(mpe,DMARequestPriority);
}

/* FinishCacheFill - finish a cache fill */
static void FinishCacheFill(MPE *mpe,cacheElement *ce)
{
    registerScoreboard *sb = mpe->scoreboard;
    int count = scoreboardCount;
    while (--count >= 0) {
        if (sb->cacheElement == ce)
            mpeLoadRegisterFromMemory(mpe,ce->iAddr + sb->offset,sb->state,sb->base);
        ++sb;
    }
    ce->transferState = ceIdle;
    ce->clean = TRUE;
}

/* FinishDMATransfer - finish a dma transfer */
static void FinishDMATransfer(MPE *mpe)
{
    mpe->dma_stat.value = mpe->dma_stat.newValue = mpe->dma_stat.value & ~dsActive;
}
