/* memulate.c - merlin emulator */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#define __MERLIN_INTERNAL__
#include "merdll.h"

/* local functions */
static void ClockBus(MMP *mmp);
static void RemoveBusRequest(MMP *mmp,busRequest *req);
static int ExecuteInstruction(MPE *mpe);
static int CollectInstruction(MPE *mpe);
static int ParseInstruction(MPE *mpe,MInst *i,char *text);
static UINT32 ConvertHex(char *token);
static UINT32 ConvertDecimal(char *token);
static scalarRegister *FindScalarRegister(MPE *mpe,char *name);
static vectorRegister *FindVectorRegister(MPE *mpe,char *name);
static accRegister *FindAccumulator(MPE *mpe,char *name);
static ioRegister *FindioRegister(MPE *mpe,char *name);
static int FindOtherRegister(MPE *mpe,char *name,UINT32 *pValue);
static char *NextToken(char *str,char *token);
static int CompareNames(char *name1,char *name2);

/* mmpClock - emulate a single clock tick */
EXPORT void mmpClock(MMP *mmp)
{
    MPE *mpe;
    int n;

    /* initialize each of the mpes for this tick */
    mpe = mmp->mpe;
    n = mmp->mpeCount;
    while (--n >= 0) {
        mpe->flags &= ~mpeLocalRAM;
        ++mpe;
    }

    /* clock the bus */
    ClockBus(mmp);

    /* clock each of the mpes */
    mpe = mmp->mpe;
    n = mmp->mpeCount;
    while (--n >= 0) {
        mpeClock(mpe);
        ++mpe;
    }

    /* update the elapsed time */
    ++mmp->pub.elapsedTicks;

    /* call the stat hook */
    if (mmp->pub.statHook) {
        for (n = 0; n < mmp->mpeCount; ++n)
            mmp->statSamples[n] = mmp->mpe[n].statSample;
        (*mmp->pub.statHook)(mmp,mmp->statSamples);
    }
}

/* mmpRequestBus - queue a bus request */
void mmpRequestBus(MMP *mmp,busRequest *req)
{
    busRequest **pNext = &mmp->busRequesters;
    busRequest *r;

    /* remove the request in case it's already in the queue */
    RemoveBusRequest(mmp,req);

    /* add the request in priority order */
    while ((r = *pNext) != NULL) {
        if (req->priority > r->priority)
            break;
        pNext = &r->next;
    }

    /* insert the request into the queue */
    req->next = r;
    *pNext = req;
}

/* RemoveBusRequest - remove a bus request from the queue */
static void RemoveBusRequest(MMP *mmp,busRequest *req)
{
    busRequest **pNext = &mmp->busRequesters;
    busRequest *r;

    /* find the request and remove it from the queue */
    while ((r = *pNext) != NULL) {
        if (r == req) {
            *pNext = r->next;
            break;
        }
        pNext = &r->next;
    }
}

#define BusSlotSize 96

/* ClockBus - emulate a single bus clock tick */
static void ClockBus(MMP *mmp)
{
    busRequest *master = mmp->busMaster;
    
    /* check for a current bus master */
    if (master) {

        /* clock the master and stop if it is done or at the end of the slot */
        if (!(*master->clockHandler)(master)
        ||  ++mmp->busTimer >= BusSlotSize) {
            (*master->stopHandler)(mmp->busMaster);
            mmp->busMaster = NULL;
        }
    }

    /* choose a new bus master */
    else {
        while ((master = mmp->busRequesters) != NULL) {
            mmp->busRequesters = master->next;
            if ((*master->startHandler)(master)) {
                mmp->busMaster = master;
                mmp->busTimer = 0;
                break;
            }
        }
    }
}

/* mpeRunState - get the run state of an mpe */
EXPORT int mpeRunState(MPE *mpe)
{
    return mpe->runState;
}

/* mpeSetRunState - set the run state of an mpe */
EXPORT int mpeSetRunState(MPE *mpe,int state)
{
    int oldState = mpe->runState;
    mpe->runState = state;
    return oldState;
}

/* mpeSingleStep - cause the mpe to execute a single instruction */
EXPORT int mpeSingleStep(MPE *mpe)
{
    int oldState = mpe->runState;
    int sts;

    /* start the mpe running */
    mpe->runState = rsRun;

    /* dma won't interfere on single step */
    mpe->flags &= ~mpeLocalRAM;

    /* clock the mpe so one instruction executes */
    sts = mpeClock(mpe);

    /* restore the previous state if still running */
    if (mpe->runState == rsRun)
        mpe->runState = oldState;

    /* return status */
    return sts;
}

/* mpeReadNext - read the next instruction register */
void mpeReadNext(MPE *mpe)
{
    mpe->next.value = mpe->executePacket->addr;
}

/* mpeWritePC - write the pc register */
void mpeWritePC(MPE *mpe)
{
    /* flush the pipeline */
    mpe->executePacket = mpe->fetchPacket;
    
    /* reprime the pipeline */
    CollectInstruction(mpe);
    
    /* latch the new pc value */
    mpe->pc.value = mpe->pc.newValue;
}

/* mpeClock - emulate a single clock tick */
EXPORT int mpeClock(MPE *mpe)
{
    /* clear the stat sample */
    mpe->statSample = 0;
    
    /* execute an instruction if in run or step mode */
    switch (mpe->runState) {
    case rsStop:
        return FALSE;
    case rsRun:
        if (mpe->executePacket == mpe->fetchPacket) {
            CollectInstruction(mpe);
            mpe->pc.value = mpe->pc.newValue;
            return FALSE;
        }
        else
            return ExecuteInstruction(mpe);
    }
}

/* ExecuteInstruction - execute a single instruction */
static int ExecuteInstruction(MPE *mpe)
{
    InstructionPacket *packet;
    MInst **pi,*i;
    int count;

    /* get the next instruction in the pipeline */
    packet = mpe->executePacket;

    /* stall on implied nops */
    if (packet->delaySlot == dsImpliedNOP) {
        packet->delaySlot = dsNone;

        /* fake executing the instruction */
        mpe->executePacket = packet->next;
        CollectInstruction(mpe);
        mpe->pc.value = mpe->pc.newValue;
        return TRUE;
    }

    /* stall on accesses to scoreboarded registers */
    else if (packet->registerMask & mpe->registerMask)
        return FALSE;

    /* perform any loads staged in the last instruction */
    PerformStagedLoads(&mpe->loadBuffer);
    PerformStagedLoads(&mpe->multiplyBuffer);

    /* update the multiply overflow condition code */
    if (mpe->multiplyOverflowBuffer != mvNone) {
        mpe->cc.newValue &= ~ccMV;
        if (mpe->multiplyOverflowBuffer == mvOverflow)
            mpe->cc.newValue |= ccMV;
        mpe->multiplyOverflowBuffer = mvNone;
    }
    
    /* halt on accesses to registers on the tick after a load */
    if (packet->registerMask & mpe->loadBuffer.mask) {
        if (!CallHaltHook(mpe,hhBadLoad,packet->addr))
            mpe->runState = rsStop;
        return FALSE;
    }
    
    /* halt on accesses to registers on the tick after a multiply */
    if (packet->registerMask & mpe->multiplyBuffer.mask) {
        if (!CallHaltHook(mpe,hhBadMultiply,packet->addr))
            mpe->runState = rsStop;
        return FALSE;
    }
    
    /* check for breakpoints and before/after methods */
    i = packet->instructions[0]; /* first instruction has the flags */
    if (i->flags & ifBreak) {
        if (!CallHaltHook(mpe,hhBreak,packet->addr))
            mpe->runState = rsStop;
        return FALSE;
    }

    /* assume the instruction won't stall */
    mpe->flags &= ~mpeStalled;

    /* execute the compound instruction */
    pi = packet->instructions;
    count = packet->instCount;
    while (--count >= 0) {
        i = *pi++;
        (*i->entry->handler)(mpe,i->args);
    }

    /* check to see if the mpe stalled */
    if (mpe->flags & mpeStalled)
        return FALSE;

    /* move ahead to the next packet */
    mpe->executePacket = packet->next;

    /* fill the instruction pipeline */
    CollectInstruction(mpe);

    /* latch the new values (except pc and rz) */
    LatchNewValues(mpe);

    /* handle mv_spr pc,rz */
    if (mpe->saveReturnP) {
        if (mpe->branchType != btNone && mpe->branchDelaySlot == dsImpliedNOP)
            mpe->rz.newValue = mpe->pc.value;
        else
            mpe->rz.newValue = mpe->pc.newValue;
        mpe->saveReturnP = FALSE;
    }

    /* handle branches */
    if (mpe->branchType != btNone) {
        if (packet->delaySlot == dsNone
        &&  CheckConditionCode(mpe,mpe->branchCondition)) {
            mpe->executePacket->delaySlot = mpe->branchDelaySlot;
            if (mpe->branchType == btAbsolute)
                mpe->pc.newValue = mpe->branchTarget;
            else
                mpe->pc.newValue = mpe->pc.value + mpe->branchTarget * bytesPerInstruction;
        }
        mpe->branchType = btNone;
    }

    /* latch pc and rz */
    mpe->pc.value = mpe->pc.newValue;
    mpe->rz.value = mpe->rz.newValue;

    /* return successfully */
    return TRUE;
}

/* CallHaltHook - call the halt hook function */
int CallHaltHook(MPE *mpe,int code,MScalar addr)
{
    if (!mpe->mmp->pub.haltHook)
        return FALSE;
    return (*mpe->mmp->pub.haltHook)(mpe,code,addr);
}

/* CollectInstruction - collect a compound instruction */
static int CollectInstruction(MPE *mpe)
{
    MScalar flags,pc = mpe->pc.value;
    InstructionPacket *packet;
    MInst **pdst,*i,*top;

    /* initialize the packet */
    packet = mpe->fetchPacket;
    mpe->fetchPacket = packet->next;
    pdst = packet->instructions;
    packet->addr = pc;
    packet->registerMask = 0;
    packet->instCount = 0;
    packet->delaySlot = dsNone;
    
    /* fetch the next instuction */
    if (!mpeInstructionAddress(mpe,pc,&i,&top))
        return FALSE;
    
    /* collect the compound instruction */
    while (++packet->instCount <= MAXINSTPERTICK) {

        /* make sure there is another instruction available */
        if (i >= top)
            return FALSE;

        /* get the instruction flags */
        flags = i->flags;
            
        /* store the instruction */
        packet->registerMask |= i->registerMask;
        *pdst++ = i;
            
        /* update the location counter */
        if (i->flags & if48Bits) {
            pc += 3 * bytesPerInstruction;
            i += 3;
        }
        else if (i->flags & if32Bits) {
            pc += 2 * bytesPerInstruction;
            i += 2;
        }
        else {
            pc += bytesPerInstruction;
            i += 1;
        }

        /* check for the end of the packet */
        if (flags & ifStop)
            break;
    }

    /* store the updated pc and return successfully */
    mpe->pc.newValue = pc;
    return TRUE;
}

/* LatchNewValues - latch new register values at the end of a tick */
void LatchNewValues(MPE *mpe)
{
    scalarRegister *r = mpe->scalarRegisters;
    int cnt;

    /* latch the scalar registers */
    for (cnt = scalarRegisterCount; --cnt >= 0; ++r)
        r->value = r->newValue;

    /* latch the accumulator */
    mpe->ac.valueHigh = mpe->ac.newValueHigh;
    mpe->ac.valueLow = mpe->ac.newValueLow;

    /* latch the special registers */
    mpe->sp.value = mpe->sp.newValue;
    mpe->cc.value = mpe->cc.newValue;
    mpe->rc0.value = mpe->rc0.newValue;
    mpe->rc1.value = mpe->rc1.newValue;
    mpe->sv_shift.value = mpe->sv_shift.newValue;
    mpe->ac_shift.value = mpe->ac_shift.newValue;
    mpe->rx.value = mpe->rx.newValue;
    mpe->ry.value = mpe->ry.newValue;
    mpe->xy_width.value = mpe->xy_width.newValue;
    mpe->xy_itype.value = mpe->xy_itype.newValue;
    mpe->xy_base.value = mpe->xy_base.newValue;
    mpe->xy_mipmap.value = mpe->xy_mipmap.newValue;
    mpe->xy_tile.value = mpe->xy_tile.newValue;
    mpe->ru.value = mpe->ru.newValue;
    mpe->rv.value = mpe->rv.newValue;
    mpe->uv_width.value = mpe->uv_width.newValue;
    mpe->uv_itype.value = mpe->uv_itype.newValue;
    mpe->uv_base.value = mpe->uv_base.newValue;
    mpe->uv_mipmap.value = mpe->uv_mipmap.newValue;
    mpe->uv_tile.value = mpe->uv_tile.newValue;
    mpe->dma_eaddr.value = mpe->dma_eaddr.newValue;
    mpe->dma_flags.value = mpe->dma_flags.newValue;
    mpe->dma_length.value = mpe->dma_length.newValue;
    mpe->dma_iaddr.value = mpe->dma_iaddr.newValue;
    mpe->dma_x.value = mpe->dma_x.newValue;
    mpe->dma_y.value = mpe->dma_y.newValue;
    mpe->dma_stat.value = mpe->dma_stat.newValue;
}

/* mpeRegisterValue - get the value of a register */
EXPORT int mpeRegisterValue(MPE *mpe,char *name,MScalar *pValue)
{
    scalarRegister *sr;
    ioRegister *r;
    
    /* check the scalar registers */
    if ((sr = FindScalarRegister(mpe,name)) != NULL) {
        *pValue = sr->value;
        return TRUE;
    }

    /* check the special registers */
    else if ((r = FindioRegister(mpe,name)) != NULL) {
        if (r->readHandler)
            (*r->readHandler)(mpe);
        *pValue = r->value;
        return TRUE;
    }

    /* not found */
    return FALSE;
}

/* mpeSetRegisterValue - set the value of a register */
EXPORT int mpeSetRegisterValue(MPE *mpe,char *name,MScalar value)
{
    scalarRegister *sr;
    ioRegister *r;
    
    /* check the scalar registers */
    if ((sr = FindScalarRegister(mpe,name)) != NULL) {
        sr->value = sr->newValue = value;
        return TRUE;
    }

    /* check the special registers */
    else if ((r = FindioRegister(mpe,name)) != NULL) {
        r->value = r->newValue = value;
        if (r->writeHandler)
            (*r->writeHandler)(mpe);
        return TRUE;
    }

    /* not found */
    return FALSE;
}

/* mpeAccumulatorValue - get the value of the accumulator */
EXPORT void mpeAccumulatorValue(MPE *mpe,MScalar *pValueHigh,MScalar *pValueLow)
{
    *pValueHigh = mpe->ac.valueHigh;
    *pValueLow = mpe->ac.valueLow;
}

/* mpeSetAccumulatorValue - set the value of the accumulator */
EXPORT void mpeSetAccumulatorValue(MPE *mpe,MScalar valueHigh,MScalar valueLow)
{
    mpe->ac.valueHigh = mpe->ac.newValueHigh = valueHigh;
    mpe->ac.valueLow = mpe->ac.newValueLow = valueLow;
}

/* CheckConditionCode - check a condition code */
int CheckConditionCode(MPE *mpe,int cc)
{
    MScalar ccv = mpe->cc.value;
    int c = (ccv & ccC) != 0;
    int n = (ccv & ccN) != 0;
    int v = (ccv & ccV) != 0;
    int z = (ccv & ccZ) != 0;
    switch (cc) {
    case ccT:   return TRUE;
    case ccEQ:  return z;  
    case ccNE:  return !z;
    case ccLT:  return (n && !v) || (!n && v);
    case ccLE:  return z || (n && !v) || (!n && v);
    case ccGT:  return (n && v && !z) || (!n && !v && !z);
    case ccGE:  return (n && v) || (!n && !v);
    case ccCC:  return !c;
    case ccCS:  return c;
    case ccVC:  return !v;
    case ccVS:  return v;
    case ccC0EQ:return (mpe->cc.value & ccC0Z) != 0;
    case ccC0NE:return (mpe->cc.value & ccC0Z) == 0;
    case ccC1NE:return (mpe->cc.value & ccC1Z) == 0;
    case ccMVS: return (mpe->cc.value & ccMV) != 0;
    case ccMVC: return (mpe->cc.value & ccMV) == 0;
    }
}

/* mpeBreakpointP - check for a breakpoint on an instruction */
EXPORT int mpeBreakpointP(MPE *mpe,MScalar addr)
{
    MInst *i,*top;

    /* get the instruction address */
    if (!mpeInstructionAddress(mpe,addr,&i,&top))
        return FALSE;

    /* check for a breakpoint on this instruction */
    return (i->flags & ifBreak) != 0;
}

/* mpeSetBreakpoint - set a breakpoint on an instruction */
EXPORT int mpeSetBreakpoint(MPE *mpe,MScalar addr)
{
    MInst *i,*top;

    /* get the instruction address */
    if (!mpeInstructionAddress(mpe,addr,&i,&top))
        return FALSE;

    /* set the breakpoint and return successfully */
    i->flags |= ifBreak;
    return TRUE;
}

/* mpeClearBreakpoint - clear a breakpoint on an instruction */
EXPORT int mpeClearBreakpoint(MPE *mpe,MScalar addr)
{
    MInst *i,*top;

    /* get the instruction address */
    if (!mpeInstructionAddress(mpe,addr,&i,&top))
        return FALSE;

    /* clear the breakpoint and return successfully */
    i->flags &= ~ifBreak;
    return TRUE;
}

/* mpeStoreInstruction - store an instruction in mpe memory */
EXPORT int mpeStoreInstruction(MPE *mpe,
                               MScalar addr,
                               UINT32 flags,
                               char *text)
{
    MInst *i,*top;

    /* get the instruction address */
    if (!mpeInstructionAddress(mpe,addr,&i,&top))
        return FALSE;

    /* initialize the instruction */
    i->flags = flags;
    i->registerMask = 0;
    if (!ParseInstruction(mpe,i,text))
        return FALSE;

    /* return successfully */
    return TRUE;
}

/* ParseInstruction - parse the text form of an iinstruction */
static int ParseInstruction(MPE *mpe,MInst *i,char *text)
{
    OpcodeEntry *entry;
    scalarRegister *sr;
    vectorRegister *vr;
    accRegister *ac;
    ioRegister *r;
    char token[256];
    int n = 0;

    /* setup the opcode */
    if ((text = NextToken(text,token)) != NULL
    &&  (entry = FindOpcodeEntry(token)) != NULL) {
        i->flags |= entry->flags;
        i->entry = entry;
    }

    /* unknown opcode */
    else {
        i->entry = NULL;
        return FALSE;
    }

    /* setup each argument */
    while ((text = NextToken(text,token)) != NULL) {

        /* handle numeric arguments */
        if (isdigit(token[0]) || token[0] == '-')
            i->args[n] = ConvertDecimal(token);

        /* handle hex numeric arguments */
        else if (token[0] == '$')
            i->args[n] = ConvertHex(token);

        /* check the scalar registers */
        else if ((sr = FindScalarRegister(mpe,token)) != NULL) {
            i->registerMask |= sr->registerMask;
            i->args[n] = (UINT32)sr;
        }

        /* check the vector (and pixel) registers */
        else if ((vr = FindVectorRegister(mpe,token)) != NULL) {
            i->registerMask |= vr->registerMask;
            i->args[n] = (UINT32)vr->base;
        }

        /* check the accumulators */
        else if ((ac = FindAccumulator(mpe,token)) != NULL)
            i->args[n] = (UINT32)ac;

        /* check the special registers */
        else if ((r = FindioRegister(mpe,token)) != NULL)
            i->args[n] = (UINT32)r;

        /* check the other registers */
        else if (FindOtherRegister(mpe,token,&i->args[n]))
            ;

        /* shouldn't ever be reached */
        else
            return FALSE;
        ++n;
    }

    /* return successfully */
    return TRUE;
}

/* FindOpcodeEntry - find an opcode by name */
OpcodeEntry *FindOpcodeEntry(char *name)
{
    OpcodeEntry *e;
    for (e = optable; e->name != NULL; ++e)
        if (CompareNames(name,e->name) == 0)
            return e;
    return NULL;
}

/* ConvertHex - convert a hex number */
static UINT32 ConvertHex(char *token)
{
    UINT32 value;
    sscanf(&token[1],"%lx",&value);
    return value;
}

/* ConvertDecimal - convert a decimal token */
static UINT32 ConvertDecimal(char *token)
{
    return (UINT32)atol(token);
}

/* FindScalarRegister - find a scalar register */
static scalarRegister *FindScalarRegister(MPE *mpe,char *name)
{
    if (CompareNames(name,"r0") == 0)
        return &mpe->scalarRegisters[0];
    else if (CompareNames(name,"r1") == 0)
        return &mpe->scalarRegisters[1];
    else if (CompareNames(name,"r2") == 0)
        return &mpe->scalarRegisters[2];
    else if (CompareNames(name,"r3") == 0)
        return &mpe->scalarRegisters[3];
    else if (CompareNames(name,"r4") == 0)
        return &mpe->scalarRegisters[4];
    else if (CompareNames(name,"r5") == 0)
        return &mpe->scalarRegisters[5];
    else if (CompareNames(name,"r6") == 0)
        return &mpe->scalarRegisters[6];
    else if (CompareNames(name,"r7") == 0)
        return &mpe->scalarRegisters[7];
    else if (CompareNames(name,"r8") == 0)
        return &mpe->scalarRegisters[8];
    else if (CompareNames(name,"r9") == 0)
        return &mpe->scalarRegisters[9];
    else if (CompareNames(name,"r10") == 0)
        return &mpe->scalarRegisters[10];
    else if (CompareNames(name,"r11") == 0)
        return &mpe->scalarRegisters[11];
    else if (CompareNames(name,"r12") == 0)
        return &mpe->scalarRegisters[12];
    else if (CompareNames(name,"r13") == 0)
        return &mpe->scalarRegisters[13];
    else if (CompareNames(name,"r14") == 0)
        return &mpe->scalarRegisters[14];
    else if (CompareNames(name,"r15") == 0)
        return &mpe->scalarRegisters[15];
    else if (CompareNames(name,"r16") == 0)
        return &mpe->scalarRegisters[16];
    else if (CompareNames(name,"r17") == 0)
        return &mpe->scalarRegisters[17];
    else if (CompareNames(name,"r18") == 0)
        return &mpe->scalarRegisters[18];
    else if (CompareNames(name,"r19") == 0)
        return &mpe->scalarRegisters[19];
    else if (CompareNames(name,"r20") == 0)
        return &mpe->scalarRegisters[20];
    else if (CompareNames(name,"r21") == 0)
        return &mpe->scalarRegisters[21];
    else if (CompareNames(name,"r22") == 0)
        return &mpe->scalarRegisters[22];
    else if (CompareNames(name,"r23") == 0)
        return &mpe->scalarRegisters[23];
    else if (CompareNames(name,"r24") == 0)
        return &mpe->scalarRegisters[24];
    else if (CompareNames(name,"r25") == 0)
        return &mpe->scalarRegisters[25];
    else if (CompareNames(name,"r26") == 0)
        return &mpe->scalarRegisters[26];
    else if (CompareNames(name,"r27") == 0)
        return &mpe->scalarRegisters[27];
    else if (CompareNames(name,"r28") == 0)
        return &mpe->scalarRegisters[28];
    else if (CompareNames(name,"r29") == 0)
        return &mpe->scalarRegisters[29];
    else if (CompareNames(name,"r30") == 0)
        return &mpe->scalarRegisters[30];
    else if (CompareNames(name,"r31") == 0)
        return &mpe->scalarRegisters[31];

    /* not found */
    return NULL;
}

/* FindVectorRegister - find a vector (or pixel) register */
static vectorRegister *FindVectorRegister(MPE *mpe,char *name)
{
    /* check the vector registers */
    if (CompareNames(name,"v0") == 0)
        return &mpe->vectorRegisters[0];
    else if (CompareNames(name,"v1") == 0)
        return &mpe->vectorRegisters[1];
    else if (CompareNames(name,"v2") == 0)
        return &mpe->vectorRegisters[2];
    else if (CompareNames(name,"v3") == 0)
        return &mpe->vectorRegisters[3];
    else if (CompareNames(name,"v4") == 0)
        return &mpe->vectorRegisters[4];
    else if (CompareNames(name,"v5") == 0)
        return &mpe->vectorRegisters[5];
    else if (CompareNames(name,"v6") == 0)
        return &mpe->vectorRegisters[6];
    else if (CompareNames(name,"v7") == 0)
        return &mpe->vectorRegisters[7];

    /* check the pixel registers */
    else if (CompareNames(name,"p0") == 0)
        return &mpe->vectorRegisters[0];
    else if (CompareNames(name,"p1") == 0)
        return &mpe->vectorRegisters[1];
    else if (CompareNames(name,"p2") == 0)
        return &mpe->vectorRegisters[2];
    else if (CompareNames(name,"p3") == 0)
        return &mpe->vectorRegisters[3];
    else if (CompareNames(name,"p4") == 0)
        return &mpe->vectorRegisters[4];
    else if (CompareNames(name,"p5") == 0)
        return &mpe->vectorRegisters[5];
    else if (CompareNames(name,"p6") == 0)
        return &mpe->vectorRegisters[6];
    else if (CompareNames(name,"p7") == 0)
        return &mpe->vectorRegisters[7];

    /* not found */
    return NULL;
}

/* FindAccumulator - find an accumulator (just for debugging now) */
static accRegister *FindAccumulator(MPE *mpe,char *name)
{
    /* the accumulator */
    if (CompareNames(name,"ac") == 0)
        return &mpe->ac;

    /* not found */
    return NULL;
}

/* FindioRegister - find a special register */
static ioRegister *FindioRegister(MPE *mpe,char *name)
{
    if (CompareNames(name,"pc") == 0)
        return &mpe->pc;
    else if (CompareNames(name,"next") == 0)
        return &mpe->next;
    else if (CompareNames(name,"rz") == 0)
        return &mpe->rz;
    else if (CompareNames(name,"sp") == 0)
        return &mpe->sp;
    else if (CompareNames(name,"cc") == 0)
        return &mpe->cc;
    else if (CompareNames(name,"rc0") == 0)
        return &mpe->rc0;
    else if (CompareNames(name,"rc1") == 0)
        return &mpe->rc1;
    else if (CompareNames(name,"rx") == 0)
        return &mpe->rx;
    else if (CompareNames(name,"ry") == 0)
        return &mpe->ry;
    else if (CompareNames(name,"ru") == 0)
        return &mpe->ru;
    else if (CompareNames(name,"rv") == 0)
        return &mpe->rv;
    else if (CompareNames(name,"sv_shift") == 0)
        return &mpe->sv_shift;
    else if (CompareNames(name,"ac_shift") == 0)
        return &mpe->ac_shift;
    else if (CompareNames(name,"xy_width") == 0)
        return &mpe->xy_width;
    else if (CompareNames(name,"xy_itype") == 0)
        return &mpe->xy_itype;
    else if (CompareNames(name,"xy_base") == 0)
        return &mpe->xy_base;
    else if (CompareNames(name,"xy_mipmap") == 0)
        return &mpe->xy_mipmap;
    else if (CompareNames(name,"xy_tile") == 0)
        return &mpe->xy_tile;
    else if (CompareNames(name,"uv_width") == 0)
        return &mpe->uv_width;
    else if (CompareNames(name,"uv_itype") == 0)
        return &mpe->uv_itype;
    else if (CompareNames(name,"uv_base") == 0)
        return &mpe->uv_base;
    else if (CompareNames(name,"uv_mipmap") == 0)
        return &mpe->uv_mipmap;
    else if (CompareNames(name,"uv_tile") == 0)
        return &mpe->uv_tile;
    else if (CompareNames(name,"dma_eaddr") == 0)
        return &mpe->dma_eaddr;
    else if (CompareNames(name,"dma_flags") == 0)
        return &mpe->dma_flags;
    else if (CompareNames(name,"dma_length") == 0)
        return &mpe->dma_length;
    else if (CompareNames(name,"dma_iaddr") == 0)
        return &mpe->dma_iaddr;
    else if (CompareNames(name,"dma_x") == 0)
        return &mpe->dma_x;
    else if (CompareNames(name,"dma_y") == 0)
        return &mpe->dma_y;
    else if (CompareNames(name,"dma_stat") == 0)
        return &mpe->dma_stat;

    /* not found */
    return NULL;
}

/* FindOtherRegister - find a bilinear register, dma group, etc. */
static int FindOtherRegister(MPE *mpe,char *name,UINT32 *pValue)
{
    /* bilinear registers */
    if (CompareNames(name,"xy") == 0)
        *pValue = (UINT32)&mpe->xy;
    else if (CompareNames(name,"uv") == 0)
        *pValue = (UINT32)&mpe->uv;
    
    /* dma vector registers */
    else if (CompareNames(name,"dma_1") == 0)
        *pValue = (UINT32)&mpe->dma_1;
    else if (CompareNames(name,"dma_2") == 0)
        *pValue = (UINT32)&mpe->dma_2;
    
    /* data ram/rom indicators */
    else if (CompareNames(name,"dtram") == 0)
        *pValue = (UINT32)dtRAMSpace;
    else if (CompareNames(name,"dtrom") == 0)
        *pValue = (UINT32)dtROMSpace;

    /* condition codes */
    else if (CompareNames(name,"cc_t") == 0)
        *pValue = (UINT32)ccT;
    else if (CompareNames(name,"cc_eq") == 0)
        *pValue = (UINT32)ccEQ;
    else if (CompareNames(name,"cc_ne") == 0)
        *pValue = (UINT32)ccNE;
    else if (CompareNames(name,"cc_lt") == 0)
        *pValue = (UINT32)ccLT;
    else if (CompareNames(name,"cc_le") == 0)
        *pValue = (UINT32)ccLE;
    else if (CompareNames(name,"cc_gt") == 0)
        *pValue = (UINT32)ccGT;
    else if (CompareNames(name,"cc_ge") == 0)
        *pValue = (UINT32)ccGE;
    else if (CompareNames(name,"cc_cc") == 0)
        *pValue = (UINT32)ccCC;
    else if (CompareNames(name,"cc_cs") == 0)
        *pValue = (UINT32)ccCS;
    else if (CompareNames(name,"cc_vc") == 0)
        *pValue = (UINT32)ccVC;
    else if (CompareNames(name,"cc_vs") == 0)
        *pValue = (UINT32)ccVS;
    else if (CompareNames(name,"cc_c0eq") == 0)
        *pValue = (UINT32)ccC0EQ;
    else if (CompareNames(name,"cc_c0ne") == 0)
        *pValue = (UINT32)ccC0NE;
    else if (CompareNames(name,"cc_c1ne") == 0)
        *pValue = (UINT32)ccC1NE;
    else if (CompareNames(name,"cc_mvs") == 0)
        *pValue = (UINT32)ccMVS;
    else if (CompareNames(name,"cc_mvc") == 0)
        *pValue = (UINT32)ccMVC;
    
    /* stallops */
    else if (CompareNames(name,"so_nop") == 0)
        *pValue = (UINT32)soNOP;
    else if (CompareNames(name,"so_next") == 0)
        *pValue = (UINT32)soNEXT;

    /* nothing */
    else if (CompareNames(name,"nil") == 0)
        *pValue = (UINT32)0;

    /* not found */
    else
        return FALSE;
    
    /* return successfully */
    return TRUE;
}

/* NextToken - get the next space delimited token */
static char *NextToken(char *str,char *token)
{
    int ch;

    /* skip leading spaces */
    while ((ch = *str) != '\0' && isspace(ch))
        ++str;

    /* check for end of string */
    if (ch == '\0')
        return NULL;

    /* collect the next token */
    while (ch != '\0' && !isspace(ch)) {
        *token++ = islower(ch) ? toupper(ch) : ch;
        ch = *++str;
    }
    *token = '\0';

    /* return the remainder of the string */
    return str;
}

/* CompareNames - compare names to see if they are equal */
static int CompareNames(char *name1,char *name2)
{
    int ch1,ch2;
    for (;;) {
        ch1 = *name1++;
        ch2 = *name2++;
        if (!ch1 || !ch2) break;
        if (isupper(ch1)) ch1 = tolower(ch1);
        if (isupper(ch2)) ch2 = tolower(ch2);
        if (ch1 != ch2)
            return ch1 < ch2 ? -1 : 1;
    }
    return ch1 == ch2 ? 0 : (ch1 ? 1 : -1);
}