/* memulate.h - definitions for the merlin emulator */

#ifndef __MEMULATE_H__
#define __MEMULATE_H__

#ifndef TRUE
#define TRUE	1
#define FALSE	0
#endif

/* scalar register count */
#define scalarRegisterCount 32

/* macro to generate register bit numbers */
#define RegBit(n)           (1 << (n))

/* scalar register structure */
typedef struct scalarRegister {
    char name[20];              /* register name */
    UINT32 registerMask;        /* bit corresponding to this register */
    MScalar value;		/* current value */
    MScalar newValue;		/* value to be latched at end of current tick */    
} scalarRegister;

/* vector register count */
#define vectorRegisterCount (scalarRegisterCount / 4)

/* vector register structure */
typedef struct vectorRegister {
    char name[20];              /* register name */
    UINT32 registerMask;        /* bits corresponding to this register */
    scalarRegister *base;       /* first scalar register in this vector */
} vectorRegister;

/* accumulator structure */
typedef struct accRegister {
    char name[20];              /* register name */
    MScalar valueHigh;          /* high 32 bits of value */
    MScalar valueLow;           /* low 32 bits of value */
    MScalar newValueHigh;       /* high value to be latched at end of current tick */
    MScalar newValueLow;        /* low value to be latched at end of current tick */
} accRegister;

/* i/o register structure */
typedef struct ioRegister {
    char name[20];              /* register name */
    MScalar value;              /* current value */
    MScalar newValue;           /* value to be latched at end of current tick */
    void (*readHandler)(struct MPE *mpe);
    void (*writeHandler)(struct MPE *mpe);
} ioRegister;

/* i/o vector register */
typedef struct ioVectorRegister {
    char name[20];              /* register name */
    ioRegister *data[4];        /* data registers */
    void (*readHandler)(struct MPE *mpe);
    void (*writeHandler)(struct MPE *mpe);
} ioVectorRegister;

/* bilinear register structure */
typedef struct bilinearRegister {
    char name[20];              /* register name */
    ioRegister *x;              /* x/u register */
    ioRegister *y;              /* y/v register */
    ioRegister *width;
    ioRegister *itype;
    ioRegister *base;
    ioRegister *mipmap;
    ioRegister *tile;
} bilinearRegister;

/* memory segment structure */
typedef struct memSegment {
    char *data;
    char *dataTop;
    MScalar base;
    MScalar top;
} memSegment;

/* bus request structure */
typedef struct busRequest {
    struct busRequest *next;
    int priority;
    int (*startHandler)(struct busRequest *req);
    int (*clockHandler)(struct busRequest *req);
    void (*stopHandler)(struct busRequest *req);
} busRequest;

/* mpe bus request structure */
typedef struct mpeBusRequest {
    busRequest busRequest;
    struct MPE *mpe;
} mpeBusRequest;

/* cache element structure */
typedef struct cacheElement {
    struct cacheElement *next;	/* next cache element in active list */
    int transferState;	        /* transfer state */
    MScalar eAddr;		/* external address */
    MScalar iAddr;		/* internal address */
    int type : 2;
    int clean : 1;
    int lrr : 4;
} cacheElement;

/* address mask to enforce aligned accesses */
#define scalarAddressMask       0xfffffffc 
#define smallVectorAddressMask  0xfffffff8 
#define vectorAddressMask       0xfffffff0 

/* cache element constants */
#define ceMaxCount	16		/* maximum number of cache elements */
#define cePageSize	32		/* number of bytes per cache element */
#define cePageMask	0xffffffe0	/* page address mask */
#define ceMaxLRUCount   15              /* maximum lru count */

/* cache element types */
#define ceLinear	0
#define ceBilinearXY	1
#define ceBilinearUV	2

/* cache element states */
#define ceIdle              0	/* no transfer in progress */
#define cePending           1   /* transfer pending */
#define ceInProgress        2	/* transfer in progress */

/* maximum number of scoreboarded registers */
#define scoreboardCount     2

/* register scoreboard structure */
typedef struct registerScoreboard {
    int state;                  /* scoreboard state */
    cacheElement *cacheElement; /* pending cache element */
    scalarRegister *base;       /* scalar register or first register of a vector */
    MScalar offset;             /* offset for scalar loads or x/y for pixel loads */
} registerScoreboard;

/* scoreboard load types */
#define sbIdle              0
#define sbScalarLoad        1
#define sbSmallVectorLoad   2
#define sbVectorLoad        3
#define sbPixelLoad         4

/* opcode handler */
typedef void (*OpcodeHandler)(struct MPE *mpe,UINT32 *args);

/* opcode table entry */
typedef struct {
    char *name;
    Opcode opcode;
    UINT32 flags;
    OpcodeHandler handler;
} OpcodeEntry;

/* opcode table */
extern OpcodeEntry optable[];

#define maxArgs	4

/* instruction structure */
typedef struct MInst {
    UINT32 flags;
    UINT32 registerMask;
    OpcodeEntry *entry;
    UINT32 args[maxArgs];
} MInst;

/* instruction flags */
#define ifALU                   0x00000001      /* uses the alu */
#define ifALUShifter            0x00000002      /* uses the alu shifter */
#define ifMultiplier            0x00000004      /* uses the multiplier */
#define ifRegister              0x00000008      /* uses the register unit */
#define ifMemory                0x00000010      /* uses the memory unit */
#define ifExecute               0x00000020      /* uses the execute unit */
#define ifUnitMask              0x0000003f      /* function unit usage mask */
#define ifArg1IsSR              0x00000040      /* argument 1 is a scalar */ 
#define ifArg1IsHR              0x00000080      /* argument 1 is a half vector */ 
#define ifArg1IsVR              0x00000100      /* argument 1 is a vector */ 
#define ifArg2IsSR              0x00000200      /* argument 2 is a scalar */ 
#define ifArg2IsHR              0x00000400      /* argument 2 is a half vector */ 
#define ifArg2IsVR              0x00000800      /* argument 2 is a vector */ 
#define ifArg3IsSR              0x00001000      /* argument 3 is a scalar */ 
#define ifArg3IsHR              0x00002000      /* argument 3 is a half vector */ 
#define ifArg3IsVR              0x00004000      /* argument 3 is a vector */ 
#define ifArg4IsSR              0x00008000      /* argument 4 is a scalar */ 
#define ifArg4IsHR              0x00010000      /* argument 4 is a half vector */ 
#define ifArg4IsVR              0x00020000      /* argument 4 is a vector */ 

/* maximum number of instructions per packet */
#define MAXINSTPERTICK  5

/* instruction packet structure */
typedef struct InstructionPacket {
    struct InstructionPacket *next;
    MScalar addr;
    UINT32 registerMask;
    MInst *instructions[MAXINSTPERTICK];
    int instCount;
    int delaySlot;
} InstructionPacket;

/* branch types */
#define btNone          0   /* no branch pending */
#define btRelative      1   /* relative branch pending */
#define btAbsolute      2   /* absolute branch pending */

/* delay slot modes */
#define dsNone          0   /* this packet is not in the delay slot */
#define dsDelay         1   /* this packet is in the delay slot */
#define dsImpliedNOP    2   /* an implied nop is in the delay slot */

/* load target structure */
typedef struct {
    scalarRegister *target;
    MScalar value;
} LoadTarget;

/* staging buffer structure */
typedef struct stagingBuffer {
    LoadTarget targets[4];
    UINT32 mask;
    int count;
} stagingBuffer;

/* multiply overflow buffer values */
#define mvNone          0
#define mvOverflow      1
#define mvNoOverflow    2

/* mpe flags */
#define mpeStalled      0x0001  /* mpe stalled on this tick */
#define mpeLocalRAM     0x0002  /* local ram in use on this tick */

/* mpe structure */
typedef struct MPE {
    MPEPublic pub;
    struct MMP *mmp;
    int runState;
    int flags;
    int branchType;
    int branchDelaySlot;
    int branchCondition;
    MScalar branchTarget;
    InstructionPacket packet1;
    InstructionPacket packet2;
    InstructionPacket *fetchPacket;
    InstructionPacket *executePacket;
    mpeBusRequest busRequest;
    scalarRegister scalarRegisters[scalarRegisterCount];
    vectorRegister vectorRegisters[vectorRegisterCount];
    registerScoreboard scoreboard[scoreboardCount];
    stagingBuffer loadBuffer;
    stagingBuffer multiplyBuffer;
    MScalar multiplyOverflowBuffer;
    int saveReturnP;
    UINT32 statSample;
    UINT32 registerMask;
    accRegister ac;
    ioRegister pc,next,rz,sp,cc,rc0,rc1,sv_shift,ac_shift;
    bilinearRegister xy,uv;
    ioRegister rx,ry,xy_width,xy_itype,xy_base,xy_mipmap,xy_tile;
    ioRegister ru,rv,uv_width,uv_itype,uv_base,uv_mipmap,uv_tile;
    ioVectorRegister dma_1,dma_2;
    ioRegister dma_eaddr,dma_flags,dma_length,dma_iaddr;
    ioRegister dma_x,dma_y,dma_stat;
    cacheElement cacheElements[ceMaxCount];
    int cacheElementActiveCount;
    memSegment localROM;
    memSegment localRAM;
    memSegment instructionROM;
    memSegment instructionRAM;
} MPE;

/* dma buffer size */
#define dmaBufferSize   4

/* dma structure */
typedef struct dmaUnit {
    MScalar flags;
    MScalar buffer[dmaBufferSize];
    MScalar *bufferPtr;
    int bufferCnt;
    int bufferMax;
    cacheElement *cacheElement;
    MScalar eAddr,eInc;
    MScalar iAddr,iInc;
    MScalar cnt;
} dmaUnit;

/* mmp structure */
typedef struct MMP {
    MMPPublic pub;
    MScalar totalSize;
    int mpeCount;
    UINT32 *statSamples;
    busRequest *busRequesters;
    busRequest *busMaster;
    int busTimer;
    dmaUnit dma;
    memSegment externalRAM;
    memSegment sharedRAM;
    MPE mpe[1];
} MMP;

/* data space tags */
#define dtRAMSpace  0
#define dtROMSpace  1

/* dma request types */
#define DMALongRead     0
#define DMALongWrite    1
#define DMAPixelRead    5
#define DMAPixelWrite   6
#define DMAPixelWriteZ  9

/* dma constants */
#define DMAFlagFunctionMask     0xf     /* function code mask */
#define DMAFlagBackwards        0x40    /* backwards dma */
#define DMAFlagWidthMask        0x3f    /* width mask */
#define DMAFlagWidthShift       16      /* width shift */
#define DMAZTypeLT              0x1000  /* z type lt flag */
#define DMAZTypeEQ              0x2000  /* z type eq flag */
#define DMAZTypeGT              0x4000  /* z type gt flag */
#define DMALengthMask           0x3f    /* dma length mask (zero is maximum) */
#define DMAMaxLength            64      /* maximum transfer length */
#define DMARequestPriority      3

/* prototypes for mmemory.c */
int mpeInstructionAddress(MPE *mpe,MScalar addr,struct MInst **pAddr,MInst **pTop);
int mpeFetchLocalScalar(MPE *mpe,MScalar addr,MScalar *pValue);
int mpeStoreLocalScalar(MPE *mpe,MScalar addr,MScalar value);
int mpeLoadRegister(MPE *mpe,MScalar addr,int type,scalarRegister *r);
int mpeLoadRegisterFromMemory(MPE *mpe,MScalar addr,int type,scalarRegister *r);
int StageScalarLoad(stagingBuffer *buf,scalarRegister *r,MScalar value);
int StagePixelLoad(stagingBuffer *buf,scalarRegister *r,MScalar *pValues);
int StageVectorLoad(stagingBuffer *buf,scalarRegister *r,MScalar *pValues);
void PerformStagedLoads(stagingBuffer *buf);

/* prototypes for memulate.c */
void mmpRequestBus(MMP *mmp,busRequest *req);
void mpeReadNext(MPE *mpe);
void mpeWritePC(MPE *mpe);
int CallHaltHook(MPE *mpe,int code,MScalar addr);
OpcodeEntry *FindOpcodeEntry(char *name);
int CheckConditionCode(MPE *mpe,int cc);
void LatchNewValues(MPE *mpe);

/* prototypes for mdma.c */
void mpeStartDMA_1(MPE *mpe);
void mpeStartDMA_2(MPE *mpe);
void mpeRequestBus(MPE *mpe,int priority);
int mpeBusStartHandler(busRequest *r);
void mpeBusStopHandler(busRequest *r);

#endif
