/* mmath.c - merlin math routines */

#define __MERLIN_INTERNAL__
#include "merdll.h"

#define HALFBITS	16
#define HALFMASK	0xFFFF
#define SIGNBIT		0x80000000

/* mul64 - multiply two 32 bit numbers giving a 64 bit product */
EXPORT void mul64(MScalar a,MScalar b,MScalar *pHigh,MScalar *pLow)
{
    UMScalar a1,a0,b1,b0,t3,t2,t1,r3,r2,r1,r0;
    int negate = 0;
    
    /* make the arguments positive */
    if (a < 0) {
	negate = !negate;
	a = -a;
    }
    if (b < 0) {
	negate = !negate;
	b = -b;
    }
    
    /* take apart the operands */
    a1 = (UMScalar)a >> HALFBITS; a0 = a & HALFMASK;
    b1 = (UMScalar)b >> HALFBITS; b0 = b & HALFMASK;
    
    /* do the multiplies */
    r0 = b0 * a0; 
    t1 = b0 * a1;
    t2 = b1 * a0;
    t3 = b1 * a1;

    /* compute the result */
    r1 = (r0 >> HALFBITS) + (t1 & HALFMASK) + (t2 & HALFMASK);
    r2 = (r1 >> HALFBITS) + (t1 >> HALFBITS) + (t2 >> HALFBITS) + (t3 & HALFMASK);
    r3 = (r2 >> HALFBITS) + (t3 >> HALFBITS);
    
    /* assemble the result */
    r0 = (r1 << HALFBITS) | (r0 & HALFMASK);
    r1 = (r3 << HALFBITS) | (r2 & HALFMASK);

    /* negate the result if necessary */
    if (negate)
	neg64((MScalar *)&r1,(MScalar *)&r0);
    
    /* return the two halves of the result */
    *pHigh = (MScalar)r1;
    *pLow = (MScalar)r0;
}

/* neg64 - negate a 64 bit number */
EXPORT void neg64(MScalar *pv1,MScalar *pv0)
{
    if (*pv0 == 0)
	*pv1 = -*pv1;
    else {
	*pv1 = ~*pv1;
	*pv0 = -*pv0;
    }
}

/* add32 - add two 32 bit numbers giving a sum, carry and overflow */
EXPORT void add32(MScalar a,MScalar b,MScalar *pSum,int *pC,int *pV)
{
    UMScalar a1,a0,b1,b0;
    MScalar r1,r0,r,overflow;
    
    /* take apart the operands */
    a1 = (UMScalar)a >> HALFBITS; a0 = a & HALFMASK;
    b1 = (UMScalar)b >> HALFBITS; b0 = b & HALFMASK;
    
    /* compute the result */
    r0 = a0 + b0;
    r1 = a1 + b1 + (r0 >> HALFBITS);
    r = (r1 << HALFBITS) | (r0 & HALFMASK);

    /* compute the overflow flag */
    overflow = ((a & SIGNBIT) == (b & SIGNBIT))
	  &&   ((a & SIGNBIT) != (r & SIGNBIT));
    
    /* return the result, carry and overflow */
    *pSum = r;
    *pC = (r1 >> HALFBITS) != 0;
    *pV = overflow;
}

/* sub32 - subtract two 32 bit numbers giving a sum, carry and overflow */
EXPORT void sub32(MScalar a,MScalar b,MScalar *pSum,int *pC,int *pV)
{
    add32(-a,b,pSum,pC,pV);
}

/* msb32 - get the number of significant bits in an unsigned 32 bit number */
EXPORT void msb32(MScalar a,MScalar *pR)
{
    MScalar r;
    
    /* find the most significant bit that is set */
    if (a > 0) {
        for (r = 32; (a & 0x80000000) == 0; a <<= 1)
	    --r;
    }
    else if (a < 0) {
        for (r = 33; (a & 0x80000000) != 0; a <<= 1)
	    --r;
    }
    else
        r = 0;
    
    /* return the result */
    *pR = r;
}

#define BitIs0(scalar,pos)  (((scalar) & (1 << (pos))) == 0)
#define BitIs1(scalar,pos)  (((scalar) & (1 << (pos))) != 0)

/* bits32 - get a 32 bit field out of a 64 bit value */
EXPORT void bits32(MScalar n1,MScalar n0,int shift,MScalar *pR,int *pV)
{
    MScalar r;
    int pos,signbitpos;
    
    /* do the shift */
    if (shift >= 32) {
        r = n1 >> (shift - 32);
        /* overflow not possible in this case */
        *pV = 0;
    }
    else if (shift > 0) {
        r = (n1 << (32 - shift)) | ((UMScalar)n0 >> shift);
        /* overflow decided by n1 alone */
        signbitpos = shift-1;
        *pV = 0;    /* innocent by default */
        if (BitIs0(n1,signbitpos)) {
            for (pos = 31; pos > signbitpos; pos--) {
                if (BitIs1(n1,pos)) {
                    *pV = 1;    /* decision reversed: guilty of overflow */
                    break;
                }
            }
        }
        else {  /* sign bit is 1 */
            for (pos = 31; pos > signbitpos; pos--) {
                if (BitIs0(n1,pos)) {
                    *pV = 1;    /* overflow */
                    break;
                }
            }
        }
    }
    else if (shift == 0) {
        r = n0;
        /* overflow if n1 doesn't propagate the sign bit of n0 */
        if (BitIs0(31,n0)) 
            *pV = (n1 != 0);
        else
            *pV = (n1 != -1);
    }
    else {
        r = n0 << -shift;
        /* funny special case if shift == -32 */
        if (shift == -32) {
            *pV = ((n0 != 0) || (n1 != 0));
        }
        else {
            *pV = 0;    /* default assumption */
            signbitpos = 31 + shift;    /* remember, shift is negative */
            if (BitIs0(n0,signbitpos)) {
                if (n1 != 0) {
                    *pV = 1;    /* overflow */
                }
                else {
                    for (pos = 31; pos > signbitpos; pos--) {
                        if (BitIs1(n0,pos)) {
                            *pV = 1;    /* overflow */
                            break;
                        }
                    }
                }
            }
            else {  /* sign bit is 1 */
                if (n1 != -1) {
                    *pV = 1;    /* overflow */
                }
                else {
                    for (pos = 31; pos > signbitpos; pos--) {
                        if (BitIs0(n0,pos)) {
                            *pV = 1;    /* overflow */
                            break;
                        }
                    }
                }
            }
        }
    }
        
    /* return the specified 32 bit field */
    *pR = r;
}

/* shift64 - shift a 64 bit value */
EXPORT void shift64(MScalar n1,MScalar n0,int shift,MScalar *pHigh,MScalar *pLow)
{
    MScalar r1;
    MScalar r0;
    
    /* do the shift */
    if (shift > 32) {
        r0 = n1 >> (shift - 32);
        r1 = n1 < 0 ? -1 : 0;
    }
    else if (shift > 0) {
        r0 = (n1 << (32 - shift)) | ((UMScalar)n0 >> shift);
        r1 = n1 >> shift;
    }
    else if (shift > -32) {
        r0 = n0 << -shift;
        r1 = (n1 << -shift) | ((UMScalar)n0 >> (32 + shift));
    }
    else {
        r0 = 0;
        r1 = n0 << (-shift - 32);
    }
    
    /* return the two halves of the result */
    *pHigh = r1;
    *pLow = r0;
}

/* ushift64 - unsigned shift a 64 bit value */
EXPORT void ushift64(MScalar n1,MScalar n0,int shift,MScalar *pHigh,MScalar *pLow)
{
    MScalar r1,r0;
    
    /* do the shift */
    if (shift > 32) {
        r0 = (UMScalar)n1 >> (shift - 32);
        r1 = n1 < 0 ? -1 : 0;
    }
    else if (shift > 0) {
        r0 = (n1 << (32 - shift)) | ((UMScalar)n0 >> shift);
        r1 = (UMScalar)n1 >> shift;
    }
    else if (shift > -32) {
        r0 = n0 << -shift;
        r1 = (n1 << -shift) | ((UMScalar)n0 >> (32 + shift));
    }
    else {
        r0 = 0;
        r1 = n0 << (-shift - 32);
    }
    
    /* return the two halves of the result */
    *pHigh = r1;
    *pLow = r0;
}

/* shift32 - shift a 32 bit value */
EXPORT void shift32(MScalar n,int shift,MScalar *pValue)
{
    if (shift > 0)
        *pValue = n >> shift;
    else
        *pValue = n << -shift;
}

/* rotate32 - rotate a 32 bit value */
EXPORT void rotate32(MScalar n,int shift,MScalar *pValue)
{
    if (shift > 0)
        *pValue = ((UMScalar)n >> shift)
                | (n << (32 - shift));
    else
        *pValue = (n << -shift)
                | ((UMScalar)n >> (32 + shift));
}




