/* mmemory.c - merlin memory routines */
/*
	Copyright (c) 1995, by VM Labs, Inc.
	All Rights Reserved
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define __MERLIN_INTERNAL__
#include "merdll.h"

/* check to see if an address is within a segment */
#define CheckAddress(s,a)	((a) >= (s)->base && (a) < (s)->top)
#define SegmentOffset(s,a)      ((a) - (s)->base)
#define MapSegmentAddress(s,a)	((s)->data + SegmentOffset(s,a))
#define MapInstructionOffset(o) (((o) / bytesPerInstruction) * sizeof(MInst))

/* local functions */
static char *InitMPE(MMP *mmp,MPE *mpe,char *p,
                     int iRAMSize,int iROMSize,
                     int dRAMSize,int dROMSize);
static char *InitSegment(memSegment *s,char *data,MScalar base,MScalar size);
static char *InitInstSegment(memSegment *s,char *data,MScalar base,MScalar size);
static void InitCacheElements(MPE *mpe);
static int MapExternalROMAddress(MMP *mmp,MScalar addr,MScalar **pAddr);
static int MapExternalRAMAddress(MMP *mmp,MScalar addr,MScalar **pAddr);
static int MapAddress(MPE *mpe,MScalar addr,MScalar **pAddr);
static int MapRAMAddress(MPE *mpe,MScalar addr,MScalar **pAddr);
static int MapLocalAddress(MPE *mpe,MScalar addr,MScalar **pAddr);
static int MapLocalROMAddress(MPE *mpe,MScalar addr,MScalar **pAddr);
static int MapLocalRAMAddress(MPE *mpe,MScalar addr,MScalar **pAddr);
static registerScoreboard *FindRegisterScoreboard(MPE *mpe);
static cacheElement *FindCacheElement(MPE *mpe,int type,MScalar addr);
static cacheElement *FindAvailableCacheElement(MPE *mpe);

/* MakeMMP - make the memory for an mmp and its mpes */
EXPORT MMP *MakeMMP(int mpeCount,
                    int iRAMSize,int iROMSize,
                    int dRAMSize,int dROMSize,
                    int sRAMSize,int eRAMSize)
{
    long sharedSize,dataSize,instSize,localSize,headerSize,totalSize;
    MMP *mmp;
    MPE *mpe;
    char *p;

    /* compute the size of shared memory */
    sharedSize = sRAMSize + eRAMSize;

    /* compute the size of mpe local data memory */
    dataSize = dRAMSize + dROMSize;

    /* compute the size of mpe local instruction memory */
    instSize = MapInstructionOffset(iRAMSize + iROMSize);

    /* compute the size of mpe local memory */
    localSize = dataSize + instSize;

    /* compute the MMP structure size */
    headerSize = sizeof(MMP) + (mpeCount - 1) * sizeof(MPE) + mpeCount * sizeof(UINT32);

    /* compute the total size */
    totalSize = headerSize + sharedSize + mpeCount * localSize;

    /* allocate the mmp memory structure and the memory buffers */
    mmp = (MMP *)calloc(1,totalSize);
    if (mmp == NULL) return NULL;

    /* get a pointer to the first memory location */
    p = (char *)mmp + headerSize;

    /* setup the stat buffer */
    mmp->statSamples = (UINT32 *)p;
    p += mpeCount * sizeof(UINT32);
        
    /* initialize the mmp */
    p = InitSegment(&mmp->sharedRAM,p,sharedRAMBase,sRAMSize);
    p = InitSegment(&mmp->externalRAM,p,externalRAMBase,eRAMSize);
    mmp->totalSize = totalSize;
    mmp->mpeCount = mpeCount;
    mmp->busRequesters = NULL;
    mmp->pub.displayHook = NULL;
    mmp->pub.haltHook = NULL;

    /* initialize each of the mpes */
    for (mpe = &mmp->mpe[0]; --mpeCount >= 0; ++mpe)
        p = InitMPE(mmp,mpe,p,iRAMSize,iROMSize,dRAMSize,dROMSize);

    /* return the new mmp structure */
    return mmp; 
}

/* InitMPE - initialize an mpe */
static char *InitMPE(MMP *mmp,MPE *mpe,char *p,
                     int iRAMSize,int iROMSize,
                     int dRAMSize,int dROMSize)
{
    int i;
    
    /* initialize */
    mpe->mmp = mmp;
    mpe->runState = rsStop;
    mpe->registerMask = 0;
    mpe->flags = 0;

    /* initialize the instruction pipeline */
    mpe->packet1.next = &mpe->packet2;
    mpe->packet2.next = &mpe->packet1;
    mpe->fetchPacket = &mpe->packet1;
    mpe->executePacket = &mpe->packet1;
    mpe->branchType = btNone;
    mpe->saveReturnP = FALSE;

    /* initialized the staged load and multiply buffers */
    mpe->loadBuffer.count = 0;
    mpe->loadBuffer.mask = 0;
    mpe->multiplyBuffer.count = 0;
    mpe->multiplyBuffer.mask = 0;
    mpe->multiplyOverflowBuffer = mvNone;

    /* initialize local memory */
    p = InitSegment(&mpe->localROM,p,localROMBase,dROMSize);
    p = InitSegment(&mpe->localRAM,p,localRAMBase,dRAMSize);
    p = InitInstSegment(&mpe->instructionROM,p,instructionROMBase,iROMSize);
    p = InitInstSegment(&mpe->instructionRAM,p,instructionRAMBase,iRAMSize);
    InitCacheElements(mpe);

    /* initialize the register sets */
    strcpy(mpe->xy.name,"xy");
    mpe->xy.x = &mpe->rx;
    mpe->xy.y = &mpe->ry;
    mpe->xy.width = &mpe->xy_width;
    mpe->xy.itype = &mpe->xy_itype;
    mpe->xy.base = &mpe->xy_base;
    mpe->xy.mipmap = &mpe->xy_mipmap;
    mpe->xy.tile = &mpe->xy_tile;
    strcpy(mpe->uv.name,"uv");
    mpe->uv.x = &mpe->ru;
    mpe->uv.y = &mpe->rv;
    mpe->uv.width = &mpe->uv_width;
    mpe->uv.itype = &mpe->uv_itype;
    mpe->uv.base = &mpe->uv_base;
    mpe->uv.mipmap = &mpe->uv_mipmap;
    mpe->uv.tile = &mpe->uv_tile;
    mpe->dma_1.data[0] = &mpe->dma_eaddr;
    mpe->dma_1.data[1] = &mpe->dma_flags;
    mpe->dma_1.data[2] = &mpe->dma_length;
    mpe->dma_1.data[3] = &mpe->dma_iaddr;
    mpe->dma_2.data[0] = &mpe->dma_x;
    mpe->dma_2.data[1] = &mpe->dma_y;
    mpe->dma_2.data[2] = &mpe->dma_length;
    mpe->dma_2.data[3] = &mpe->dma_iaddr;

    /* setup the i/o register handlers */
    mpe->dma_eaddr.writeHandler = mpeStartDMA_1;
    mpe->dma_x.writeHandler = mpeStartDMA_2;
    mpe->dma_1.writeHandler = mpeStartDMA_1;
    mpe->dma_2.writeHandler = mpeStartDMA_2;

    /* debugging support */
    mpe->next.readHandler = mpeReadNext;
    mpe->pc.writeHandler = mpeWritePC;

    /* initialize the bus request structure */
    mpe->busRequest.mpe = mpe;
    mpe->busRequest.busRequest.startHandler = mpeBusStartHandler;
    mpe->busRequest.busRequest.stopHandler = mpeBusStopHandler;

    /* initialize the scalar registers */
    for (i = 0; i < scalarRegisterCount; ++i) {
        sprintf(mpe->scalarRegisters[i].name,"r%d",i);
        mpe->scalarRegisters[i].registerMask = RegBit(i);
    }

    /* initialize the vector registers */
    for (i = 0; i < vectorRegisterCount; ++i) {
        vectorRegister *vr = &mpe->vectorRegisters[i];
        sprintf(vr->name,"v%d",i);
        vr->base = &mpe->scalarRegisters[i * 4];
        vr->registerMask = vr->base[0].registerMask
                         | vr->base[1].registerMask
                         | vr->base[2].registerMask
                         | vr->base[3].registerMask;
    }

    /* initialize the i/o register names */
    strcpy(mpe->pc.name,"pc");
    strcpy(mpe->next.name,"next");
    strcpy(mpe->rz.name,"rz");
    strcpy(mpe->sp.name,"sp");
    strcpy(mpe->cc.name,"cc");
    strcpy(mpe->rc0.name,"rc0");
    strcpy(mpe->rc1.name,"rc1");
    strcpy(mpe->sv_shift.name,"sv_shift");
    strcpy(mpe->ac_shift.name,"ac_shift");
    strcpy(mpe->rx.name,"rx");
    strcpy(mpe->ry.name,"ry");
    strcpy(mpe->xy_width.name,"xy_width");
    strcpy(mpe->xy_itype.name,"xy_itype");
    strcpy(mpe->xy_base.name,"xy_base");
    strcpy(mpe->xy_mipmap.name,"xy_mipmap");
    strcpy(mpe->xy_tile.name,"xy_tile");
    strcpy(mpe->ru.name,"ru");
    strcpy(mpe->rv.name,"rv");
    strcpy(mpe->uv_width.name,"uv_width");
    strcpy(mpe->uv_itype.name,"uv_itype");
    strcpy(mpe->uv_base.name,"uv_base");
    strcpy(mpe->uv_mipmap.name,"uv_mipmap");
    strcpy(mpe->uv_tile.name,"uv_tile");
    strcpy(mpe->dma_eaddr.name,"dma_eaddr");
    strcpy(mpe->dma_flags.name,"dma_flags");
    strcpy(mpe->dma_length.name,"dma_length");
    strcpy(mpe->dma_iaddr.name,"dma_iaddr");
    strcpy(mpe->dma_x.name,"dma_x");
    strcpy(mpe->dma_y.name,"dma_y");
    strcpy(mpe->dma_stat.name,"dma_stat");

    /* return the next memory location */
    return p;
}

/* InitSegment - initialize a memory segment */
static char *InitSegment(memSegment *s,char *data,MScalar base,MScalar size)
{
    s->data = data;
    s->base = base;
    s->top = base + size;
    return s->dataTop = data + size;
}

/* InitInstSegment - initialize an instruction memory segment */
static char *InitInstSegment(memSegment *s,char *data,MScalar base,MScalar size)
{
    OpcodeEntry *nopEntry = FindOpcodeEntry("nop");
    MScalar count = size / bytesPerInstruction;
    MInst *i = (MInst *)data;

    /* initialize instruction memory to nops */
    while (--count >= 0) {
        i->flags = ifALU | ifStop;
        i->registerMask = 0;
        i->entry = nopEntry;
        ++i;
    }

    /* setup the segment */
    s->data = data;
    s->base = base;
    s->top = base + size;
    return s->dataTop = data + MapInstructionOffset(size);
}

/* InitCacheElements - initialize the cache elements of an mpe */
static void InitCacheElements(MPE *mpe)
{
    MScalar addr = localRAMBase;
    cacheElement *ce = mpe->cacheElements;
    registerScoreboard *sb = mpe->scoreboard;
    int count = ceMaxCount;
	
    /* initialize each cache element */
    while (--count >= 0) {
        ce->type = ceLinear;
        ce->clean = FALSE;
        ce->iAddr = addr;
        addr += cePageSize;
        ++ce;
    }
	
    /* no active cache elements to start */
    mpe->cacheElementActiveCount = 0;

    /* initialize the register scoreboard */
    count = scoreboardCount;
    while (--count >= 0) {
        sb->state = sbIdle;
        ++sb;
    }
}

/* FreeMMP - free an mmp memory structure */
EXPORT void FreeMMP(MMP *mmp)
{
    free(mmp);
}

/* mmpGetMPE - get an mpe memory structure */
EXPORT MPE *mmpGetMPE(MMP *mmp,int n)
{
    if (n < 0 || n >= mmp->mpeCount)
        return NULL;
    return &mmp->mpe[n];
}

/* mmpMapAddress - map an address in mmp memory */
int mmpMapAddress(MMP *mmp,MScalar addr,MScalar **pAddr)
{
    addr &= scalarAddressMask;
    return MapExternalROMAddress(mmp,addr,pAddr)
        || MapExternalRAMAddress(mmp,addr,pAddr);
}

/* mpeGetMMP - get the mmp associated with an mpe */
EXPORT MMP *mpeGetMMP(MPE *mpe)
{
    return mpe->mmp;
}

/* MapExternalROMAddress - map an external rom address in mpe memory */
static int MapExternalROMAddress(MMP *mmp,MScalar addr,MScalar **pAddr)
{
    return FALSE;
}

/* MapExternalRAMAddress - map an external ram address in mpe memory */
static int MapExternalRAMAddress(MMP *mmp,MScalar addr,MScalar **pAddr)
{
    if (CheckAddress(&mmp->sharedRAM,addr))
        *pAddr = (MScalar *)MapSegmentAddress(&mmp->sharedRAM,addr);
    else if (CheckAddress(&mmp->externalRAM,addr))
        *pAddr = (MScalar *)MapSegmentAddress(&mmp->externalRAM,addr);
    else
	return FALSE;
    return TRUE;
}

/* MapAddress - map an address in mpe memory */
static int MapAddress(MPE *mpe,MScalar addr,MScalar **pAddr)
{
    return MapLocalAddress(mpe,addr,pAddr)
	|| mmpMapAddress(mpe->mmp,addr,pAddr);
}

/* MapRAMAddress - map a ram address in mpe memory */
static int MapRAMAddress(MPE *mpe,MScalar addr,MScalar **pAddr)
{
    return MapLocalRAMAddress(mpe,addr,pAddr)
	|| MapExternalRAMAddress(mpe->mmp,addr,pAddr);
}

/* MapLocalAddress - map a local address in mpe memory */
static int MapLocalAddress(MPE *mpe,MScalar addr,MScalar **pAddr)
{
    return MapLocalROMAddress(mpe,addr,pAddr)
        || MapLocalRAMAddress(mpe,addr,pAddr);
}

/* MapLocalROMAddress - map a local rom address in mpe memory */
static int MapLocalROMAddress(MPE *mpe,MScalar addr,MScalar **pAddr)
{
    if (CheckAddress(&mpe->localROM,addr))
        *pAddr = (MScalar *)MapSegmentAddress(&mpe->localROM,addr);
    else if (CheckAddress(&mpe->instructionROM,addr))
        *pAddr = (MScalar *)MapSegmentAddress(&mpe->instructionROM,addr);
    else
        return FALSE;
    return TRUE;
}

/* MapLocalRAMAddress - map a local ram address in mpe memory */
static int MapLocalRAMAddress(MPE *mpe,MScalar addr,MScalar **pAddr)
{
    if (CheckAddress(&mpe->localRAM,addr))
        *pAddr = (MScalar *)MapSegmentAddress(&mpe->localRAM,addr);
    else if (CheckAddress(&mpe->instructionRAM,addr))
        *pAddr = (MScalar *)MapSegmentAddress(&mpe->instructionRAM,addr);
    else
        return FALSE;
    return TRUE;
}

/* mpeInstructionAddress - get an instruction address in mpe memory */
int mpeInstructionAddress(MPE *mpe,MScalar addr,MInst **pAddr,MInst **pTop)
{
    if (CheckAddress(&mpe->instructionROM,addr)) {
        MScalar offset = SegmentOffset(&mpe->instructionROM,addr);
        offset = MapInstructionOffset(offset);
        *pAddr = (MInst *)(mpe->instructionROM.data + offset);
        *pTop = (MInst *)mpe->instructionROM.dataTop;
    }
    else if (CheckAddress(&mpe->instructionRAM,addr)) {
        MScalar offset = SegmentOffset(&mpe->instructionRAM,addr);
        offset = MapInstructionOffset(offset);
        *pAddr = (MInst *)(mpe->instructionRAM.data + offset);
        *pTop = (MInst *)mpe->instructionRAM.dataTop;
    }
    else
        return FALSE;
    return TRUE;
}

/* mpeFetchScalar - fetch a scalar from mpe memory */
EXPORT int mpeFetchScalar(MPE *mpe,MScalar addr,MScalar *pValue)
{
    MScalar *p;

    /* mask out the low order address bits */
    addr &= scalarAddressMask;

    /* map the memory address */
    if (!MapAddress(mpe,addr,&p))
        return FALSE;

    /* fetch the value and return successfully */
    *pValue = *p;
    return TRUE;
}

/* mpeStoreScalar - store a scalar into mpe ram memory */
EXPORT int mpeStoreScalar(MPE *mpe,MScalar addr,MScalar value)
{
    MMP *mmp = mpe->mmp;
    MScalar *p;
    
    /* mask out the low order address bits */
    addr &= scalarAddressMask;

    /* map the memory address */
    if (!MapRAMAddress(mpe,addr,&p))
        return FALSE;
    
    /* store the new value */
    *p = value;

    /* update the display if necessary */
    if (mmp->pub.displayHook) {
        if (addr >= mmp->pub.displayBase
	&&  addr < mmp->pub.displayTop)
            (*mmp->pub.displayHook)(mmp,addr,value);
    }

    /* return successfully */
    return TRUE;
}

/* mpeInitScalar - initialize a scalar in mpe ram or rom memory */
EXPORT int mpeInitScalar(MPE *mpe,MScalar addr,MScalar value)
{
    MScalar *p;

    /* mask out the low order address bits */
    addr &= scalarAddressMask;

    /* map the memory address */
    if (!MapAddress(mpe,addr,&p))
        return FALSE;

    /* store the new value and return successfully */
    *p = value;
    return TRUE;
}

/* mpeLoadBinary - load binary data into mpe ram or rom memory */
EXPORT int mpeLoadBinary(MPE *mpe,MScalar addr,char *fname)
{
    MScalar value,*p;
    FILE *fp;

    /* mask out the low order address bits */
    addr &= scalarAddressMask;

    /* open the binary file */
    if ((fp = fopen(fname,"rb")) == NULL)
        return FALSE;

    /* read each scalar from the file */
    while (fread(&value,sizeof(MScalar),1,fp) == 1) {
        if (!MapAddress(mpe,addr,&p)) {
            fclose(fp);
            return FALSE;
        }
        addr += sizeof(MScalar);
        *p = value;
    }

    /* close the file and return successfully */
    fclose(fp);
    return TRUE;
}

/* mpeFetchLocalScalar - fetch a local scalar from mpe memory */
int mpeFetchLocalScalar(MPE *mpe,MScalar addr,MScalar *pValue)
{
    MScalar *p;

    /* mask out the low order address bits */
    addr &= scalarAddressMask;

    /* map the memory address */
    if (!MapLocalAddress(mpe,addr,&p))
        return FALSE;

    /* fetch the value and return successfully */
    *pValue = *p;
    return TRUE;
}

/* mpeStoreLocalScalar - store a local scalar into mpe ram memory */
int mpeStoreLocalScalar(MPE *mpe,MScalar addr,MScalar value)
{
    MMP *mmp = mpe->mmp;
    MScalar *p;
    
    /* mask out the low order address bits */
    addr &= scalarAddressMask;

    /* map the memory address */
    if (!MapLocalRAMAddress(mpe,addr,&p))
        return FALSE;
    
    /* store the new value and return successfully */
    *p = value;
    return TRUE;
}

/* StageScalarLoad - stage a scalar register load */
int StageScalarLoad(stagingBuffer *buf,scalarRegister *r,MScalar value)
{
    LoadTarget *t = buf->targets;
    if (buf->count != 0)
        return FALSE;
    t->target = r;
    t->value = value;
    buf->count = 1;
    return TRUE;
}

/* StagePixelLoad - stage a pixel register load */
int StagePixelLoad(stagingBuffer *buf,scalarRegister *r,MScalar *pValues)
{
    LoadTarget *t = buf->targets;
    if (buf->count != 0)
        return FALSE;
    while (t < &buf->targets[3]) {
        t->target = r++;
        t->value = *pValues++;
        ++t;
    }
    buf->count = 3;
    return TRUE;
}

/* StageVectorLoad - stage a vector register load */
int StageVectorLoad(stagingBuffer *buf,scalarRegister *r,MScalar *pValues)
{
    LoadTarget *t = buf->targets;
    if (buf->count != 0)
        return FALSE;
    while (t < &buf->targets[4]) {
        t->target = r++;
        t->value = *pValues++;
        ++t;
    }
    buf->count = 4;
    return TRUE;
}

/* PerformStagedLoads - perform staged loads */
void PerformStagedLoads(stagingBuffer *buf)
{
    LoadTarget *t = buf->targets;
    int count = buf->count;
    buf->count = 0;
    buf->mask = 0;
    while (--count >= 0) {
        buf->mask |= t->target->registerMask;
        t->target->newValue = t->value;
        ++t;
    }
}

/* mpeLoadRegister - load an mpe register from memory */
int mpeLoadRegister(MPE *mpe,MScalar addr,int type,scalarRegister *r)
{
    MScalar pageAddr,pageOffset,*p;
    registerScoreboard *sb;
    cacheElement *ce;

    /* first check to see if it's a local address */
    if (mpeLoadRegisterFromMemory(mpe,addr,type,r))
	return TRUE;

    /* compute the page and offset */
    pageAddr = addr & cePageMask;
    pageOffset = addr & ~cePageMask;

    /* find a cache element containing this page */
    if ((ce = FindCacheElement(mpe,ceLinear,pageAddr)) != NULL)
        return mpeLoadRegisterFromMemory(mpe,ce->iAddr + pageOffset,type,r);

    /* make sure it's an external address */
    if (!mmpMapAddress(mpe->mmp,addr,&p))
	return FALSE;
	
    /* find a cache element to read from external memory */
    if ((ce = FindAvailableCacheElement(mpe)) == NULL)
	return FALSE;

    /* setup the read */
    ce->transferState = cePending;
    ce->type = ceLinear;
    ce->clean = FALSE;
    ce->eAddr = pageAddr;

    /* scoreboard the register */
    if ((sb = FindRegisterScoreboard(mpe)) == NULL)
        return FALSE;

    /* setup the scoreboard */
    sb->state = type;
    sb->cacheElement = ce;
    sb->base = r;
    sb->offset = pageOffset;
    return TRUE;
}

/* mpeLoadRegisterFromMemory - load a register from internal memory */
int mpeLoadRegisterFromMemory(MPE *mpe,MScalar addr,int type,scalarRegister *r)
{
    MScalar *p;

    /* map the internal address */
    if (!MapLocalAddress(mpe,addr,&p))
        return FALSE;

    /* do the load */
    switch (type) {
    case sbScalarLoad:
        r->newValue = *p;
        break;
    case sbSmallVectorLoad:
        break;
    case sbVectorLoad:
        (*r++).newValue = *p++;
        (*r++).newValue = *p++;
        (*r++).newValue = *p++;
        (*r++).newValue = *p++;
        break;
    case sbPixelLoad:
        break;
    }
    
    /* return successfully */
    return TRUE;
}

/* FindRegisterScoreboard - find an available register scoreboard */
static registerScoreboard *FindRegisterScoreboard(MPE *mpe)
{
    registerScoreboard *sb = mpe->scoreboard;
    int count = scoreboardCount;
    while (--count >= 0) {
        if (sb->state == sbIdle)
            return sb;
        ++sb;
    }
    return NULL;
}

/* FindCacheElement - find a matching cache element */
static cacheElement *FindCacheElement(MPE *mpe,int type,MScalar addr)
{
    int count = mpe->cacheElementActiveCount;
    cacheElement *ce = mpe->cacheElements;
    for (; --count >= 0; ++ce)
	if (ce->type == type && ce->clean && ce->eAddr == addr)
	    return ce;
    return NULL;
}

/* FindAvailableCacheElement - find an available cache element */
static cacheElement *FindAvailableCacheElement(MPE *mpe)
{
    int count = mpe->cacheElementActiveCount;
    cacheElement *ce = mpe->cacheElements;
    cacheElement *lru = NULL;
    while (--count >= 0) {
	if (ce->transferState == ceIdle)
	    if (lru == NULL || ce->lrr > lru->lrr)
		lru = ce;
        if (ce->lrr < ceMaxLRUCount)
            ++ce->lrr;
        ++ce;
    }
    if (lru) lru->lrr = 0;
    return lru;
}

