/* mexpr.c - expression evaluator */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "mparse.h"

/* partial value structure */
typedef struct {
    Symbol *sym;
    Value val;
} PVAL;

/* prototypes */
static void do_expr(ScanContext *c,PVAL *pv);
static void getrvalue(ScanContext *c,PVAL *pv);
static void chklvalue(ScanContext *c,PVAL *pv);
static void do_expr2(ScanContext *c,PVAL *pv);
static void do_assignment(ScanContext *c,PVAL *pv,int op);
static void do_expr3(ScanContext *c,PVAL *pv);
static void do_expr4(ScanContext *c,PVAL *pv);
static void do_expr5(ScanContext *c,PVAL *pv);
static void do_expr6(ScanContext *c,PVAL *pv);
static void do_expr7(ScanContext *c,PVAL *pv);
static void do_expr8(ScanContext *c,PVAL *pv);
static void do_expr9(ScanContext *c,PVAL *pv);
static void do_expr10(ScanContext *c,PVAL *pv);
static void do_expr11(ScanContext *c,PVAL *pv);
static void do_expr12(ScanContext *c,PVAL *pv);
static void do_expr13(ScanContext *c,PVAL *pv);
static void do_binary(ScanContext *c,PVAL *pv,PVAL *pv2,int op);
static void binary_op(ScanContext *c,Value *v,int op,Value *p1,Value *p2);
static void binary_fop(ScanContext *c,Value *v,int op,Value *p1,Value *p2);
static void do_expr14(ScanContext *c,PVAL *pv);
static void do_preincrement(ScanContext *c,PVAL *pv,int op);
static void do_expr15(ScanContext *c,PVAL *pv);
static void do_postincrement(ScanContext *c,PVAL *pv,int op);
static void do_unary(ScanContext *c,PVAL *pv,int op);
static void do_primary(ScanContext *c,PVAL *pv);
static void do_number(ScanContext *c,PVAL *pv);
static void do_string(ScanContext *c,PVAL *pv);
static void do_identifier(ScanContext *c,PVAL *pv);
static void do_call(ScanContext *c,PVAL *pv);
static void CheckArgCount(ScanContext *c,int argc,int min,int max);

/* built-in functions */
static void fcn_fix(ScanContext *c,Value *v,int argc,Value *argv);
static void fcn_sin(ScanContext *c,Value *v,int argc,Value *argv);
static void fcn_cos(ScanContext *c,Value *v,int argc,Value *argv);
static void fcn_tan(ScanContext *c,Value *v,int argc,Value *argv);
static void fcn_pow(ScanContext *c,Value *v,int argc,Value *argv);

/* InitExpressionEvaluator - initialize the expression evaluator */
void InitExpressionEvaluator(ScanContext *c)
{
    AddNamedFunction(c,"fix",fcn_fix);
    AddNamedFunction(c,"sin",fcn_sin);
    AddNamedFunction(c,"cos",fcn_cos);
    AddNamedFunction(c,"tan",fcn_tan);
    AddNamedFunction(c,"pow",fcn_pow);
}

/* EvalExpression - parse and evaluate an expression */
void EvalExpression(ScanContext *c,Value *value)
{
    PVAL pv;
    do_expr(c,&pv);
    *value = pv.val;
}

/* EvalAssignmentExpression - evaluate an assignment expression */
void EvalAssignmentExpression(ScanContext *c,char *name,int op)
{
    PVAL pv;
    if ((pv.sym = FindSymbol(c,name)) == NULL)
        pv.sym = AddSymbol(c,name,nilType,NULL);
    do_assignment(c,&pv,op);
}

/* AssignmentOperatorP - is this token an assignment operator? */
int AssignmentOperatorP(int tkn)
{
    switch (tkn) {
    case '=':
    case T_ADDEQ:
    case T_SUBEQ:
    case T_MULEQ:
    case T_DIVEQ:
    case T_REMEQ:
    case T_ANDEQ:
    case T_OREQ:
    case T_XOREQ:
    case T_SHLEQ:
    case T_SHREQ:
        return TRUE;
    default:
        return FALSE;
    }
}

/* do_expr - parse an expression */
static void do_expr(ScanContext *c,PVAL *pv)
{
    do_expr2(c,pv);
    getrvalue(c,pv);
}

/* getrvalue - get the getrvalue of a partial expression */
static void getrvalue(ScanContext *c,PVAL *pv)
{
    if (pv->sym) {
	CopyValue(c,&pv->val,&pv->sym->value);
	pv->sym = NULL;
    }
}

/* chklvalue - make sure we've got an lvalue */
static void chklvalue(ScanContext *c,PVAL *pv)
{
    if (pv->sym == NULL)
	ParseError(c,"Expecting an lvalue");
}

/* do_expr2 - handle the assignment operators */
static void do_expr2(ScanContext *c,PVAL *pv)
{
    int tkn;
    do_expr3(c,pv);
    while (AssignmentOperatorP(tkn = Token(c))) 
	do_assignment(c,pv,tkn);
    SaveToken(c,tkn);
}

/* do_assignment - handle assignment operations */
static void do_assignment(ScanContext *c,PVAL *pv,int op)
{
    PVAL pv2;

    /* make sure we've got an lvalue */
    chklvalue(c,pv);

    /* make sure the value is a number */
    if (op != '=' && !NumericP(&pv->sym->value))
	ParseError(c,"Expecting a number");

    /* get the second operand */
    do_expr(c,&pv2);

    /* do the operation */
    switch (op) {
    case '=':
	CopyValue(c,&pv->val,&pv2.val);
	break;
    case T_ADDEQ:
        binary_op(c,&pv->val,'+',&pv->sym->value,&pv2.val);
	break;
    case T_SUBEQ:
        binary_op(c,&pv->val,'-',&pv->sym->value,&pv2.val);
	break;
    case T_MULEQ:
        binary_op(c,&pv->val,'*',&pv->sym->value,&pv2.val);
	break;
    case T_DIVEQ:
        binary_op(c,&pv->val,'/',&pv->sym->value,&pv2.val);
	break;
    case T_REMEQ:
        binary_fop(c,&pv->val,'%',&pv->sym->value,&pv2.val);
	break;
    case T_ANDEQ:
        binary_fop(c,&pv->val,'&',&pv->sym->value,&pv2.val);
	break;
    case T_OREQ:
        binary_fop(c,&pv->val,'!',&pv->sym->value,&pv2.val);
	break;
    case T_XOREQ:
        binary_fop(c,&pv->val,'^',&pv->sym->value,&pv2.val);
	break;
    case T_SHLEQ:
        binary_fop(c,&pv->val,T_SHL,&pv->sym->value,&pv2.val);
	break;
    case T_SHREQ:
        binary_fop(c,&pv->val,T_SHR,&pv->sym->value,&pv2.val);
	break;
    }
    
    /* set the symbol value */
    CopyValue(c,&pv->sym->value,&pv->val);

    /* not an lvalue anymore */
    pv->sym = NULL;
}

/* do_expr3 - handle the '?:' operator */
static void do_expr3(ScanContext *c,PVAL *pv)
{
    int tkn;
    do_expr4(c,pv);
    while ((tkn = Token(c)) == '?') {
	do_expr2(c,pv); getrvalue(c,pv);
	FRequire(c,':');
	do_expr2(c,pv); getrvalue(c,pv);
    }
    SaveToken(c,tkn);
}

/* do_expr4 - handle the '||' operator */
static void do_expr4(ScanContext *c,PVAL *pv)
{
    int tkn;
    do_expr5(c,pv);
    while ((tkn = Token(c)) == T_OR) {
	getrvalue(c,pv);
	do_expr5(c,pv); getrvalue(c,pv);
    }
    SaveToken(c,tkn);
}

/* do_expr5 - handle the '&&' operator */
static void do_expr5(ScanContext *c,PVAL *pv)
{
    int tkn;
    do_expr6(c,pv);
    while ((tkn = Token(c)) == T_AND) {
	getrvalue(c,pv);
	do_expr6(c,pv); getrvalue(c,pv);
    }
    SaveToken(c,tkn);
}

/* do_expr6 - handle the '|' operator */
static void do_expr6(ScanContext *c,PVAL *pv)
{
    PVAL pv2;
    int tkn;
    do_expr7(c,pv);
    while ((tkn = Token(c)) == '|') {
	do_expr7(c,&pv2);
	do_binary(c,pv,&pv2,tkn);
    }
    SaveToken(c,tkn);
}

/* do_expr7 - handle the '^' operator */
static void do_expr7(ScanContext *c,PVAL *pv)
{
    PVAL pv2;
    int tkn;
    do_expr8(c,pv);
    while ((tkn = Token(c)) == '^') {
	do_expr8(c,&pv2);
	do_binary(c,pv,&pv2,tkn);
    }
    SaveToken(c,tkn);
}

/* do_expr8 - handle the '&' operator */
static void do_expr8(ScanContext *c,PVAL *pv)
{
    PVAL pv2;
    int tkn;
    do_expr9(c,pv);
    while ((tkn = Token(c)) == '&') {
	do_expr9(c,&pv2);
	do_binary(c,pv,&pv2,tkn);
    }
    SaveToken(c,tkn);
}

/* do_expr9 - handle the '==' and '!=' operators */
static void do_expr9(ScanContext *c,PVAL *pv)
{
    PVAL pv2;
    int tkn;
    do_expr10(c,pv);
    while ((tkn = Token(c)) == T_EQ || tkn == T_NE) {
	do_expr10(c,&pv2);
	do_binary(c,pv,&pv2,tkn);
    }
    SaveToken(c,tkn);
}

/* do_expr10 - handle the '<', '<=', '>=' and '>' operators */
static void do_expr10(ScanContext *c,PVAL *pv)
{
    PVAL pv2;
    int tkn;
    do_expr11(c,pv);
    while ((tkn = Token(c)) == '<'
		    ||  tkn == T_LE
		    ||  tkn == T_GE
		    ||  tkn == '>') {
	do_expr11(c,&pv2);
	do_binary(c,pv,&pv2,tkn);
    }
    SaveToken(c,tkn);
}

/* do_expr11 - handle the '<<' and '>>' operators */
static void do_expr11(ScanContext *c,PVAL *pv)
{
    PVAL pv2;
    int tkn;
    do_expr12(c,pv);
    while ((tkn = Token(c)) == T_SHL || tkn == T_SHR) {
	do_expr12(c,&pv2);
	do_binary(c,pv,&pv2,tkn);
    }
    SaveToken(c,tkn);
}

/* do_expr12 - handle the '+' and '-' operators */
static void do_expr12(ScanContext *c,PVAL *pv)
{
    PVAL pv2;
    int tkn;
    do_expr13(c,pv);
    while ((tkn = Token(c)) == '+' || tkn == '-') {
	do_expr13(c,&pv2);
	do_binary(c,pv,&pv2,tkn);
    }
    SaveToken(c,tkn);
}

/* do_expr13 - handle the '*', '/' and '%' operators */
static void do_expr13(ScanContext *c,PVAL *pv)
{
    PVAL pv2;
    int tkn;
    do_expr14(c,pv);
    while ((tkn = Token(c)) == '*' || tkn == '/' || tkn == '%') {
	do_expr14(c,&pv2);
	do_binary(c,pv,&pv2,tkn);
    }
    SaveToken(c,tkn);
}

/* do_binary - do a binary operation */
static void do_binary(ScanContext *c,PVAL *pv,PVAL *pv2,int op)
{
    /* get the values of both operands */
    getrvalue(c,pv); getrvalue(c,pv2);

    /* do the operation */
    switch (op) {
    case '%':
    case '|':
    case '^':
    case '&':
    case T_SHL:
    case T_SHR:
        binary_fop(c,&pv->val,op,&pv->val,&pv2->val);
	break;
    case T_EQ:
    case T_NE:
    case '<':
    case T_LE:
    case T_GE:
    case '>':
    case '+':
    case '-':
    case '*':
    case '/':
        binary_op(c,&pv->val,op,&pv->val,&pv2->val);
	break;
    }
}

/* binary_op - perform a binary operation */
static void binary_op(ScanContext *c,Value *v,int op,Value *p1,Value *p2)
{
    int real = FALSE;
    
    /* convert operands if necessary */
    if (RealP(p1)) {
        if (FixedNumericP(p2))
            ConvertToReal(p2);
        else if (!RealP(p2))
            ParseError(c,"Expecting a number");
        real = TRUE;
    }
    else if (RealP(p2)) {
        if (FixedNumericP(p1))
            ConvertToReal(p1);
        else if (!RealP(p1))
            ParseError(c,"Expecting a number");
        real = TRUE;
    }
    else if (!FixedNumericP(p1) || !FixedNumericP(p2))
        ParseError(c,"Expecting a number");

    /* set the value type */
    v->type = (real ? realType : fixedType);

    /* perform the operation */
    switch (op) {
    case T_EQ:
        if (real)
            v->rvalue = (p1->rvalue == p2->rvalue);
        else {
	    v->fvalue = (p1->fvalue == p2->fvalue);
            v->ffracbits = 0;
        }
	break;
    case T_NE:
        if (real)
             v->rvalue = (p1->rvalue != p2->rvalue);
       else {
	    v->fvalue = (p1->fvalue != p2->fvalue);
            v->ffracbits = 0;
        }
	break;
    case '<':
        if (real)
            v->rvalue = (p1->rvalue < p2->rvalue);
        else {
	    v->fvalue = (p1->fvalue < p2->fvalue);
            v->ffracbits = 0;
        }
	break;
    case T_LE:
        if (real)
            v->rvalue = (p1->rvalue <= p2->rvalue);
        else {
	    v->fvalue = (p1->fvalue <= p2->fvalue);
            v->ffracbits = 0;
        }
	break;
    case T_GE:
        if (real)
            v->rvalue = (p1->rvalue >= p2->rvalue);
       else {
	    v->fvalue = (p1->fvalue >= p2->fvalue);
            v->ffracbits = 0;
        }
	break;
    case '>':
        if (real)
            v->rvalue = (p1->rvalue > p2->rvalue);
        else {
	    v->fvalue = (p1->fvalue > p2->fvalue);
            v->ffracbits = 0;
        }
	break;
    case '+':
        if (real)
            v->rvalue = p1->rvalue + p2->rvalue;
        else {
	    v->fvalue = p1->fvalue + p2->fvalue;
            v->ffracbits = p1->ffracbits;
        }
	break;
    case '-':
        if (real)
            v->rvalue = p1->rvalue - p2->rvalue;
        else {
	    v->fvalue = p1->fvalue - p2->fvalue;
            v->ffracbits = p1->ffracbits;
        }
	break;
    case '*':
        if (real)
            v->rvalue = p1->rvalue * p2->rvalue;
        else {
	    v->fvalue = p1->fvalue * p2->fvalue;
            v->ffracbits = p1->ffracbits + p2->ffracbits;
        }
	break;
    case '/':
        if (real)
            v->rvalue = p1->rvalue / p2->rvalue;
        else {
	    v->fvalue = p1->fvalue / p2->fvalue;
            v->ffracbits = p1->ffracbits + p2->ffracbits;
        }
	break;
    }
}

/* binary_fop - perform a binary fixed operation */
static void binary_fop(ScanContext *c,Value *v,int op,Value *p1,Value *p2)
{
    /* convert operands if necessary */
    if (RealP(p1))
        ConvertToFixed(p1,0);
    else if (!FixedNumericP(p1))
        ParseError(c,"Expecting a number");
    if (RealP(p2))
        ConvertToFixed(p2,0);
    else if (!FixedNumericP(p2))
        ParseError(c,"Expecting a number");

    /* set the value type */
    v->type = fixedType;

    /* do the operation */
    switch (op) {
    case '%':
        v->fvalue = p1->fvalue % p2->fvalue;
        v->ffracbits = p1->ffracbits;
        break;
    case '|':
	v->fvalue = p1->fvalue | p2->fvalue;
        v->ffracbits = p1->ffracbits;
	break;
    case '^':
	v->fvalue = p1->fvalue ^ p2->fvalue;
        v->ffracbits = p1->ffracbits;
	break;
    case '&':
	v->fvalue = p1->fvalue & p2->fvalue;
        v->ffracbits = p1->ffracbits;
	break;
    case T_SHL:
	v->fvalue = p1->fvalue << p2->fvalue;
        v->ffracbits = p1->ffracbits + p2->fvalue;
	break;
    case T_SHR:
	v->fvalue = p1->fvalue >> p2->fvalue;
        v->ffracbits = p1->ffracbits - p2->fvalue;
	break;
    }
}

/* do_expr14 - handle unary operators */
static void do_expr14(ScanContext *c,PVAL *pv)
{
    int tkn;
    switch (tkn = Token(c)) {
    case '-':
	do_unary(c,pv,'-');
	break;
    case '!':
	do_unary(c,pv,'!');
	break;
    case '~':
	do_unary(c,pv,'~');
	break;
    case T_INC:
	do_preincrement(c,pv,'+');
	break;
    case T_DEC:
	do_preincrement(c,pv,'-');
	break;
    default:
	SaveToken(c,tkn);
	do_expr15(c,pv);
	break;
    }
}

/* do_unary - do a unary operation */
static void do_unary(ScanContext *c,PVAL *pv,int op)
{
    /* get the expression */
    do_expr15(c,pv); getrvalue(c,pv);

    /* make sure the value is a number */
    if (!NumericP(&pv->val))
	ParseError(c,"Expecting a number");

    /* do the operation */
    switch (op) {
    case '-':
        if (RealP(&pv->val))
            pv->val.rvalue = -pv->val.rvalue;
        else if (FixedNumericP(&pv->val))
	    pv->val.fvalue = -pv->val.fvalue;
	break;
    case '!':
        if (RealP(&pv->val))
            pv->val.fvalue = !pv->val.rvalue;
        else
	    pv->val.fvalue = !pv->val.fvalue;
        SetFixedValue(c,&pv->val,pv->val.fvalue,0);
	break;
    case '~':
        if (RealP(&pv->val))
            ConvertToFixed(&pv->val,0);
	SetFixedValue(c,&pv->val,~pv->val.fvalue,0);
	break;
    }
}

/* do_preincrement - handle prefix '++' and '--' */
static void do_preincrement(ScanContext *c,PVAL *pv,int op)
{
    FixedValue ivalue = {1,0};
    Value increment;

    /* get the expression */
    do_expr15(c,pv);

    /* make sure we've got an lvalue */
    chklvalue(c,pv);

    /* make the increment value */
    SetValue(c,&increment,fixedType,&ivalue);

    /* do the increment operation */
    switch (op) {
    case '+':
	binary_op(c,&pv->sym->value,'+',&pv->sym->value,&increment);
	break;
    case '-':
	binary_op(c,&pv->sym->value,'-',&pv->sym->value,&increment);
	break;
    }

    /* get the current value */
    CopyValue(c,&pv->val,&pv->sym->value);
    
    /* not an lvalue anymore */
    pv->sym = NULL;
}

/* do_expr15 - handle function calls and suffix '++' and '--' */
static void do_expr15(ScanContext *c,PVAL *pv)
{
    int tkn;
    do_primary(c,pv);
    while ((tkn = Token(c)) == '(' || tkn == T_INC || tkn == T_DEC)
	switch (tkn) {
	case '(':
	    do_call(c,pv);
	    break;
	case T_INC:
	    do_postincrement(c,pv,'+');
	    break;
	case T_DEC:
	    do_postincrement(c,pv,'-');
	    break;
	}
    SaveToken(c,tkn);
}

/* do_postincrement - handle postfix '++' and '--' */
static void do_postincrement(ScanContext *c,PVAL *pv,int op)
{
    FixedValue ivalue = {1,0};
    Value increment;

    /* make sure we've got an lvalue */
    chklvalue(c,pv);

    /* get the current value */
    CopyValue(c,&pv->val,&pv->sym->value);
    
    /* make the increment value */
    SetValue(c,&increment,fixedType,&ivalue);

    /* do the increment operation */
    switch (op) {
    case '+':
	binary_op(c,&pv->sym->value,'+',&pv->sym->value,&increment);
	break;
    case '-':
	binary_op(c,&pv->sym->value,'-',&pv->sym->value,&increment);
	break;
    }

    /* not an lvalue anymore */
    pv->sym = NULL;
}

/* do_primary - parse a primary expression and unary operators */
static void do_primary(ScanContext *c,PVAL *pv)
{
    switch (Token(c)) {
    case '(':
	do_expr2(c,pv);
	FRequire(c,')');
	break;
    case T_NUMBER:
	do_number(c,pv);
	break;
    case T_STRING:
	do_string(c,pv);
	break;
    case T_IDENTIFIER:
	do_identifier(c,pv);
	break;
    default:
	ParseError(c,"Expecting a primary expression");
	break;
    }
}

/* do_number - handle a numeric constant */
static void do_number(ScanContext *c,PVAL *pv)
{
    CopyValue(c,&pv->val,&c->t_value);
    pv->sym = NULL;
}

/* do_string - handle a string constant */
static void do_string(ScanContext *c,PVAL *pv)
{
    SetValue(c,&pv->val,stringType,c->t_token);
    pv->sym = NULL;
}

/* do_identifier - handle an expression starting with an identifier */
static void do_identifier(ScanContext *c,PVAL *pv)
{
    Symbol *sym;
    if (sym = FindSymbol(c,c->t_token))
        pv->sym = sym;
    else {
        char *fullName = MakeFullName(c,c->t_token);
        if (sym = FindSymbol(c,fullName))
	    pv->sym = sym;
        else  {
	    SetValue(c,&pv->val,identifierType,fullName);
	    pv->sym = NULL;
        }
    }
}

/* MakeFullName - make a (possibly) mangled name */
char *MakeFullName(ScanContext *c,char *name)
{
    static char fullName[TKNSIZE+1];
    if (IsKeyword(name) || FindExportSymbol(c,name) || !c->file->moduleName[0])
        strcpy(fullName,name);
    else
        sprintf(fullName,"%s%%%s",c->file->moduleName,name);
    return fullName;
}

#define MAXARGS 10

/* do_call - compile a function call */
static void do_call(ScanContext *c,PVAL *pv)
{
    Value argv[MAXARGS];
    int tkn,argc = 0;
    PVAL arg;
    
    /* get the value of the function */
    getrvalue(c,pv);

    /* make sure it's a function */
    if (!FunctionP(&pv->val))
        ParseError(c,"Expecting a function");

    /* compile each argument expression */
    if ((tkn = Token(c)) != ')') {
	SaveToken(c,tkn);
	do {
	    do_expr2(c,&arg); getrvalue(c,&arg);
            CopyValue(c,&argv[argc],&arg.val);
	    ++argc;
	} while ((tkn = Token(c)) == ',');
    }
    Require(c,tkn,')');

    /* call the function */
    (*pv->val.fhandler)(c,&pv->val,argc,argv);
    pv->sym = NULL;
}

/* CheckArgCount - check the argument count to a function */
static void CheckArgCount(ScanContext *c,int argc,int min,int max)
{
    if (argc < min || argc > max)
        ParseError(c,"Wrong number of arguments");
}

/* fcn_fix - convert a floating point number to fixed point */
static void fcn_fix(ScanContext *c,Value *v,int argc,Value *argv)
{
    int fracbits;
    
    /* check the argument count */
    CheckArgCount(c,argc,1,2);
    
    /* get the number to convert */
    if (FixedNumericP(&argv[0]))
        ConvertToReal(&argv[0]);
    else if (!RealP(&argv[0]))
        ParseError(c,"Expecting a number");

    /* get the number of fracbits */
    if (argc == 2) {
        if (!IntegerP(&argv[1]))
            ParseError(c,"Expecting an integer");
        fracbits = argv[1].fvalue;
    }
    else
        fracbits = 0;

    /* compute the fixed point number */
    CopyValue(c,v,&argv[0]);
    ConvertToFixed(v,fracbits);
}

/* fcn_sin - compute the sine of a number */
static void fcn_sin(ScanContext *c,Value *v,int argc,Value *argv)
{
    double value;
    
    /* check the argument count */
    CheckArgCount(c,argc,1,1);
    
    /* get the number */
    if (FixedNumericP(&argv[0]))
        ConvertToReal(&argv[0]);
    else if (!RealP(&argv[0]))
        ParseError(c,"Expecting a number");

    /* compute the sin of the number */
    value = sin(argv[0].rvalue);
    SetValue(c,v,realType,&value);
}

/* fcn_cos - compute the cosine of a number */
static void fcn_cos(ScanContext *c,Value *v,int argc,Value *argv)
{
    double value;
    
    /* check the argument count */
    CheckArgCount(c,argc,1,1);
    
    /* get the number */
    if (FixedNumericP(&argv[0]))
        ConvertToReal(&argv[0]);
    else if (!RealP(&argv[0]))
        ParseError(c,"Expecting a number");

    /* compute the sin of the number */
    value = cos(argv[0].rvalue);
    SetValue(c,v,realType,&value);
}

/* fcn_tan - compute the tangent of a number */
static void fcn_tan(ScanContext *c,Value *v,int argc,Value *argv)
{
    double value;
    
    /* check the argument count */
    CheckArgCount(c,argc,1,1);
    
    /* get the number */
    if (FixedNumericP(&argv[0]))
        ConvertToReal(&argv[0]);
    else if (!RealP(&argv[0]))
        ParseError(c,"Expecting a number");

    /* compute the sin of the number */
    value = tan(argv[0].rvalue);
    SetValue(c,v,realType,&value);
}

/* fcn_pow - compute m to the n */
static void fcn_pow(ScanContext *c,Value *v,int argc,Value *argv)
{
    double value;
    
    /* check the argument count */
    CheckArgCount(c,argc,2,2);
    
    /* get the numbers */
    if (FixedNumericP(&argv[0]))
        ConvertToReal(&argv[0]);
    else if (!RealP(&argv[0]))
        ParseError(c,"Expecting a number");
    if (FixedNumericP(&argv[1]))
        ConvertToReal(&argv[1]);
    else if (!RealP(&argv[1]))
        ParseError(c,"Expecting a number");

    /* compute the sin of the number */
    value = pow(argv[0].rvalue,argv[1].rvalue);
    SetValue(c,v,realType,&value);
}

