/* mparse.c - merlin parser */
/*
	Copyright (c) 1996, by VM Labs, Inc.
	All Rights Reserved
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#ifdef __MWERKS__
#include <unix.h>
#else
#include <sys/stat.h>
#endif
#include "mparse.h"

/* instruction alias structure */
typedef struct Alias {
    char *alias;
    char *opcode;
} Alias;

/* instruction aliases */
Alias instructionAliases[] = {
    {   "add_reg",    "addr"    },
    {   "butta",      "butt"    },
    {   NULL,         NULL      },
};

/* nop forms */
Form nop_forms[] = {
    {{0},                           aluFlag,                    NULL},
    {{END}}
};
    
/* abs, neg and not forms */
Form abs_forms[] = {
    {{SR,0},                        aluFlag,                    NULL},
    {{END}}
};

/* add and sub forms */
Form add_forms[] = {
    {{SR,SR,0},                     aluFlag,                    NULL},
    {{UIMM,5,SR,0},		    aluFlag|hasLongFormFlag,    NULL},
    {{SR,SR,SR,0},		    aluFlag|if32Bits,           NULL},
    {{UIMM,5,SR,SR,0},		    aluFlag|if32Bits,           NULL},
    {{UIMM,10,SR,0},		    aluFlag|if32Bits,           NULL},
    {{UIMM,5,ROTATE,IMM,5,SR,0},    aluFlag|if32Bits,           NULL},
    {{SR,UIMM,5,SR,0},              aluFlag|if32Bits,           NULL},
    {{UIMM,5,SHIFT,SR,SR,0},	    aluFlag|if32Bits,           NULL},
    {{SR,SHIFT,IMM,5,SR,0},	    aluFlag|if32Bits,           NULL},
    {{SR,SHIFT,SR,SR,0},	    aluFlag|if32Bits,           NULL},
    {{END}}
};

/* add_sv and sub_sv forms */
Form add_sv_forms[] = {
    {{VR,VR,0},	                    aluFlag,                    NULL},
    {{VR,VR,VR,0},		    aluFlag|if32Bits,           NULL},
    {{END}}
};

/* cmp forms */
Form cmp_forms[] = {
    {{SR,SR,0},	                    aluFlag,                    NULL},
    {{UIMM,5,SR,0},		    aluFlag|hasLongFormFlag,    NULL},
    {{UIMM,10,SR,0},		    aluFlag|if32Bits,           NULL},
    {{UIMM,5,ROTATE,IMM,5,SR,0},    aluFlag|if32Bits,           NULL},
    {{SR,UIMM,5,0},                 aluFlag|if32Bits,           NULL},
    {{UIMM,5,SHIFT,SR,SR,0},	    aluFlag|if32Bits,           NULL},
    {{SR,SHIFT,IMM,5,SR,0},	    aluFlag|if32Bits,           NULL},
    {{SR,SHIFT,SR,SR,0},	    aluFlag|if32Bits,           NULL},
    {{END}}
};

/* as and ls forms */
Form as_forms[] = {
    {{SR,SR,SR,0},                  aluFlag|if32Bits,           NULL},
    {{SR,SR,DUP,0},                 aluFlag|if32Bits,           NULL},
    {{END}}
};

/* asl, lsl, asr, lsr forms */
Form asl_forms[] = {
    {{UIMM,5,SR,0},                 aluFlag,                    NULL},
    {{UIMM,5,SR,SR,0},              aluFlag|if32Bits,           NULL},
    {{END}}
};

/* and, or and eor forms */
Form and_forms[] = {
    {{SR,SR,0},                     aluFlag,                    NULL},
    {{SR,SR,SR,0},		    aluFlag|if32Bits,           NULL},
    {{IMM,5,SR,SR,0},		    aluFlag|if32Bits,           NULL},
    {{IMM,5,SR,DUP,0},		    aluFlag|if32Bits,           NULL},
    {{IMM,5,ROTATE,IMM,5,SR,0},     aluFlag|if32Bits,           NULL},
    {{IMM,5,SHIFT,SR,SR,0},	    aluFlag|if32Bits,           NULL},
    {{SR,SHIFT,IMM,5,SR,0},	    aluFlag|if32Bits,           NULL},
    {{SR,SHIFT,SR,SR,0},	    aluFlag|if32Bits,           NULL},
    {{SR,ROTATE,SR,SR,0},	    aluFlag|if32Bits,           NULL},
    {{END}}
};

/* bits forms */
Form bits_forms[] = {
    {{UIMM,5,SHIFT,UIMM,5,SR,0},    aluFlag|if32Bits,           NULL},
    {{END}}
};

/* msb forms */
Form msb_forms[] = {
    {{SR,SR,0},	                    aluFlag,                    NULL}, 
    {{SR,SR,SR,0},		    aluFlag|if32Bits,           NULL},
    {{IMM,5,SR,SR,0},		    aluFlag|if32Bits,           NULL},
    {{IMM,5,SR,DUP,0},  	    aluFlag|if32Bits,           NULL},
    {{END}}
};

/* bclr and bset forms */
Form bclr_forms[] = {
    {{UIMM,5,SR,0},	            aluFlag,                    NULL},
    {{END}}
};

/* btst forms */
Form btst_forms[] = {
    {{UIMM,5,SR,0},	            aluFlag,                    NULL},
    {{IMM,5,ROTATE,IMM,5,SR,0},	    aluFlag|if32Bits,           NULL},
    {{IMM,5,SHIFT,SR,SR,0},	    aluFlag|if32Bits,           NULL},
    {{SR,SHIFT,IMM,5,SR,0},	    aluFlag|if32Bits,           NULL},
    {{SR,SHIFT,SR,SR,0},	    aluFlag|if32Bits,           NULL},
    {{END}}
};

/* copy forms */
Form copy_forms[] = {
    {{SR,SR,0},	                    aluFlag,                    NULL},
    {{END}}
};

/* butt forms */
Form butt_forms[] = {
    {{HR,HR,0},	                    aluFlag,                    NULL},
    {{SR,SR,HR,0},	            aluFlag|if32Bits,           NULL},
    {{END}}
};

/* alu instruction table */
InstDef aluInstructions[] = {
{       "nop",      0,  nop_forms           },
{       "abs",      1,  abs_forms           },
{	"add",	    2,	add_forms	    },
{	"add_sv",   3,	add_sv_forms	    },
{	"sub",	    4,	add_forms	    },
{	"sub_sv",   6,	add_sv_forms	    },
{       "cmp",      7,  cmp_forms           },
{	"and",	    8,	and_forms	    },
{	"or",	    9,	and_forms	    },
{	"eor",	    10,	and_forms	    },
{       "neg",      11, abs_forms           },
{       "not",      12, abs_forms           },
{       "bits",     13, bits_forms          },
{       "as",       15, as_forms            },
{       "asl",      16, asl_forms           },
{       "asr",      17, asl_forms           },
{       "ls",       18, as_forms            },
{       "lsl",      19, asl_forms           },
{       "lsr",      20, asl_forms           },
{	"msb",	    21,	msb_forms	    },
{       "bclr",     22, bclr_forms          },
{       "bset",     23, bclr_forms          },
{	"btst",	    24,	btst_forms	    },
{       "copy",     26, copy_forms          },
{       "butt",     27, butt_forms          },
{	NULL		   		    }
};
InstructionTable aluTable = { aluFlag, aluInstructions };

/* dotp forms */
Form dotp_forms[] = {
    {{SR,VR,SHIFT,SV_SHIFT,SR,0},	    multiplierFlag|if32Bits,    NULL},
    {{SR,VR,SHIFT,MULIMM,SR,0},		    multiplierFlag|if32Bits,    NULL},
    {{VR,VR,SHIFT,SV_SHIFT,SR,0},	    multiplierFlag,             NULL},
    {{VR,VR,SHIFT,MULIMM,SR,0},		    multiplierFlag,             NULL},
    {{END}}
};

/* mul forms */
Form mul_forms[] = {
    {{SR,SR,0},                             multiplierFlag,             NULL},
    {{SR,SR,SHIFT,AC_SHIFT,SR,0},           multiplierFlag|if32Bits,    NULL},
    {{SR,SR,SHIFT,UIMM,7,SR_EQ,2,0},        multiplierFlag|if32Bits,    NULL},
    {{SR,SR,SHIFT,SR,SR_EQ,2,0},            multiplierFlag|if32Bits,    NULL},
    {{UIMM,5,SR,SHIFT,AC_SHIFT,SR,0},       multiplierFlag|if32Bits,    NULL},
    {{UIMM,5,SR,SHIFT,UIMM,7,SR_EQ,2,0},    multiplierFlag|if32Bits,    NULL},
    {{UIMM,5,SR,SHIFT,SR,SR_EQ,2,0},	    multiplierFlag|if32Bits,    NULL},
    {{END}}
};

/* mul_p forms */
Form mul_p_forms[] = {
    {{SR,VR,SHIFT,SV_SHIFT,VR,0},	    multiplierFlag|if32Bits,    NULL},
    {{SR,VR,SHIFT,MULIMM,VR,0},		    multiplierFlag|if32Bits,    NULL},
    {{RUV,VR,SHIFT,SV_SHIFT,VR,0},	    multiplierFlag|if32Bits,    NULL},
    {{RUV,VR,SHIFT,MULIMM,VR,0},	    multiplierFlag|if32Bits,    NULL},
    {{VR,VR,SHIFT,SV_SHIFT,VR,0},	    multiplierFlag,             NULL},
    {{VR,VR,SHIFT,MULIMM,VR,0},		    multiplierFlag,             NULL},
    {{END}}
};

/* mul_sv forms */
Form mul_sv_forms[] = {
    {{VR,VR,0},                             multiplierFlag,             NULL},
    {{SR,VR,SHIFT,SV_SHIFT,VR,0},	    multiplierFlag|if32Bits,    NULL},
    {{SR,VR,SHIFT,MULIMM,VR,0},		    multiplierFlag|if32Bits,    NULL},
    {{RUV,VR,SHIFT,SV_SHIFT,VR,0},	    multiplierFlag|if32Bits,    NULL},
    {{RUV,VR,SHIFT,MULIMM,VR,0},	    multiplierFlag|if32Bits,    NULL},
    {{VR,VR,SHIFT,SV_SHIFT,VR,0},	    multiplierFlag,             NULL},
    {{VR,VR,SHIFT,MULIMM,VR,0},		    multiplierFlag,             NULL},
    {{END}}
};

/* addm/subm forms */
Form addm_forms[] = {
    {{SR,SR,SR,0},                          multiplierFlag|if32Bits,    NULL},
    {{END}}
};


/* buttm forms */
Form buttm_forms[] = {
    {{HR,HR,0},	                            multiplierFlag,             NULL},
    {{SR,SR,HR,0},	                    multiplierFlag|if32Bits,    NULL},
    {{END}}
};

/* multiplier instructions */
InstDef multiplierInstructions[] = {
{       "dotp",     1,  dotp_forms          },
{	"mul",	    1,	mul_forms	    },
{       "mul_p",    2,  mul_p_forms         },
{	"mul_sv",   3,	mul_sv_forms	    },
{       "addm",     4,  addm_forms          },
{       "subm",     5,  addm_forms          },
{       "buttm",    6,  buttm_forms         },
{	NULL				    }
};
InstructionTable multiplierTable = { multiplierFlag, multiplierInstructions };

/* addr forms */
Form addr_forms[] = {
    {{RUIMM,5,RUR,DEC,RC0,DEC,RC1,0},	registerFlag,   NULL},
    {{RUIMM,5,RUR,SKIP,RC0,SKIP,RC1,0},	registerFlag,   NULL}, /* must be first */
    {{RUIMM,5,RUR,DEC,RC0,SKIP,RC1,0},	registerFlag,   NULL},
    {{RUIMM,5,RUR,SKIP,RC0,DEC,RC1,0},	registerFlag,   NULL},
    {{SR,RUR,DEC,RC0,DEC,RC1,0},	registerFlag,   NULL},
    {{SR,RUR,SKIP,RC0,SKIP,RC1,0},	registerFlag,   NULL},
    {{SR,RUR,DEC,RC0,SKIP,RC1,0},	registerFlag,   NULL},
    {{SR,RUR,SKIP,RC0,DEC,RC1,0},	registerFlag,   NULL},
    {{END}}
};

/* dec forms */
Form dec_forms[] = {
    {{RC0,0},	                        regDecFlag,     NULL},
    {{RC1,0},	                        regDecFlag,     NULL},
    {{END}}
};

/* mv_spr forms */
Form mv_spr_forms[] = {
    {{PC,RZ,DEC,RC0,DEC,RC1,0},	        registerFlag,   NULL},
    {{PC,RZ,SKIP,RC0,SKIP,RC1,0},	registerFlag,   NULL},
    {{PC,RZ,DEC,RC0,SKIP,RC1,0},	registerFlag,   NULL},
    {{PC,RZ,SKIP,RC0,DEC,RC1,0},	registerFlag,   NULL},
    {{END}}
};

/* register instructions */
InstDef registerInstructions[] = {
{	"addr",     0,	addr_forms	    },  /* must be first */
{	"dec",      1,	dec_forms	    },
{	"mv_spr",   2,  mv_spr_forms	    },
{	NULL				    }
};
InstructionTable registerTable = { registerFlag, registerInstructions };

/* ld_io forms */
Form ld_io_forms[] = {
    {{IND,IO,SR,0},	    memoryFlag,          NULL},
    {{IND,IO,VR,0},	    memoryFlag|if32Bits, NULL},
    {{END}}
};

/* ld_p forms */
Form ld_p_forms[] = {
    {{IND,BILINEAR,PR,0},   memoryFlag|if32Bits, NULL},
    {{END}}
};
    
/* ld_s forms */
Form ld_s_forms[] = {
    {{IND,SR,SR,0},	    memoryFlag,          NULL},
    {{DSOFF4,SR,0},	    memoryFlag|if32Bits, NULL},
    {{IND,BILINEAR,SR,0},   memoryFlag|if32Bits, NULL},
    {{END}}
};

/* ld_sv forms */
Form ld_sv_forms[] = {
    {{IND,SR,VR,0},	    memoryFlag|if32Bits, NULL},
    {{DSOFF8,VR,0},	    memoryFlag|if32Bits, NULL},
    {{IND,BILINEAR,VR,0},   memoryFlag|if32Bits, NULL},
    {{END}}
};

/* ld_v forms */
Form ld_v_forms[] = {
    {{IND,SR,VR,0},	    memoryFlag|if32Bits, NULL},
    {{DSOFF16,VR,0},	    memoryFlag|if32Bits, NULL},
    {{END}}
};

/* st_io forms */
Form st_io_forms[] = {
    {{SR,IND,IO,0},	    memoryFlag,          NULL},
    {{IMM,5,IND,IO,0},	    memoryFlag|if32Bits, NULL},
    {{VR,IND,IO,0},	    memoryFlag|if32Bits, NULL},
    {{END}}
};

/* st_p forms */
Form st_p_forms[] = {
    {{PR,IND,BILINEAR,0},   memoryFlag|if32Bits, NULL},
    {{END}}
};

/* st_s forms */
Form st_s_forms[] = {
    {{SR,IND,SR,0},	    memoryFlag,          NULL},
    {{SR,DSOFF4,0},	    memoryFlag|if32Bits, NULL},
    {{SR,IND,BILINEAR,0},   memoryFlag|if32Bits, NULL},
    {{END}}
};

/* st_sv forms */
Form st_sv_forms[] = {
    {{VR,IND,SR,0},	    memoryFlag|if32Bits, NULL},
    {{VR,DSOFF8,0},	    memoryFlag|if32Bits, NULL},
    {{VR,IND,BILINEAR,0},   memoryFlag|if32Bits, NULL},
    {{END}}
};

/* st_v forms */
Form st_v_forms[] = {
    {{VR,IND,SR,0},	    memoryFlag|if32Bits, NULL},
    {{VR,DSOFF16,0},	    memoryFlag|if32Bits, NULL},
    {{END}}
};

/* mv_s forms */
Form mv_s_forms[] = {
    {{SR,SR,0},		    memoryFlag,                 NULL},
    {{IMM,5,SR,0},          memoryFlag|hasLongFormFlag, NULL},
    {{IMM,32,SR,0},	    memoryFlag|if48Bits,        NULL},
    {{END}}
};

/* mv_v forms */
Form mv_v_forms[] = {
    {{VR,VR},		    memoryFlag,          NULL},
    {{END}}
};

/* push/pop forms */
Form push_pop_forms[] = {
    {{VR,0},		    memoryFlag,          NULL},
    {{VR,RZ,0},             memoryFlag,          NULL},
    {{END}}
};

/* memory instructions */
InstDef memoryInstructions[] = {
{	"ld_io",    5,  ld_io_forms         },
{	"ld_p",     3,	ld_p_forms	    },
{	"ld_s",	    1,	ld_s_forms	    },
{	"ld_sv",    2,	ld_sv_forms	    },
{	"ld_v",	    2,	ld_v_forms	    },
{	"st_io",    9,  st_io_forms	    },
{	"st_p",	    8,	st_p_forms	    },
{	"st_s",	    6,	st_s_forms	    },
{	"st_sv",    7,	st_sv_forms	    },
{	"st_v",	    7,	st_v_forms	    },
{	"mv_s",	    10,	mv_s_forms	    },
{	"mv_v",	    11,	mv_v_forms	    },
{       "push",     14, push_pop_forms      },
{       "pop",      15, push_pop_forms      },
{	NULL				    }
};
InstructionTable memoryTable = { memoryFlag, memoryInstructions };

/* bra forms */
Form bra_forms[] = {
    {{CC,RLABEL,6,0},           executeFlag, NULL},
    {{RLABEL,10,0},	        executeFlag, NULL},
    {{END}}
};

/* jmp forms */
Form jmp_forms[] = {
    {{CC,LABEL,STALLOP,0},	executeFlag, NULL},
    {{T,LABEL,STALLOP,0},	executeFlag, NULL},
    {{CC,LABEL,NO_STALLOP,0},	executeFlag, NULL},
    {{T,LABEL,NO_STALLOP,0},	executeFlag, NULL},
    {{CC,IND,SR,STALLOP,0},	executeFlag, NULL},
    {{T,IND,SR,STALLOP,0},	executeFlag, NULL},
    {{CC,IND,SR,NO_STALLOP,0},	executeFlag, NULL},
    {{T,IND,SR,NO_STALLOP,0},	executeFlag, NULL},
    {{END}}
};

/* rts forms */
Form rts_forms[] = {
    {{CC,IND,RZ,0},	        executeFlag, NULL},
    {{T,IND,RZ,0},	        executeFlag, NULL},
    {{END}}
};

/* halt forms (halt is a fake instruction!)  */
Form halt_forms[] = {
    {{0},			executeFlag, NULL},
    {{END}}
};

/* execute instructions */
InstDef executeInstructions[] = {
{	"bra",	    1,	bra_forms	    },
{	"jmp",      2,	jmp_forms	    },
{	"rts",	    3,  rts_forms	    },
{       "halt",     4,  halt_forms          },
{	NULL				    }
};
InstructionTable executeTable = { executeFlag, executeInstructions };

/* scalar registers */
char *scalarRegisters[] = {
    "r0",  "r1",  "r2",  "r3",  "r4",  "r5",  "r6",  "r7",
    "r8",  "r9",  "r10", "r11", "r12", "r13", "r14", "r15",
    "r16", "r17", "r18", "r19", "r20", "r21", "r22", "r23",
    "r24", "r25", "r26", "r27", "r28", "r29", "r30", "r31",
    0
};

/* half vector registers */
char *halfVectorRegisters[] = {
    "r0",  "r2",  "r4",  "r6",
    "r8",  "r10", "r12", "r14",
    "r16", "r18", "r20", "r22",
    "r24", "r26", "r28", "r30",
    0
};

/* vector registers */
char *vectorRegisters[] = {
    "v0", "v1", "v2", "v3", "v4", "v5", "v6", "v7",
    "p0", "p1", "p2", "p3", "p4", "p5", "p6", "p7",
    0
};

/* pixel registers (do we really need these?) */
char *pixelRegisters[] = {
    "p0", "p1", "p2", "p3", "p4", "p5", "p6", "p7",
    "v0", "v1", "v2", "v3", "v4", "v5", "v6", "v7",
    0
};

/* bilinear index registers */
char *bilinearIndexRegisters[] = {
    "xy", "uv",
    0
};

/* register unit registers */
char *registerUnitRegisters[] = {
    "rx", "ry", "ru", "rv",
    0
};

/* register ru or rv */
char *registerRUOrRV[] = {
    "ru", "rv",
    0
};

/* counter registers */
char *counterRegisters[] = {
    "rc0", "rc1",
    0
};

/* i/o registers */
char *ioRegisters[] = {
    "rx", "ry", "ru", "rv", "pc", "rz", "rc0", "rc1", "sp",
    "xy_width", "xy_itype", "xy_base", "xy_mipmap", "xy_tile",
    "uv_width", "uv_itype", "uv_base", "uv_mipmap", "uv_tile",
    "sv_shift", "ac_shift",
    "dma_1", "dma_eaddr", "dma_flags", "dma_length", "dma_iaddr",
    "dma_2", "dma_x", "dma_y", "dma_stat",
    0
};

/* condition codes */
char *conditionCodes[] = {
    "t","eq", "ne", "lt", "le", "gt", "ge",
    "cc", "cs", "vc", "vs",
    "c0eq", "c0ne", "c1ne",
    "mvc", "mvs",
    0
};
    
/* jmp stall ops */
char *stallOps[] = {
    "nop",
    0
};

/* prototypes */
int AssembleFile(char *filename,FILE *ifp,FILE *ofp,int flags,mpErrorInfo *errInfo);
Instruction *CopyInstruction(Instruction *i);
void InitializeInstruction(Instruction *i);
void FreeInstructions(ScanContext *c);
void FreeInstruction(Instruction *i);
void FreeInstructionData(Instruction *i);
int ParseLine(ScanContext *c);
int MatchAndProcessPseudoOp(ScanContext *c,char *name);
void DoModule(ScanContext *c);
void DoExport(ScanContext *c);
void DoInclude(ScanContext *c);
void DoBinaryInclude(ScanContext *c);
FILE *OpenInputFile(char *fname,char *mode,char *path);
void DoSegment(ScanContext *c);
void DoAlign(ScanContext *c,int bytes);
void DoOrigin(ScanContext *c);
void DoStart(ScanContext *c);
void DoScalarConstant(ScanContext *c);
ConstantData *AddConstantData(ScanContext *c);
void FreeConstants(ScanContext *c);
void DoStorage(ScanContext *c,int nbytes);
void ParseAssignment(ScanContext *c,char *name,int op);
int ParseInstruction(ScanContext *c,char *opname);
void ParseCompoundInstruction(ScanContext *c,CompoundInstruction *ci);
void ParseSingleInstruction(ScanContext *c,CompoundInstruction *ci,char *opname);
int BuildInstruction(ScanContext *c,CompoundInstruction *ci,InstDef *e);
void MakeDummyRegisterInstruction(ScanContext *c,Instruction *i);
Instruction *StoreInstruction(ScanContext *c,CompoundInstruction *ci,Instruction *i);
void CheckFunctionUnits(ScanContext *c,CompoundInstruction *ci,Instruction *i);
void FinishCompoundInstruction(ScanContext *c,CompoundInstruction *ci);
void ParseBeforeMethod(ScanContext *c,CompoundInstruction *ci);
void ParseAfterMethod(ScanContext *c,CompoundInstruction *ci);
void ParseOperands(ScanContext *c,Instruction *i);
void ParseOperand(ScanContext *c,Operand *op);
void GetOperand(ScanContext *c,Value *value);
void GetNumericOperand(ScanContext *c,Value *value);
int Lookup(char **table,char *name);
InstDef *LookupOpcode(InstructionTable *t,char *opname);
int MatchForm(ScanContext *c,Form *f,Instruction *i,int checkp);
int MatchIdentifier(char **table,Value *value);
int MatchRegister(char **table,Value *value);
int MatchSpecificRegister(char *name,Value *value);
int MatchImmediateValue(ScanContext *c,int bits,Operand *op,int checkp,int errorp);
int RangeCheckImmediateValue(ScanContext *c,int bits,long value,int errorp);
int MatchUnsignedImmediateValue(ScanContext *c,int bits,Operand *op,int checkp,int errorp);
int RangeCheckUnsignedImmediateValue(ScanContext *c,int bits,unsigned long value,int errorp);
int MatchMultiplierImmediateValue(ScanContext *c,Operand *op,int checkp,int errorp);
int RangeCheckMultiplierImmediateValue(ScanContext *c,long value,int errorp);
int MatchDataSpaceLabel(ScanContext *c,int shift,Operand *op,int checkp);
int RangeCheckDataSpaceLabel(ScanContext *c,int shift,Operand *op);
int MatchIndirect(ScanContext *c,int arg,Operand *op,int *pflags,int checkp);
char *CollectLispExpression(ScanContext *c);
char *CopyString(ScanContext *c,char *str);
void FreeString(char *str);
void AddSegment(ScanContext *c,char *name,UINT32 lc);
void FreeSegments(ScanContext *c);
Segment *FindSegment(ScanContext *c,char *name);
void *AllocMemory(ScanContext *c,size_t size);
void OutputEmulatorTables(void);

#ifdef WINDLL

// DLLEntryPoint - entry point for Win32
BOOL WINAPI DLLEntryPoint (HINSTANCE hinstDLL,DWORD fdwReason,LPVOID lpvReserved)
{
   switch (fdwReason) {
   case DLL_PROCESS_ATTACH:
      // The DLL is being mapped into the process's address space.
      break;

   case DLL_THREAD_ATTACH:
      // A thread is being created.
      break;

   case DLL_THREAD_DETACH:
      // A thread is exiting cleanly.
      break;

   case DLL_PROCESS_DETACH:
      // The DLL is being unmapped from the process's address 
      // space.
      break;
   }
   return(TRUE);
}

#endif

/* mpAssemble - assemble a file */
EXPORT int mpAssemble(char *ifile,char *ofile,int flags,mpErrorInfo *errInfo)
{
    FILE *ifp,*ofp;
    char path[256];
    int sts;

    /* check for table output request */
    if (flags & mpfOutputTables) {
        OutputEmulatorTables();
        return mpsSuccess;
    }

    /* initialize the error info structure */
    errInfo->text[0] = '\0';
    errInfo->file[0] = '\0';
    errInfo->line = errInfo->offset = 0;
    errInfo->status = mpsSuccess;

    /* open the input file */
    if ((ifp = OpenInputFile(ifile,"r",path)) == NULL) {
        sprintf(errInfo->text,"can't open input file: %s",ifile);
	return errInfo->status = mpsInputFile;
    }

    /* create the output file */
    if (ofile) {
        char *mode = flags & mpfOutputBinary ? "wb" : "w";
	if ((ofp = fopen(ofile,mode)) == NULL) {
	    fclose(ifp);
            sprintf(errInfo->text,"can't create output file: %s",ofile);
	    return errInfo->status = mpsOutputFile;
	}
    }
    else
	ofp = stdout;

    /* assemble the file */
    sts = AssembleFile(path,ifp,ofp,flags,errInfo);

    /* close the output file */
    fclose(ofp);

    /* return status */
    return sts;
}

/* AssembleFile - assemble a file */
int AssembleFile(char *path,FILE *ifp,FILE *ofp,int flags,mpErrorInfo *errInfo)
{
    int sts = mpsSuccess;
    ScanContext *c;

    /* initialize the scanner */
    if ((c = InitScan(errInfo)) == NULL || !IncludeFile(c,path,ifp))
	return errInfo->status = mpsError;

    /* setup the memory segments */
    AddSegment(c,"local_ram",localRAMBase);
    AddSegment(c,"local_rom",localROMBase);
    AddSegment(c,"instruction_ram",instructionRAMBase);
    AddSegment(c,"instruction_rom",instructionROMBase);
    AddSegment(c,"shared_ram",sharedRAMBase);
    AddSegment(c,"external_ram",externalRAMBase);

    /* initialize the expression evaluator */
    InitExpressionEvaluator(c);
    
    /* set the default segment */
    c->segment = FindSegment(c,"instruction_rom");

    /* setup the error trap */
    if ((sts = setjmp(c->errorTarget)) == 0) {
    
    	/* parse each line */
    	while (ParseLine(c))
	    ;

    	/* generate output */
 	if (flags & mpfOutputBinary)
            OutputBinaryData(c,ofp);
        else
            OutputInstructions(c,ofp);
    }

    /* free up allocated data */
    FreeInstructions(c);
    FreeConstants(c);
    FreeSegments(c);
    FreeScan(c);

    /* return status */
    return sts;
}

/* CopyInstruction - make an instruction */
Instruction *CopyInstruction(Instruction *src)
{
    Instruction *dst;

    /* make an instruction object */
    if ((dst = (Instruction *)malloc(sizeof(Instruction))) == NULL)
        return NULL;

    /* copy the instruction data and return the new instruction */
    *dst = *src;
    return dst;
}

/* InitializeInstruction - initialize an instruction */
void InitializeInstruction(Instruction *i)
{
    i->beforeMethod = i->afterMethod = NULL;
    i->flags = 0;
    i->nOperands = 0;
    i->instDef = NULL;
    i->form = NULL;
}

/* FreeInstructions - free all instructions */
void FreeInstructions(ScanContext *c)
{
    Instruction *i,*next;
    for (i = c->instructions; i != NULL; i = next) {
    	next = i->next;
	FreeInstruction(i);
    }
}

/* FreeInstruction - free a single instruction */
void FreeInstruction(Instruction *i)
{
    FreeInstructionData(i);
    free(i);
}

/* FreeInstructionData - free instruction data */
void FreeInstructionData(Instruction *i)
{
    int nOperands = i->nOperands;
    Operand *op = i->operands;
    for (; --nOperands >= 0; ++op) {
    	FreeValue(&op->value);
	if (op->flags & value2Flag)
	    FreeValue(&op->value2);
    }
    FreeString(i->beforeMethod);
    FreeString(i->afterMethod);
    i->beforeMethod = i->afterMethod = NULL;
    i->nOperands = 0;
}

/* ParseLine - parse a single line (or multi-line instruction) */
int ParseLine(ScanContext *c)
{
    char opname[TKNSIZE+1];
    int gotOpcode = FALSE;
    int tkn;

    /* check for end of file */
    if ((tkn = NonEOLToken(c)) == T_EOF)
	return FALSE;

    /* parse the label, assignment or opcode */
    else if (tkn == T_IDENTIFIER) {

	/* check for pseudo-ops */
	if (MatchAndProcessPseudoOp(c,c->t_token))
	    return TRUE;

	/* assume it's an opcode for the moment */
	strcpy(opname,c->t_token);

	/* check for a label */
	if ((tkn = Token(c)) == ':') {
            FixedValue label = {0,0};
            label.value = c->segment->lc;
	    AddSymbol(c,MakeFullName(c,opname),labelType,&label);
	    return TRUE;
	}

	/* check for an assignment */
	else if (AssignmentOperatorP(tkn)) {
	    ParseAssignment(c,MakeFullName(c,opname),tkn);
	    return TRUE;
	}

	/* handle an opcode */
	else {
	    SaveToken(c,tkn);
	    gotOpcode = TRUE;
	}
    }
    else
	SaveToken(c,tkn);

    /* parse an instruction */
    return ParseInstruction(c,gotOpcode ? opname : NULL);
}

/* MatchAndProcessPseudoOp - match and process a pseudo-op */
int MatchAndProcessPseudoOp(ScanContext *c,char *name)
{
    if (strcmp(name,".module") == 0)
	DoModule(c);
    else if (strcmp(name,".import") == 0)
        DoExport(c); /* same as .export for now */
    else if (strcmp(name,".export") == 0)
        DoExport(c);
    else if (strcmp(name,".include") == 0)
	DoInclude(c);
    else if (strcmp(name,".binclude") == 0)
	DoBinaryInclude(c);
    else if (strcmp(name,".segment") == 0)
        DoSegment(c);
    else if (strcmp(name,".origin") == 0)
	DoOrigin(c);
    else if (strcmp(name,".start") == 0)
	DoStart(c);
    else if (strcmp(name,".align.v") == 0)
        DoAlign(c,16);
    else if (strcmp(name,".dc.s") == 0)
	DoScalarConstant(c);
    else if (strcmp(name,".ds.s") == 0)
        DoStorage(c,4);
    else if (strcmp(name,".ds.sv") == 0)
        DoStorage(c,8);
    else if (strcmp(name,".ds.v") == 0)
        DoStorage(c,16);
    else
	return FALSE;
    return TRUE;
}

/* DoModule - handle the .module pseudo-op */
void DoModule(ScanContext *c)
{
    int tkn;
    if ((tkn = Token(c)) == T_IDENTIFIER) {
        strcpy(c->file->moduleName,c->t_token);
        tkn = Token(c);
    }
    else
        c->file->moduleName[0] = '\0';
   Require(c,tkn,T_EOL);
}

/* DoExport - handle the .export pseudo-op */
void DoExport(ScanContext *c)
{
    int tkn;
    do {
        FRequire(c,T_IDENTIFIER);
        AddExportSymbol(c,c->t_token);
    } while ((tkn = Token(c)) == ',');
    Require(c,tkn,T_EOL);
}

/* DoInclude - handle the .include pseudo-op */
void DoInclude(ScanContext *c)
{
    char path[256];
    Value value;
    FILE *fp;
    
    /* get the file name */
    GetOperand(c,&value);
    FRequire(c,T_EOL);

    /* make sure it's a file name string */
    if (value.type != stringType)
	ParseError(c,"Expecting a file name string");

    /* open the input file */
    fp = OpenInputFile(value.svalue,"r",path);

    /* make sure we found the file */
    if (fp == NULL)
	ParseError(c,"Can't open include file %s",value.svalue);

    /* setup input from the include file */
    if (!IncludeFile(c,path,fp))
	ParseError(c,"Can't include file %s",value.svalue);
}

/* DoBinaryInclude - handle the .binclude pseudo-op */
void DoBinaryInclude(ScanContext *c)
{
    char srcpath[256],binpath[256],*ext;
    struct stat srcinfo,bininfo;
    ConstantData *d;
    Value value;
    FILE *fp;
    int sts;
    
    /* get the file name */
    GetOperand(c,&value);
    FRequire(c,T_EOL);

    /* make sure it's a file name string */
    if (value.type != stringType)
	ParseError(c,"Expecting a file name string");

    /* open the input file */
    fp = OpenInputFile(value.svalue,"rb",binpath);

    /* make sure we found the file */
    if (fp == NULL)
	ParseError(c,"Can't open binary include file %s",value.svalue);

    /* get the file extension */
    ext = strrchr(binpath,'.');
    
    /* check to see if it is a source file */
    if (ext && (strcmp(ext,".a") == 0 || strcmp(ext,".A") == 0)) {

        /* make the source and binary paths */
        strcpy(srcpath,binpath);
        strcpy(ext,".bin");

        /* get the last modification times of both files */
        fstat(fileno(fp),&srcinfo); fclose(fp);
        sts = stat(binpath,&bininfo);

        /* assemble the binary source if necessary */
        if (sts != 0 || srcinfo.st_mtime > bininfo.st_mtime) {
            mpErrorInfo errInfo;
            
            /* assemble the binary file */
            sts = mpAssemble(srcpath,binpath,mpfOutputBinary,&errInfo);

            /* make sure the assembly was successful */
            if (sts != mpsSuccess)
                ParseError(c,"Error assembling binary data file %s",value.svalue);
        }

        /* open the binary file */
        fp = fopen(binpath,"rb");

        /* make sure we found the file */
        if (fp == NULL)
	    ParseError(c,"Can't open binary include file %s",binpath);
    }

    /* add a constant data entry with the binary path */
    d = AddConstantData(c);
    SetValue(c,&d->value,stringType,binpath);
    
    /* set the new origin */
    fseek(fp,0,SEEK_END);
    c->segment->lc += ftell(fp);
}

/* OpenInputFile - open an input file */
FILE *OpenInputFile(char *fname,char *mode,char *path)
{
#ifdef XLPATH
    FILE *xlLoadOpen(char *fname,char *mode,char *pathsym,char *path);
    FILE *fp = xlLoadOpen(fname,mode,"*MPARSE-PATH*",path);
#else
    char *NextIncludeDirectory(char **pp);
    char *dir,*next = NULL;
    FILE *fp = fopen(fname,mode);
    if (fp)
        strcpy(path,fname);
    else {
	/* try each directory in the path */
        while ((dir = NextIncludeDirectory(&next)) != NULL) {
	    sprintf(path,"%s%s",dir,fname);
	    if ((fp = fopen(path,mode)) != NULL)
	        break;
	}
    }
#endif
    return fp;
}

#ifndef XLPATH
/* NextIncludeDirectory - get the next include directory */
char *NextIncludeDirectory(char **pp)
{
    static char buf[256];
    char *src,*dst;
    
    /* initialize */
    if (!(src = *pp) && !(src = getenv("MPARSE")))
	return NULL;

    /* find the next directory in the path */
    for (dst = buf; *src != '\0'; *dst++ = *src++)
#ifdef macintosh
	if (*src == ',') {
#else
	if (*src == ';') {
#endif
	    ++src;
	    break;
	}
    *dst = '\0';

    /* make sure the directory ends with a backslash */
#ifndef macintosh
    if (dst > buf && dst[-1] != '\\')
	strcat(dst,"\\");
#endif
    
    /* return this directory and position in the path */
    *pp = src;
    return dst == buf ? NULL : buf;
}
#endif

/* DoSegment - handle the .segment pseudo-op */
void DoSegment(ScanContext *c)
{
    Segment *segment;

    /* get the origin expression */
    FRequire(c,T_IDENTIFIER);
    
    /* make sure the value is a segment name */
    if ((segment = FindSegment(c,c->t_token)) == NULL)
	ParseError(c,"Expecting a segment name");

    /* set the new segment */
    c->segment = segment;
    FRequire(c,T_EOL);
}

/* DoAlign - handle the .align.v pseudo-op */
void DoAlign(ScanContext *c,int bytes)
{
    long mask = bytes - 1; /* bytes must be a power of 2 */
    c->segment->lc = (c->segment->lc + mask) & ~mask;
    FRequire(c,T_EOL);
}

/* DoOrigin - handle the .origin pseudo-op */
void DoOrigin(ScanContext *c)
{
    Value value;

    /* get the origin expression */
    GetOperand(c,&value);
    
    /* make sure the value is a number */
    if (!IntegerP(&value))
	ParseError(c,"Expecting an integer");

    /* set the new origin */
    c->segment->lc = value.fvalue;
    FRequire(c,T_EOL);
}

/* DoStart - handle the .start pseudo-op */
void DoStart(ScanContext *c)
{
    GetOperand(c,&c->start);
    FRequire(c,T_EOL);
}

/* GetStartAddress - get the program start address */
long GetStartAddress(ScanContext *c)
{
    long addr = GetNumericValue(c,&c->start,"Invalid start address");
    return addr;
}

/* CheckInstructionAddress - check an instruction address */
int CheckInstructionAddress(ScanContext *c,long addr)
{
    if (addr >= instructionRAMBase
    &&  addr < instructionRAMBase + instructionRAMSize)
	return TRUE;
    else if (addr >= instructionROMBase
	 &&  addr < instructionROMBase + instructionROMSize)
	return TRUE;
    ParseError(c,"Bad instruction address %08lx",addr);
    return FALSE;
}
	
/* GetNumericValue - get a numeric value */
long GetNumericValue(ScanContext *c,Value *v,char *err)
{
    Symbol *sym;
    if (v->type == identifierType) {
	if ((sym = FindSymbol(c,v->svalue)) == NULL)
	    ParseError(c,"Undefined symbol %s",v->svalue);
	if (!LabelP(&sym->value) && !IntegerP(&sym->value))
	    ParseError(c,err);
	return sym->value.fvalue;
    }
    else if (FixedNumericP(v))
	return v->fvalue;
    ParseError(c,err);
    return 0;
}

/* DoScalarConstant - handle the .dc.s pseudo-op */
void DoScalarConstant(ScanContext *c)
{
    ConstantData *d;
    int tkn;

    /* parse a list of data values */
    do {

	/* allocate a constant data structure */
	d = AddConstantData(c);
        c->segment->lc += bytesPerScalar;

        /* get the next data value */
	GetOperand(c,&d->value);

    } while ((tkn = Token(c)) == ',');
    Require(c,tkn,T_EOL);
}

/* AddConstantData - add a constant data entry */
ConstantData *AddConstantData(ScanContext *c)
{
    ConstantData *d;

    /* allocate a constant data structure */
    if ((d = (ConstantData *)malloc(sizeof(ConstantData))) == NULL)
        ParseError(c,"Insufficient memory");

    /* save the file and line */
    strcpy(d->file,c->file->fileName);
    d->line = c->file->lineNumber;
	
    /* store the new data structure */
    d->addr = c->segment->lc;
    *c->pNextData = d;
    c->pNextData = &d->next;
    d->next = NULL;
    return d;
}

/* FreeConstants - free all constants */
void FreeConstants(ScanContext *c)
{
    ConstantData *d,*next;
    for (d = c->data; d != NULL; d = next) {
    	next = d->next;
	FreeValue(&d->value);
	free(d);
    }
}

/* DoStorage - handle the .ds.s and .ds.v pseudo-ops */
void DoStorage(ScanContext *c,int nbytes)
{
    Value value;

    /* get the storage expression */
    GetOperand(c,&value);
    
    /* make sure the value is a number */
    if (!IntegerP(&value))
	ParseError(c,"Expecting an integer");

    /* set the new origin */
    c->segment->lc += value.fvalue * nbytes;
    FRequire(c,T_EOL);
}

/* ParseAssignment - parse an assignment statement */
void ParseAssignment(ScanContext *c,char *name,int op)
{
    EvalAssignmentExpression(c,name,op);
    FRequire(c,T_EOL);
}

/* ParseInstruction - parse a single instruction */
int ParseInstruction(ScanContext *c,char *opname)
{
    CompoundInstruction ci;
    
    /* initialize compound instruction */
    ci.firstInst = NULL;
    ci.registerInst = NULL;
    ci.beforeMethod = NULL;
    ci.afterMethod = NULL;
    ci.flags = 0;
    ci.length = 0;

    /* handle single line instructions */
    if (opname)
	ParseSingleInstruction(c,&ci,opname);

    /* handle compound instructions */
    else
	ParseCompoundInstruction(c,&ci);

    /* finish the compound instruction */
    FinishCompoundInstruction(c,&ci);

    /* return successfully */
    FRequire(c,T_EOL);
    return TRUE;
}

/* ParseCompoundInstruction - parse a compound instruction */
void ParseCompoundInstruction(ScanContext *c,CompoundInstruction *ci)
{
    char opname[TKNSIZE+1];
    int tkn;

    /* get the start of the instruction */
    FRequire(c,'{');
    
    /* parse each single instruction */
    while ((tkn = NonEOLToken(c)) != T_EOF && tkn != '}') {

	/* get the opcode */
	Require(c,tkn,T_IDENTIFIER);
	strcpy(opname,c->t_token);

        /* check for a before method */
        if (strcmp(opname,"before") == 0)
	    ParseBeforeMethod(c,ci);

        /* check for after method */
        else if (strcmp(opname,"after") == 0)
	    ParseAfterMethod(c,ci);
    
	/* check for a label */
	else if ((tkn = Token(c)) == ':') {
            FixedValue label = {0,0};
            label.value = c->segment->lc;
	    AddSymbol(c,MakeFullName(c,opname),labelType,&label);
	    continue;
	}

        /* parse the single instruction */
        else {
            SaveToken(c,tkn);
	    ParseSingleInstruction(c,ci,opname);
        }

	/* check for end of instruction */
	if ((tkn = Token(c)) == '}')
	    break;
	Require(c,tkn,T_EOL);
    }
}

/* ParseSingleInstruction - parse a single instruction */
void ParseSingleInstruction(ScanContext *c,CompoundInstruction *ci,char *opname)
{
    InstDef *e;
    Alias *a;
    
    /* first translate instruction aliases */
    for (a = instructionAliases; a->alias != NULL; ++a)
        if (strcmp(a->alias,opname) == 0) {
            opname = a->opcode;
            break;
        }
    
    /* lookup the opcode */
    if ((e = LookupOpcode(&aluTable,opname)) != NULL
    ||  (e = LookupOpcode(&multiplierTable,opname)) != NULL
    ||  (e = LookupOpcode(&registerTable,opname)) != NULL
    ||  (e = LookupOpcode(&memoryTable,opname)) != NULL
    ||  (e = LookupOpcode(&executeTable,opname)) != NULL)
        BuildInstruction(c,ci,e);

    /* no matching instruction */
    else
	ParseError(c,"Unknown instruction");
}

/* BuildInstruction - parse operands and build an instruction */
int BuildInstruction(ScanContext *c,CompoundInstruction *ci,InstDef *e)
{
    Instruction tmpi,*i = &tmpi;
    int found = FALSE;
    Form *f;

    /* initialize an instruction object */
    InitializeInstruction(i);
    strcpy(i->file,c->file->fileName);
    i->line = c->file->lineNumber;
    i->instDef = e;

    /* parse the operands */
    ParseOperands(c,i);
    
    /* find the matching form */
    for (f = e->forms; f->pattern[0] != END; ++f)
        if (MatchForm(c,f,i,FALSE) && MatchForm(c,f,i,TRUE)) {
            found = TRUE;
 	    break;
	}
    
    /* no matching form found */
    if (!found) {
        ParseError(c,"Unknown instruction form");
        return FALSE;
    }

    /* store the instruction flags */
    i->flags |= f->flags & (if48Bits | if32Bits | functionUnitMask);
    i->form = f;

    /* make sure the necessary function units are available */
    CheckFunctionUnits(c,ci,i);

    /* check for a register decrement instruction */
    if (f->flags & regDecFlag) {
        int rc0p = MatchSpecificRegister("rc0",&i->operands[0].value);
        if (!ci->registerInst) {
            FreeInstructionData(i);
            MakeDummyRegisterInstruction(c,i);
            if ((i = StoreInstruction(c,ci,i)) == NULL)
                return FALSE;
            ci->flags |= registerFlag;
            ci->registerInst = i;
        }
 	if (rc0p) {
            if (ci->flags & hasRegDec0Flag)
                ParseError(c,"Duplicate decrement instruction");
            ci->flags |= hasRegDec0Flag;
            SetValue(c,&ci->registerInst->operands[2].value,registerType,"rc0");
        }
        else {
            if (ci->flags & hasRegDec1Flag)
                ParseError(c,"Duplicate decrement instruction");
            ci->flags |= hasRegDec1Flag;
            SetValue(c,&ci->registerInst->operands[3].value,registerType,"rc1");
        }
        return TRUE;
    }
    
    /* just store any other instruction */
    else {
        i = StoreInstruction(c,ci,i);
        if (f->flags & registerFlag)
            ci->registerInst = i;
        return i != NULL;
    }
}

/* MakeDummyRegisterInstruction - make a dummy register instruction */
void MakeDummyRegisterInstruction(ScanContext *c,Instruction *i)
{
    Form *f = &addr_forms[0];
    FixedValue zero = {0,0};

    /* build an addr #0,rx instruction */
    i->flags = f->flags & functionUnitMask;
    i->instDef = &registerInstructions[0];
    i->nOperands = 4;
    i->operands[0].flags = 0;
    SetValue(c,&i->operands[0].value,fixedType,&zero);
    i->operands[1].flags = 0;
    SetValue(c,&i->operands[1].value,registerType,"rx");
    i->operands[2].flags = 0;
    SetValue(c,&i->operands[2].value,nilType,NULL);
    i->operands[3].flags = 0;
    SetValue(c,&i->operands[3].value,nilType,NULL);
    i->form = f;
}

/* StoreInstruction - store a new instruction */
Instruction *StoreInstruction(ScanContext *c,CompoundInstruction *ci,Instruction *i)
{
    /* copy the instruction */
    if ((i = CopyInstruction(i)) == NULL)
	return NULL;

    /* initialize the instruction */
    i->addr = c->segment->lc;
    i->beforeMethod = CopyString(c,ci->beforeMethod);
   
    /* update the location counter */
    if (i->flags & if48Bits) {
        c->segment->lc += 3 * bytesPerInstruction;
        ci->length += 3 * bytesPerInstruction;
    }
    else if (i->flags & if32Bits) {
        c->segment->lc += 2 * bytesPerInstruction;
        ci->length += 2 * bytesPerInstruction;
    }
    else {
        c->segment->lc += bytesPerInstruction;
        ci->length += bytesPerInstruction;
    }

    /* store the new instruction */
    *c->pNextInst = i;
    c->pNextInst = &i->next;
    i->next = NULL;

    /* remember the start of this compound instruction */
    if (!ci->firstInst)
        ci->firstInst = i;
    
    /* return the instruction */
    return i;
}

/* CheckFunctionUnits - check to make sure required units are available */
void CheckFunctionUnits(ScanContext *c,CompoundInstruction *ci,Instruction *i)
{
    int requiredUnits = i->flags & functionUnitMask;
    char units[100] = "";
    int count = 0;

    /* check each unit */
    if ((requiredUnits & aluFlag) && (ci->flags & aluFlag))
        strcat(units," alu"), ++count;
    if ((requiredUnits & multiplierFlag) && (ci->flags & multiplierFlag))
        strcat(units," multiplier"), ++count;
    if ((requiredUnits & registerFlag) && (ci->flags & registerFlag))
        strcat(units," register"), ++count;
    if ((requiredUnits & memoryFlag) && (ci->flags & memoryFlag))
        strcat(units," memory"), ++count;
    if ((requiredUnits & executeFlag) && (ci->flags & executeFlag))
        strcat(units," execute"), ++count;

    /* make sure all units were available */
    if (count > 0)
        ParseError(c,"Unit%s not available:%s",(count == 1 ? "" : "s"),units);

    /* mark the required units as busy */
    ci->flags |= requiredUnits;
}

/* FinishCompoundInstruction - finish up processing of a compound instruction */
void FinishCompoundInstruction(ScanContext *c,CompoundInstruction *ci)
{
    Instruction *i,*lastInst = NULL,*aluInst = NULL;
    long lastLine = c->file->lineNumber;
    MScalar maxAddr;

    /* make sure there is at least one instruction */
    if ((i = ci->firstInst) == NULL)
        ParseError(c,"No instructions in packet");

    /* compute the address after the maximum size instruction */
    maxAddr = (i->addr & ~7) + 16; /* two 64 bit units */

    /* check to see if the instruction is too long */
    if (ci->length > 16)
        ParseError(c,"Instruction too long, addr: %08lx, length: %ld",i->addr,ci->length);
    else if (i->addr + ci->length > maxAddr)
        ParseWarning(c,"Instruction alignment problem, addr: %08lx, length: %ld",i->addr,ci->length);

    /* find the alu instruction and the last instruction */
    do {
        lastInst = i;
        if (i->flags & aluFlag)
            aluInst = i;
    } while ((i = i->next) != NULL);

    /* set the stop bit */
    if (lastInst)
        lastInst->flags |= ifStop;

    /* make sure 16 bit alu instructions are last */
    if (aluInst && aluInst != lastInst && !(aluInst->flags & if32Bits))
        ParseError(c,"16 bit ALU instructions must be last");
    
    /* fixup the instruction lengths and after methods */
    if ((i = ci->firstInst) != NULL) {
        for (;;) {
            i->lineCount = lastLine - i->line + 1;
            i->afterMethod = CopyString(c,ci->afterMethod);
            if (i->beforeMethod || i->afterMethod)
                i->flags |= ifBreak;
            if (i == lastInst)
                break;
            i = i->next;
        }
    }

    /* free the before and after methods */
    FreeString(ci->beforeMethod);
    FreeString(ci->afterMethod);
}

/* ParseBeforeMethod - parse a before method */
void ParseBeforeMethod(ScanContext *c,CompoundInstruction *ci)
{
    if (ci->beforeMethod)
	FreeString(ci->beforeMethod);
    ci->beforeMethod = CollectLispExpression(c);
}

/* ParseAfterMethod - parse an after method */
void ParseAfterMethod(ScanContext *c,CompoundInstruction *ci)
{
    if (ci->afterMethod)
	ParseError(c,"Instruction already has an after method");
    ci->afterMethod = CollectLispExpression(c);
}

/* ParseOperands - parse instruction operands */
void ParseOperands(ScanContext *c,Instruction *i)
{
    Operand *op = i->operands;
    int tkn;

    /* initialize */
    i->nOperands = 0;

    /* get the first operand */
    tkn = Token(c);
    SaveToken(c,tkn);

    /* check for an instruction with no operands */
    if (tkn == T_EOL || tkn == '}')
	return;

    /* parse each operand */
    do {
	if (++i->nOperands > MAXOPERANDS)
	    ParseError(c,"Too many operands");
	ParseOperand(c,op++);
    } while ((tkn = Token(c)) == ',');
    SaveToken(c,tkn);
}

/* ParseOperand - parse an operand */
void ParseOperand(ScanContext *c,Operand *op)
{
    int tkn;

    /* initialize */
    op->flags = 0;

    /* look for the shift, rotate or decrement indicators */
    if ((tkn = Token(c)) == T_SHIFT) {
	op->flags |= shiftFlag;
	tkn = Token(c);
    }
    else if (tkn == T_ROTATE) {
	op->flags |= rotateFlag;
	tkn = Token(c);
    }
    else if (tkn == T_DEC) {
        op->flags |= decrementFlag;
        tkn = Token(c);
    }
 
    /* look for an immediate operand */
    if (tkn == '#') {
	op->flags |= immediateFlag;
	GetNumericOperand(c,&op->value);
    }

    /* check for an indirect address */
    else if (tkn == '(') {
	op->flags |= indirectFlag;
	GetOperand(c,&op->value);
	FRequire(c,')');
    }

    /* handle simple operands */
    else {
	SaveToken(c,tkn);
	GetOperand(c,&op->value);
    }
}

/* GetOperand - get the value of an operand expression */
void GetOperand(ScanContext *c,Value *value)
{
    EvalExpression(c,value);
}

/* GetNumericOperand - get a numeric operand */
void GetNumericOperand(ScanContext *c,Value *value)
{
    GetOperand(c,value);
    if (!FixedNumericP(value))
	ParseError(c,"Expecting a numeric value");
}

/* Lookup - lookup a name in a table */
int Lookup(char **table,char *name)
{
    int i;
    for (i = 0; *table != NULL; ++table, ++i)
	if (strcmp(*table,name) == 0)
	    return i;
    return -1;
}

/* LookupOpcode - lookup and parse an instruction */
InstDef *LookupOpcode(InstructionTable *t,char *opname)
{
    InstDef *e;
    
    /* look for a matching entry */
    for (e = t->entries; e->opname != NULL; ++e)
	if (strcmp(opname,e->opname) == 0)
	    return e;
    
    /* no matching entry */
    return NULL;
}

/* MatchForm - match an instruction operands to a form */
int MatchForm(ScanContext *c,Form *f,Instruction *i,int checkp)
{
    char *pat = f->pattern;
    Operand *op = i->operands;
    int nOperands = i->nOperands;
    int code,flags,next,errorp;

    /* initialize */
    flags = op->flags;
    next = FALSE;
    
    /* match each pattern element */
    while ((code = *pat++) != 0) {

	/* dispatch on the pattern code */
        switch (code) {
	case SR:	/* scalar register */
	    if (--nOperands < 0) return FALSE;
	    if (!MatchRegister(scalarRegisters,&op->value))
	        return FALSE;
	    next = TRUE;
	    break;
        case SR_EQ:
	    if (--nOperands < 0) return FALSE;
	    if (!MatchRegister(scalarRegisters,&op->value))
	        return FALSE;
            if (strcmp(i->operands[*pat++ - 1].value.svalue,op->value.svalue) != 0)
                ParseError(c,"This operand must match operand %d",pat[-1]);
	    next = TRUE;
            break;
	case HR:	/* half vector register */
	    if (--nOperands < 0) return FALSE;
	    if (!MatchRegister(halfVectorRegisters,&op->value))
	        return FALSE;
	    next = TRUE;
	    break;
	case VR:	/* vector register */
	    if (--nOperands < 0) return FALSE;
	    if (!MatchRegister(vectorRegisters,&op->value))
	        return FALSE;
	    next = TRUE;
	    break;
	case PR:	/* pixel register */
	    if (--nOperands < 0) return FALSE;
	    if (!MatchRegister(pixelRegisters,&op->value))
	        return FALSE;
	    next = TRUE;
	    break;
	case RZ:	/* rz register */
	    if (--nOperands < 0) return FALSE;
	    if (!MatchSpecificRegister("rz",&op->value)) return FALSE;
	    next = TRUE;
	    break;
	case PC:	/* pc register */
	    if (--nOperands < 0) return FALSE;
	    if (!MatchSpecificRegister("pc",&op->value)) return FALSE;
	    next = TRUE;
	    break;
	case RC0:	/* rc0 register */
	    if (--nOperands < 0) return FALSE;
	    if (!MatchSpecificRegister("rc0",&op->value)) return FALSE;
	    next = TRUE;
	    break;
	case RC1:	/* rc1 register */
	    if (--nOperands < 0) return FALSE;
	    if (!MatchSpecificRegister("rc1",&op->value)) return FALSE;
	    next = TRUE;
	    break;
	case RUR:	/* register unit register */
	    if (--nOperands < 0) return FALSE;
	    if (!MatchRegister(registerUnitRegisters,&op->value))
	        return FALSE;
	    next = TRUE;
	    break;
	case RUV:	/* register ru or rv */
	    if (--nOperands < 0) return FALSE;
	    if (!MatchRegister(registerRUOrRV,&op->value))
	        return FALSE;
	    next = TRUE;
	    break;
	case AC_SHIFT:	/* ac_shift register */
	    if (--nOperands < 0) return FALSE;
	    if (!MatchSpecificRegister("ac_shift",&op->value)) return FALSE;
	    next = TRUE;
	    break;
	case SV_SHIFT:	/* sv_shift register */
	    if (--nOperands < 0) return FALSE;
	    if (!MatchSpecificRegister("sv_shift",&op->value)) return FALSE;
	    next = TRUE;
	    break;
	case IMM:	/* immediate value */
	case RUIMM:	/* register unit immediate value */
	    if (--nOperands < 0) return FALSE;
            errorp = (f->flags & hasLongFormFlag) == 0;
	    if (!MatchImmediateValue(c,*pat++,op,checkp,errorp))
	        return FALSE;
	    flags &= ~immediateFlag;
	    next = TRUE;
	    break;
	case UIMM:	/* unsigned immediate value */
	    if (--nOperands < 0) return FALSE;
            errorp = (f->flags & hasLongFormFlag) == 0;
	    if (!MatchUnsignedImmediateValue(c,*pat++,op,checkp,errorp))
	        return FALSE;
	    flags &= ~immediateFlag;
	    next = TRUE;
	    break;
	case MULIMM:	/* multiplier immediate value */
	    if (--nOperands < 0) return FALSE;
	    if (!MatchMultiplierImmediateValue(c,op,checkp,errorp))
	        return FALSE;
	    flags &= ~immediateFlag;
	    next = TRUE;
	    break;
	case SHIFT:	/* shift operator */
	    if (nOperands == 0) return FALSE;
	    if ((op->flags & shiftFlag) == 0)
	        return FALSE;
	    flags &= ~shiftFlag;
	    break;
	case ROTATE:	/* rotate operator */
	    if (nOperands == 0) return FALSE;
	    if ((op->flags & rotateFlag) == 0)
	        return FALSE;
	    flags &= ~rotateFlag;
	    break;
        case DEC:       /* decrement operator */
            if (nOperands == 0) return FALSE;
            if ((op->flags & decrementFlag) == 0)
                return FALSE;
            flags &= ~decrementFlag;
            break;
	case IND:	/* indirect operator */
	    if (--nOperands < 0) return FALSE;
	    if (!(flags & indirectFlag)) return FALSE;
	    if (!MatchIndirect(c,*pat++,op,&flags,checkp))
		return FALSE;
	    flags &= ~indirectFlag;
	    next = TRUE;
	    break;
	case CC:	/* condition code */
	    if (--nOperands < 0) return FALSE;
	    if (!MatchIdentifier(conditionCodes,&op->value))
	        return FALSE;
	    next = TRUE;
	    break;
	case T:	        /* implicit alw condition code */
        case DUP:       /* duplicate preceeding operand */
        case NO_STALLOP:/* no stallop specified */
	    break;
	case STALLOP:	/* bra stall op */
	    if (--nOperands < 0) return FALSE;
	    if (!MatchIdentifier(stallOps,&op->value))
	        return FALSE;
	    next = TRUE;
	    break;
	case RLABEL:	/* branch target (relative branches) */
	    ++pat;
	case LABEL:	/* branch target */
	    if (--nOperands < 0) return FALSE;
	    if (!IdentifierP(&op->value)
	    &&  !LabelP(&op->value)
	    &&  !IntegerP(&op->value))
		return FALSE;
	    next = TRUE;
	    break;
	case DSOFF4:
	    if (--nOperands < 0) return FALSE;
	    if (!MatchDataSpaceLabel(c,2,op,checkp))
	        return FALSE;
	    next = TRUE;
	    break;
	case DSOFF8:
	    if (--nOperands < 0) return FALSE;
	    if (!MatchDataSpaceLabel(c,3,op,checkp))
	        return FALSE;
	    next = TRUE;
	    break;
	case DSOFF16:
	    if (--nOperands < 0) return FALSE;
	    if (!MatchDataSpaceLabel(c,4,op,checkp))
	        return FALSE;
	    next = TRUE;
	    break;
        case SKIP:      /* move the remaining operands to the right */
            if (checkp) {
                if (nOperands > 0) {
                    Operand *src = &op[nOperands];
                    Operand *dst = src + 1;
                    while (src != op)
                        *--dst = *--src;
                }
                ++i->nOperands;
                SetValue(c,&op->value,nilType,NULL);
                flags = op->flags = 0;
                next = TRUE;
            }
            ++pat;
            break;
	default:	/* unknown pattern element */
	    ParseError(c,"Unknown pattern element %d",code);
	    return FALSE;
	}

	/* move to next operand */
	if (next) {
	    if (flags & (shiftFlag
                       | rotateFlag
                       | decrementFlag
                       | immediateFlag
                       | indirectFlag))
		return FALSE;
	    ++op; flags = op->flags;
	    next = FALSE;
	}
    }
	
    /* form matches if there are no more operands */
    return nOperands == 0;
}

/* IsKeyword - check to see if an identifier is a keyword */
int IsKeyword(char *id)
{
    return Lookup(conditionCodes,id) != -1
        || Lookup(stallOps,id) != -1;
}

/* MatchIdentifier - match an identifier */
int MatchIdentifier(char **table,Value *value)
{
    return value->type == identifierType
	&& Lookup(table,value->svalue) != -1;
}

/* MatchRegister - match a register */
int MatchRegister(char **table,Value *value)
{
    return value->type == registerType
	&& Lookup(table,value->svalue) != -1;
}

/* MatchSpecificRegister - match a specific register */
int MatchSpecificRegister(char *name,Value *value)
{
    return value->type == registerType
	&& strcmp(value->svalue,name) == 0;
}

/* MatchImmediateValue - match immediate value */
int MatchImmediateValue(ScanContext *c,int bits,Operand *op,int checkp,int errorp)
{
    if (!(op->flags & immediateFlag) || !FixedNumericP(&op->value))
	return FALSE;
    else if (checkp)
        return RangeCheckImmediateValue(c,bits,op->value.fvalue,errorp);
    return TRUE;
}

/* RangeCheckImmediateValue - range check immediate value */
int RangeCheckImmediateValue(ScanContext *c,int bits,long value,int errorp)
{
    long range2 = 1 << (bits - 1);
    long lo = -range2;
    long hi = range2 - 1;
    if (bits == 32 || value >= lo && value <= hi)
        return TRUE;
    if (errorp)
        ParseError(c,"Immediate value out of range %ld (%ld..%ld)",value,lo,hi);
    return FALSE;
}

/* MatchMultiplierImmediateValue - match multiplier immediate value */
int MatchMultiplierImmediateValue(ScanContext *c,Operand *op,int checkp,int errorp)
{
    if (!(op->flags & immediateFlag) || !FixedNumericP(&op->value))
	return FALSE;
    else if (checkp)
        return RangeCheckMultiplierImmediateValue(c,op->value.fvalue,errorp);
    return TRUE;
}

/* RangeCheckMultiplierImmediateValue - range check multiplier immediate value */
int RangeCheckMultiplierImmediateValue(ScanContext *c,long value,int errorp)
{
    if (value == 0 || value == 8 || value == 12 || value == 14)
        return TRUE;
    if (errorp)
        ParseError(c,"Illegal value %ld (0, 8, 12, 14)",value);
    return FALSE;
}

/* MatchUnsignedImmediateValue - match immediate value */
int MatchUnsignedImmediateValue(ScanContext *c,int bits,Operand *op,int checkp,int errorp)
{
    if (!(op->flags & immediateFlag) || !FixedNumericP(&op->value))
	return FALSE;
    else if (checkp)
        return RangeCheckUnsignedImmediateValue(c,bits,op->value.fvalue,errorp);
    return TRUE;
}

/* RangeCheckUnsignedImmediateValue - range check immediate value */
int RangeCheckUnsignedImmediateValue(ScanContext *c,int bits,unsigned long value,int errorp)
{
    unsigned long hi = (1 << bits) - 1;
    if (bits == 32 || value <= hi)
        return TRUE;
    if (errorp)
        ParseError(c,"Immediate value out of range %ld (0..%ld)",value,hi);
    return FALSE;
}

/* MatchDataSpaceLabel - match a data space label */
int MatchDataSpaceLabel(ScanContext *c,int shift,Operand *op,int checkp)
{
    if (!IntegerP(&op->value) && !LabelP(&op->value))
	return FALSE;
    if (checkp)
        return RangeCheckDataSpaceLabel(c,shift,op);
    return TRUE;
}

/* RangeCheckDataSpaceLabel - range check data space label */
int RangeCheckDataSpaceLabel(ScanContext *c,int shift,Operand *op)
{
    long hi = (1 << 11) - 1;
    long mask = (1 << shift) - 1;
    FixedValue value = {0,0};
    long offset;
    
    /* check for a local rom access */
    offset = op->value.fvalue - localROMBase;
    if (offset >= 0 && offset <= hi) {
	if (offset & mask)
	    ParseError(c,"Offset not aligned %ld (%ld)",offset,mask + 1);
        value.value = offset;
	op->flags |= value2Flag;
	SetValue(c,&op->value,identifierType,"dtrom");
	SetValue(c,&op->value2,fixedType,&value);
	return TRUE;
    }

    /* check for a local ram access */
    offset = op->value.fvalue - localRAMBase;
    if (offset >= 0 && offset <= hi) {
	if (offset & mask)
	    ParseError(c,"Offset not aligned %ld (%ld)",offset,mask + 1);
        value.value = offset;
	op->flags |= value2Flag;
	SetValue(c,&op->value,identifierType,"dtram");
	SetValue(c,&op->value2,fixedType,&value);
	return TRUE;
    }

    /* label out of range */
    ParseError(c,"Offset out of range %ld (0..%ld)",op->value.fvalue,hi);
    return FALSE;
}

/* MatchIndirect - match an indirect operand */
int MatchIndirect(ScanContext *c,int arg,Operand *op,int *pflags,int checkp)
{
    switch (arg) {
    case SR:
	return MatchRegister(scalarRegisters,&op->value);
    case RZ:	/* rz register */
	return MatchSpecificRegister("rz",&op->value);
    case BILINEAR:
	return MatchRegister(bilinearIndexRegisters,&op->value);
    case IO:
	if (op->value.type != registerType) return FALSE;
	if (checkp && Lookup(ioRegisters,op->value.svalue) == -1)
	    ParseError(c,"Not an i/o register %s",op->value.svalue);
	return TRUE;
    default:
	ParseError(c,"Unknown indirect form");
	return FALSE;
    }
}

/* CollectLispExpression - collect a complete Lisp expression */
char *CollectLispExpression(ScanContext *c)
{
    char expr[LSIZE],*p;
    long length = 1; /* space for terminating nul */
    long level = 1;
    int ch;

    /* skip leading spaces */
    if ((ch = SkipSpaces(c)) == '(') {
	p = expr; *p++ = ch;
	while ((ch = GetCh(c)) != EOF) {
	    if (++length >= LSIZE)
		ParseError(c,"Expression too long");
	    if ((*p++ = ch) == ')' && --level <= 0)
		break;
	    else if (ch == '(')
		++level;
	}
	*p = '\0';
    }

    /* handle single tokens */
    else {
	UngetCh(c,ch);
	Token(c);
	strcpy(expr,c->t_token);
    }

    /* return a copy of the string */
    return CopyString(c,expr);
}

/* CopyString - make a copy of a string */
char *CopyString(ScanContext *c,char *str)
{
    char *p;

    /* check for a null string */
    if (str == NULL)
        return NULL;
    
    /* allocate a buffer for the copy */
    p = AllocMemory(c,strlen(str) + 1);

    /* return the copy */
    strcpy(p,str);
    return p;
}

/* FreeString - free a string */
void FreeString(char *str)
{
    if (str)
        free(str);
}

/* AddSegment - add a memory segment */
void AddSegment(ScanContext *c,char *name,UINT32 lc)
{
    Segment *s = FindSegment(c,name);
    if (!s) {
        s = (Segment *)AllocMemory(c,sizeof(Segment));
        s->name = (char *)AllocMemory(c,strlen(name) + 1);
        s->next = c->segments;
        strcpy(s->name,name);
    }   c->segments = s;
    s->lc = lc;
}

/* FreeSegments - free memory segments */
void FreeSegments(ScanContext *c)
{
    Segment *s,*next;
    for (s = c->segments; s != NULL; s = next) {
        next = s->next;
        free(s->name);
        free(s);
    }
}

/* FindSegment - find a memory segment */
Segment *FindSegment(ScanContext *c,char *name)
{
    Segment *s = c->segments;
    for (; s != NULL; s = s->next)
        if (strcmp(name,s->name) == 0)
            return s;
    return NULL;
}

/* AllocMemory - allocate memory */
void *AllocMemory(ScanContext *c,size_t size)
{
    void *p = malloc(size);
    if (p == NULL)
        ParseError(c,"Insufficient memory");
    return p;
}

/* Abort - abort the current assembly */
void Abort(ScanContext *c,int sts)
{
    longjmp(c->errorTarget,sts);
}

void OutputOpcodes(FILE *fp,InstructionTable *table);
void OutputOpcodeHandlers(FILE *fp,InstructionTable *table);
void OutputOpcodeTable(FILE *fp,InstructionTable *table);
char *GetInstructionFlag(int flags);

/* OutputEmulatorTables - output tables used by the emulator */
void OutputEmulatorTables(void)
{
    FILE *fp;

    /* output the opcode definitions */
    if (fp = fopen("opcodes.hx","w")) {
        fprintf(fp,"/* opcodes.h - merlin instruction definitions */\n\n");
        fprintf(fp,"#ifndef __OPCODES_H__\n");
        fprintf(fp,"#define __OPCODES_H__\n\n");
        fprintf(fp,"/* opcodes */\n");
        fprintf(fp,"typedef enum {\n\n");
        OutputOpcodes(fp,&aluTable);
        OutputOpcodes(fp,&multiplierTable);
        OutputOpcodes(fp,&registerTable);
        OutputOpcodes(fp,&memoryTable);
        OutputOpcodes(fp,&executeTable);
        fprintf(fp,"    max_opcode\n\n");
        fprintf(fp,"} Opcode;\n\n");
        fprintf(fp,"#endif\n");
        fclose(fp);
    }

    /* output the opcode handlers and table */
    if (fp = fopen("opcodes.cx","w")) {
        OutputOpcodeHandlers(fp,&aluTable);
        OutputOpcodeHandlers(fp,&multiplierTable);
        OutputOpcodeHandlers(fp,&registerTable);
        OutputOpcodeHandlers(fp,&memoryTable);
        OutputOpcodeHandlers(fp,&executeTable);
        fprintf(fp,"/* opcode to handler table */\n");
        fprintf(fp,"OpcodeEntry optable[] = {\n");
        OutputOpcodeTable(fp,&aluTable);
        OutputOpcodeTable(fp,&multiplierTable);
        OutputOpcodeTable(fp,&registerTable);
        OutputOpcodeTable(fp,&memoryTable);
        OutputOpcodeTable(fp,&executeTable);
        fprintf(fp,"{NULL,0,0,NULL}\n");
        fprintf(fp,"};\n");
        fclose(fp);
    }
}

/* OutputOpcodes - output a single emulator table */
void OutputOpcodes(FILE *fp,InstructionTable *t)
{
    char opname[100],*p;
    InstDef *e;
    Form *f;
    int code;
    
    for (e = t->entries; e->opname != NULL; ++e) {
        fprintf(fp,"    base_%s,\n",e->opname);
        for (f = e->forms; f->pattern[0] != END; ++f) {
            strcpy(opname,e->opname);
            for (p = f->pattern; (code = *p++) != 0; ) {
                if (code == DUP || code == T || code == NO_STALLOP || code == SKIP)
                    goto skip;
                AppendOperandType(opname,code,&p);
            }
            fprintf(fp,"      op_%s",opname);
            if (f == e->forms)
                fprintf(fp," = base_%s",e->opname);
            fprintf(fp,",\n");
skip:       ; /* why is this needed? */
        }
    }
}

/* OutputOpcodeHandlers - output the opcode handlers */
void OutputOpcodeHandlers(FILE *fp,InstructionTable *t)
{
    char opname[100],*p;
    InstDef *e;
    Form *f;
    int code;
    
    for (e = t->entries; e->opname != NULL; ++e) {
        for (f = e->forms; f->pattern[0] != END; ++f) {
            strcpy(opname,e->opname);
            for (p = f->pattern; (code = *p++) != 0; ) {
                if (code == DUP || code == T || code == NO_STALLOP || code == SKIP)
                    goto skip;
                AppendOperandType(opname,code,&p);
            }
            fprintf(fp,"/* %s */\n",e->opname);
            fprintf(fp,"void h_%s(MPE *mpe,UINT32 *args)\n",opname);
            fprintf(fp,"{\n");
            fprintf(fp,"    CallHaltHook(mpe,hhError,mpe->pc.value);\n");
            fprintf(fp,"}\n\n");
skip:       ; /* why is this needed? */
        }
    }
}

/* OutputOpcodeTable - output the opcode table */
void OutputOpcodeTable(FILE *fp,InstructionTable *t)
{
    char opname[100],flags[100],eflags[100],*p;
    int spaces,code,argn;
    InstDef *e;
    Form *f;
    
    for (e = t->entries; e->opname != NULL; ++e) {
        for (f = e->forms; f->pattern[0] != END; ++f) {
            if (f->flags & regDecFlag)
                continue;
            strcpy(opname,e->opname);
            strcpy(flags,GetInstructionFlag(f->flags));
            if (f->flags & if48Bits) strcat(flags," | if48Bits");
            if (f->flags & if32Bits) strcat(flags," | if32Bits");
            eflags[0] = '\0'; argn = 0;
            for (p = f->pattern; (code = *p++) != 0; ) {
                if (code == DUP || code == T || code == NO_STALLOP || code == SKIP)
                    goto skip;
                AppendOperandType(opname,code,&p);
                switch (code) {
                case SR:
                    sprintf(&eflags[strlen(eflags)]," | ifArg%dIsSR",++argn);
                    break;
                case HR:
                    sprintf(&eflags[strlen(eflags)]," | ifArg%dIsHR",++argn);
                    break;
                case VR:
                case PR:
                    sprintf(&eflags[strlen(eflags)]," | ifArg%dIsVR",++argn);
                    break;
                case SHIFT:
                case ROTATE:
                    strcat(flags," | ifALUShifter");
                case IND:
                case DEC:
                    break;
                default:
                    ++argn;
                    break;
                }
            }
            spaces = 30 - strlen(opname);
            fprintf(fp,"{ \"%s\",%*sop_%s,\n",opname,spaces,"",opname);
            fprintf(fp,"%35s%s%s,\n","",flags,eflags);
            fprintf(fp,"%35sh_%s%*s},\n","",opname,spaces,"");
skip:       /* target for the label */;
        }
    }
}

/* GetInstructionFlag - get the instruction flag for an instruction */
char *GetInstructionFlag(int flags)
{
    if (flags & aluFlag)
        return "ifALU";
    else if (flags & multiplierFlag)
        return "ifMultiplier";
    else if (flags & registerFlag)
        return "ifRegister";
    else if (flags & memoryFlag)
        return "ifMemory";
    else if (flags & executeFlag)
        return "ifExecute";
    else
        return "0";
}
