/* mparse.h - merlin parser definitions */
/*
	Copyright (c) 1995, by VM Labs, Inc.
	All Rights Reserved
*/

#ifndef __MPARSE_H__
#define __MPARSE_H__

#include <setjmp.h>
#include "mpdll.h"
#include "merlin.h"

/* limits */
#define TKNSIZE		256	/* maximum token size */
#define LSIZE		1024	/* maximum line size */
#define INSTLENGTH	128	/* instruction length in bits */
#define MAXOPERANDS	4	/* maximum number of operands */

/* useful definitions */
#ifndef TRUE
#define TRUE	1
#define FALSE	0
#endif

/* useful types */
typedef long INT32;
typedef unsigned long UINT32;

/* segment structure */
typedef struct Segment {
    struct Segment *next;
    char *name;
    UINT32 lc;
} Segment;

/* fixed point value */
typedef	struct {
    long value;
    int fracbits;
} FixedValue;

/* function handler */
typedef void (*FHandler)(struct ScanContext *c,
                         struct Value *v,
                         int argc,
                         struct Value *argv);

/* value structure */
typedef struct Value {
    int type;			/* type */
    union {			/* value */
	char *_svalue;		/*   string value */
	double _rvalue;		/*   real value */
	FixedValue _fvalue;    	/*   fixed value */
        FHandler _fhandler;     /*   function value */
    } value;
} Value;

/* shorthands */
#define svalue	        value._svalue
#define rvalue	        value._rvalue
#define fvalue	        value._fvalue.value
#define ffracbits	value._fvalue.fracbits
#define fhandler        value._fhandler

/* type predicates */
#define IdentifierP(v)  ((v)->type == identifierType)
#define StringP(v)      ((v)->type == stringType)
#define FixedP(v)       ((v)->type == fixedType)
#define RealP(v)        ((v)->type == realType)
#define LabelP(v)       ((v)->type == labelType)
#define RegisterP(v)    ((v)->type == registerType)
#define FunctionP(v)    ((v)->type == functionType)

/* derived predicates */
#define IntegerP(v)     (FixedP(v) && (v)->ffracbits == 0)
#define NumberP(v)      (FixedP(v) || RealP(v))

/* numeric predicates */
#define FixedNumericP(v)    (FixedP(v) || LabelP(v))
#define NumericP(v)         (RealP(v) || FixedNumericP(v))

/* value types */
#define nilType	        0
#define identifierType	1
#define stringType	2
#define fixedType	3
#define realType	4
#define labelType	5
#define registerType	6
#define functionType    7

/* operand structure */
typedef struct {
    Value value;
    Value value2;
    int flags;
} Operand;

/* operand flags */
#define shiftFlag	1
#define rotateFlag	2
#define immediateFlag	4
#define decrementFlag   8
#define indirectFlag	16
#define value2Flag	32

/* instruction structure */
typedef struct Instruction {
    struct Instruction *next;
    char file[TKNSIZE+1];
    long line,lineCount;
    MScalar addr;
    UINT32 flags;
    int nOperands;
    Operand operands[MAXOPERANDS];
    struct InstDef *instDef;
    struct Form *form;
    char *beforeMethod;
    char *afterMethod;
} Instruction;

/* instruction flags */
/* note: these flags can't overlap ifStop and ifBreak in merlin.h */
/* note: the function unit flags can't overlap compound instruction flags below */
#define aluFlag		    0x00000001
#define multiplierFlag	    0x00000002
#define registerFlag	    0x00000004
#define memoryFlag	    0x00000008
#define executeFlag	    0x00000010
#define hasLongFormFlag     0x00000020
#define regDecFlag          0x00000040
#define functionUnitMask    0x0000001f

/* compound instruction structure */
typedef struct {
    UINT32 flags;
    MScalar length;
    Instruction *registerInst;
    Instruction *firstInst;
    char *beforeMethod;
    char *afterMethod;
} CompoundInstruction;

/* compound instruction flags (must not conflict with instruction flags) */
#define hasRegDec0Flag      0x80000000
#define hasRegDec1Flag      0x40000000

/* maximum pattern length */
#define MAXPAT	10

/* sinstruction handler */
typedef void (*iHandler)(struct ScanContext *c,Instruction *i,void *data);

/* instruction form structure */
typedef struct Form {
    char pattern[MAXPAT+1];
    long flags;
    iHandler handler;
} Form;

/* form pattern elements */
#define END	   -1  /* marker for end of patterns */
#define SR	   1   /* scalar register */
#define RZ	   2   /* rz register */
#define VR	   3   /* vector register */
#define PR	   4   /* pixel register */
#define RC0        5   /* rc0 register */
#define RC1        6   /* rc1 register */
#define PC	   7   /* pc register */
#define RUR	   8   /* register unit register */
#define RUV	   9   /* register ru or rv */
#define IMM	   10  /* immediate value */
#define UIMM	   11  /* unsigned immediate value */
#define RUIMM	   12  /* register unit immediate value */
#define SHIFT	   13  /* shift operator */
#define ROTATE	   14  /* rotate operator */
#define CC	   15  /* condition code */
#define IND	   16  /* indirect */
#define BILINEAR   17  /* bilinear index */
#define DSOFF4	   18  /* data space offset (4 byte alignment) */
#define DSOFF8	   19  /* data space offset (8 byte alignment) */
#define DSOFF16	   20  /* data space offset (16 byte alignment) */
#define STALLOP	   21  /* bra stall op */
#define NO_STALLOP 22  /* no stallop */
#define LABEL	   23  /* branch target */
#define RLABEL	   24  /* branch target for relative branches */
#define IO	   25  /* i/o register */
#define T	   26  /* implicit t condition code */
#define AC_SHIFT   27  /* ac_shift register */
#define SV_SHIFT   28  /* sv_shift register */
#define DUP        29  /* duplicate the previous operand */
#define SKIP       30  /* skip and fill with nil */
#define DEC        31  /* decrement operator */
#define SR_EQ      32  /* source register equal to another operand */
#define MULIMM     33  /* multiplier immediate */
#define HR         34  /* half vector register */

/* instruction definition */
typedef struct InstDef {
    char *opname;
    int opcode;
    Form *forms;
} InstDef;

/* instruction table */
typedef struct {
    int flag;
    InstDef *entries;
} InstructionTable;

/* constant data structure */
typedef struct ConstantData {
    struct ConstantData *next;
    char file[TKNSIZE+1];
    long line;
    long addr;
    Value value;
} ConstantData;

/* symbol structure */
typedef struct Symbol {
    char name[TKNSIZE+1];	/* symbol name */
    Value value;		/* symbol value */
    struct Symbol *next;
} Symbol;

/* export symbol structure */
typedef struct ExportSymbol {
    char name[TKNSIZE+1];       /* symbol name */
    struct ExportSymbol *next;
} ExportSymbol;


/* warning structure */
typedef struct Warning {
    struct Warning *next;
    char text[MPERRTEXTSIZE];
    char file[MPERRFILESIZE];
    long line;
    long offset;
} Warning;

/* file context */
typedef struct FileContext {
    char fileName[TKNSIZE+1];	/* file name */
    char moduleName[TKNSIZE+1]; /* module name */
    ExportSymbol *exportSymbols;/* import/export list */
    FILE *input;		/* input stream */
    Value value;		/* token value */
    char token[TKNSIZE+1];	/* token string */
    long lineNumber;		/* line number */
    int savedToken;		/* look ahead token */
    int savedChar;		/* look ahead character */
    char line[LSIZE+1];		/* last input line */
    char *linePtr;		/* line pointer */
    int atEOF;			/* input end of file flag */
    struct FileContext *next;	/* context of enclosing file */
} FileContext;

/* scanner context */
typedef struct ScanContext {
    FileContext *file;		/* current file context */
    Value start;		/* start address */
    Symbol *symbols;		/* list of symbol definitions */
    Segment *segments;          /* memory segments */
    Segment *segment;           /* current segment */
    Instruction *instructions;	/* list of instructions */
    Instruction **pNextInst;	/* place to store next instruction */
    ConstantData *data;		/* constant data */
    ConstantData **pNextData;	/* place to store next data */
    Warning *warnings;          /* warning messages */
    Warning **pNextWarning;     /* place to store next warning */
    mpErrorInfo *errorInfo;	/* error information structure */
    jmp_buf errorTarget;	/* error target */
} ScanContext;

/* shorthands */
#define t_token	    file->token
#define t_value	    file->value

/* tokens */
#define T_NOTOKEN   0
#define T_EOF	    -1

/* non-character tokens */
#define T_STRING	256
#define T_IDENTIFIER	257
#define T_NUMBER	258
#define T_SHIFT		T_SHR	/* '>>' */
#define T_ROTATE	259	/* '<>' */
#define T_LE		260	/* '<=' */
#define T_EQ		261	/* '==' */
#define T_NE		262	/* '!=' */
#define T_GE		263	/* '>=' */
#define T_SHL		264	/* '<<' */
#define T_SHR		265	/* '>>' */
#define T_AND		266	/* '&&' */
#define T_OR		267	/* '||' */
#define T_INC		268	/* '++' */
#define T_DEC		269	/* '--' */
#define T_ADDEQ		270	/* '+=' */
#define T_SUBEQ		271	/* '-=' */
#define T_MULEQ		272	/* '*=' */
#define T_DIVEQ		273	/* '/=' */
#define T_REMEQ		274	/* '%=' */
#define T_ANDEQ		275	/* '&=' */
#define T_OREQ		276	/* '|=' */
#define T_XOREQ		277	/* '^=' */
#define T_SHLEQ		278	/* '<<=' */
#define T_SHREQ		279	/* '>>=' */
#define T_EOL		280

/* prototypes for mparse.c */
long GetStartAddress(ScanContext *c);
int CheckInstructionAddress(ScanContext *c,long addr);
long GetNumericValue(ScanContext *c,Value *v,char *err);
int IsKeyword(char *id);
void Abort(ScanContext *c,int sts);

/* prototypes for noutput.c */
void OutputInstructions(ScanContext *c,FILE *fp);
void AppendOperandType(char *opname,int code,char **ppat);

/* prototypes for boutput.c */
void OutputBinaryData(ScanContext *c,FILE *fp);

/* prototypes for mexpr.c */
void InitExpressionEvaluator(ScanContext *c);
void EvalExpression(ScanContext *c,Value *value);
void EvalAssignmentExpression(ScanContext *c,char *name,int op);
int AssignmentOperatorP(int tkn);
char *MakeFullName(ScanContext *c,char *name);

/* prototypes for msymbol.c */
void InitSymbols(ScanContext *c);
Symbol *AddSymbol(ScanContext *c,char *name,int type,void *pValue);
void AddNamedFunction(ScanContext *c,char *name,FHandler handler);
Symbol *FindSymbol(ScanContext *c,char *name);
void FreeSymbols(ScanContext *c);
ExportSymbol *AddExportSymbol(ScanContext *c,char *name);
ExportSymbol *FindExportSymbol(ScanContext *c,char *name);
void FreeExportSymbols(FileContext *c);
void SetValue(ScanContext *c,Value *v,int type,void *pValue);
void SetFixedValue(ScanContext *c,Value *v,long value,int fracbits);
void ConvertToReal(Value *v);
void ConvertToFixed(Value *v,int fracbits);
void DisplayValue(Value *v,FILE *fp);
void FreeValue(Value *v);
void CopyValue(ScanContext *c,Value *dst,Value *src);

/* prototypes for mscan.c */
ScanContext *InitScan(mpErrorInfo *errInfo);
void FreeScan(ScanContext *c);
int IncludeFile(ScanContext *c,char *filename,FILE *fp);
void FRequire(ScanContext *c,int rtkn);
void Require(ScanContext *c,int tkn,int rtkn);
int NonEOLToken(ScanContext *c);
int Token(ScanContext *c);
void SaveToken(ScanContext *c,int tkn);
char *TokenName(int tkn);
int SkipSpaces(ScanContext *c);
int GetCh(ScanContext *c);
void UngetCh(ScanContext *c,int ch);
void SetFileInfo(ScanContext *c,char *file,long line,long offset);
void ParseError(ScanContext *c,char *fmt,...);
void ParseWarning(ScanContext *c,char *fmt,...);

#endif
