/* mscan.c - merlin lexical scanner */
/*
	Copyright (c) 1995, by VM Labs, Inc.
	All Rights Reserved
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "mparse.h"

/* token name table entry */
typedef struct {
    char *name;
    int token;
} TokenNameEntry;

/* token name table */
static TokenNameEntry t_names[] = {
{	"<string>",	T_STRING	},
{	"<identifier>",	T_IDENTIFIER	},
{	"<number>",	T_NUMBER	},
{	"<>",		T_ROTATE	},
{	"<=",		T_LE		},
{	"==",		T_EQ		},
{	"!=",		T_NE		},
{	">=",		T_GE		},
{	"<<",		T_SHL		},
{	">>",		T_SHR		},
{	"&&",		T_AND		},
{	"||",		T_OR		},
{	"++",		T_INC		},
{	"--",		T_DEC		},
{	"+=",		T_ADDEQ		},
{	"-=",		T_SUBEQ		},
{	"*=",		T_MULEQ		},
{	"/=",		T_DIVEQ		},
{	"%=",		T_REMEQ		},
{	"&=",		T_ANDEQ		},
{	"|=",		T_OREQ		},
{	"^=",		T_XOREQ		},
{	"<<=",		T_SHLEQ		},
{	">>=",		T_SHREQ		},
{	"<eol>",	T_EOL		},
{	"<eof>",	T_EOF		},
{	NULL,		0		}
};

/* prototypes */
static void FreeFile(FileContext *f);
static void FreeFiles(ScanContext *c);
static int rtoken(ScanContext *c);
static int popfile(ScanContext *c);
static int singlechtoken(ScanContext *c,int ch);
static int getstring(ScanContext *c);
static int getcharacter(ScanContext *c);
static int literalch(ScanContext *c);
static int getid(ScanContext *c,int ch);
static int getnumber(ScanContext *c,int ch);
static int getxnumber(ScanContext *c);
static int chtodigit(int ch);
static int isidchar(int ch);
static void ErrorPrintF(ScanContext *c,char *fmt,...);
static void vErrorPrintF(ScanContext *c,char *fmt,va_list ap);

/* InitScan - initialize the scanner */
ScanContext *InitScan(mpErrorInfo *errInfo)
{
    FixedValue defaultStart = {instructionROMBase,0};
    ScanContext *c;

    /* allocate a scan context structure */
    if ((c = (ScanContext *)malloc(sizeof(ScanContext))) == NULL)
	return NULL;

    /* save the error info structure */
    c->errorInfo = errInfo;
    
    /* no file context yet */
    c->file = NULL;

    /* default start location */
    SetValue(c,&c->start,fixedType,&defaultStart);
    
    /* initialize symbols */
    InitSymbols(c);

    /* initialize segments */
    c->segments = NULL;

    /* no instructions yet */
    c->instructions = NULL;
    c->pNextInst = &c->instructions;

    /* no constant data yet */
    c->data = NULL;
    c->pNextData = &c->data;

    /* no warnings yet */
    c->warnings = NULL;
    c->pNextWarning = &c->warnings;
    
    /* return the scan context */
    return c;
}

/* FreeScan - free a scan context */
void FreeScan(ScanContext *c)
{
    FreeSymbols(c);
    FreeFiles(c);
    free(c);
}

/* IncludeFile - include a file */
int IncludeFile(ScanContext *c,char *filename,FILE *fp)
{
    FileContext *f;

    /* allocate a file context */
    if ((f = (FileContext *)malloc(sizeof(FileContext))) == NULL)
	return FALSE;
    
    /* remember the filename and input stream */
    strcpy(f->fileName,filename);
    f->input = fp;

    /* same module name as parent */
    strcpy(f->moduleName,c->file ? c->file->moduleName : "");
    f->exportSymbols = NULL;
    
    /* setup the line buffer */
    f->linePtr = f->line; *f->linePtr = '\0';
    f->lineNumber = 0;

    /* no lookahead yet */
    f->savedToken = T_NOTOKEN;
    f->savedChar = '\0';

    /* not eof yet */
    f->atEOF = FALSE;

    /* push the new file context */
    f->next = c->file;
    c->file = f;

    /* return successfully */
    return TRUE;
}

/* FreeFile - close free a file */
static void FreeFile(FileContext *f)
{
    FreeExportSymbols(f);
    fclose(f->input);
    free(f);
}

/* FreeFiles - free included files */
static void FreeFiles(ScanContext *c)
{
    FileContext *f,*next;
    for (f = c->file; f != NULL; f = next) {
    	next = f->next;
	FreeFile(f);
    }
}

/* FRequire - fetch a token and check it */
void FRequire(ScanContext *c,int rtkn)
{
    Require(c,Token(c),rtkn);
}

/* Require - check for a required token */
void Require(ScanContext *c,int tkn,int rtkn)
{
    char tknbuf[TKNSIZE+1];
    if (tkn != rtkn) {
	strcpy(tknbuf,TokenName(rtkn));
	ParseError(c,"Expecting '%s', found '%s'",tknbuf,TokenName(tkn));
    }
}

/* NonEOLToken - get the next token */
int NonEOLToken(ScanContext *c)
{
    int tkn;
    while ((tkn = Token(c)) == T_EOL)
	;
    return tkn;
}

/* Token - get the next token */
int Token(ScanContext *c)
{
    int tkn;
    if ((tkn = c->file->savedToken) != T_NOTOKEN)
	c->file->savedToken = T_NOTOKEN;
    else
	tkn = rtoken(c);
    return tkn;
}

/* SaveToken - save a token */
void SaveToken(ScanContext *c,int tkn)
{
    c->file->savedToken = tkn;
}

/* TokenName - get the name of a token */
char *TokenName(int tkn)
{
    static char tname[2];
    TokenNameEntry *n;
    for (n = t_names; n->name != NULL; ++n)
	if (tkn == n->token)
	    return n->name;
    tname[0] = tkn;
    tname[1] = '\0';
    return tname;
}

/* rtoken - read the next token */
static int rtoken(ScanContext *c)
{
    int ch,ch2;

    /* check the next character */
    for (;;)
	switch (ch = SkipSpaces(c)) {
	case EOF:	return popfile(c);
	case '\n':	return T_EOL;
	case '"':	return getstring(c);
	case '<':	switch (ch = GetCh(c)) {
			case '=':
			    return T_LE;
			case '>':
			    return T_ROTATE;
			case '<':
			    if ((ch = GetCh(c)) == '=')
				return T_SHLEQ;
			    c->file->savedChar = ch;
			    return T_SHL;
			default:
			    c->file->savedChar = ch;
			    return singlechtoken(c,'<');
			}
	case '=':	if ((ch = GetCh(c)) == '=')
			    return T_EQ;
			c->file->savedChar = ch;
			return singlechtoken(c,'=');
	case '!':	if ((ch = GetCh(c)) == '=')
			    return T_NE;
			c->file->savedChar = ch;
			return singlechtoken(c,'!');
	case '>':	switch (ch = GetCh(c)) {
			case '=':
			    return T_GE;
			case '>':
			    if ((ch = GetCh(c)) == '=')
				return T_SHREQ;
			    c->file->savedChar = ch;
			    return T_SHR;
			default:
			    c->file->savedChar = ch;
			    return singlechtoken(c,'>');
			}
	case '&':	switch (ch = GetCh(c)) {
			case '&':
			    return T_AND;
			case '=':
			    return T_ANDEQ;
			default:
			    c->file->savedChar = ch;
			    return singlechtoken(c,'&');
			}
	case '|':	switch (ch = GetCh(c)) {
			case '|':
			    return T_OR;
			case '=':
			    return T_OREQ;
			default:
			    c->file->savedChar = ch;
			    return singlechtoken(c,'|');
			}
	case '^':	if ((ch = GetCh(c)) == '=')
			    return T_XOREQ;
			c->file->savedChar = ch;
			return singlechtoken(c,'^');
	case '+':	switch (ch = GetCh(c)) {
			case '+':
			    return T_INC;
			case '=':
			    return T_ADDEQ;
			default:
			    c->file->savedChar = ch;
			    return singlechtoken(c,'+');
			}
	case '-':	switch (ch = GetCh(c)) {
			case '-':
			    return T_DEC;
			case '=':
			    return T_SUBEQ;
			default:
			    c->file->savedChar = ch;
			    return singlechtoken(c,'-');
			}
	case '*':	if ((ch = GetCh(c)) == '=')
			    return T_MULEQ;
			c->file->savedChar = ch;
			return singlechtoken(c,'*');
	case '/':	switch (ch = GetCh(c)) {
			case '=':
			    return T_DIVEQ;
			case '/':
			    while ((ch = GetCh(c)) != EOF)
				if (ch == '\n')
				    break;
			    break;
			case '*':
			    ch = ch2 = EOF;
			    for (; (ch2 = GetCh(c)) != EOF; ch = ch2)
				if (ch == '*' && ch2 == '/')
				    break;
			    break;
			default:
			    c->file->savedChar = ch;
			    return singlechtoken(c,'/');
			}
			break;
	case ';':	while ((ch = GetCh(c)) != EOF)
			    if (ch == '\n') {
				c->file->savedChar = ch;
				break;
			    }
			break;
	case '$':	return getxnumber(c);
	default:	if (isdigit(ch))
			    return getnumber(c,ch);
			else if (isidchar(ch))
			    return getid(c,ch);
			else
			    return singlechtoken(c,ch);
	}
}

/* popfile - pop a file off the include stack */
static int popfile(ScanContext *c)
{
    FileContext *next = c->file->next;
    FreeFile(c->file);
    c->file = next;
    return c->file ? rtoken(c) : T_EOF;
}

/* singlechtoken - make a single character token */
static int singlechtoken(ScanContext *c,int ch)
{
    c->file->token[0] = ch;
    c->file->token[1] = '\0';
    return ch;
}

/* getstring - get a string */
static int getstring(ScanContext *c)
{
    int ch,len=0;
    char *p;

    /* get the string */
    p = c->file->token;
    while ((ch = literalch(c)) != EOF && ch != '"') {
	if (++len > TKNSIZE)
	    ParseError(c,"String too long");
	*p++ = ch;
    }
    if (ch == EOF)
	c->file->savedChar = EOF;
    *p = '\0';
    return T_STRING;
}

/* getcharacter - get a character constant */
static int getcharacter(ScanContext *c)
{
    c->file->value.fvalue = literalch(c);
    c->file->value.ffracbits = 0;
    c->file->token[0] = (char)c->file->value.fvalue;
    c->file->token[1] = '\0';
    if (GetCh(c) != '\'')
	ParseError(c,"Expecting a closing single quote");
    return T_NUMBER;
}

/* literalch - get a character from a literal string */
static int literalch(ScanContext *c)
{
    int ch;
    if ((ch = GetCh(c)) == '\\')
	switch (ch = GetCh(c)) {
	case 'n':  ch = '\n'; break;
	case 'r':  ch = '\r'; break;
	case 't':  ch = '\t'; break;
	case EOF:  ch = '\\'; c->file->savedChar = EOF; break;
	}
    return ch;
}

/* getid - get an identifier */
static int getid(ScanContext *c,int ch)
{
    int len = 1;
    char *p;

    /* get the identifier */
    p = c->file->token; *p++ = ch;
    while ((ch = GetCh(c)) != EOF && isidchar(ch)) {
	if (++len > TKNSIZE)
	    ParseError(c,"Identifier too long");
	*p++ = ch;
    }
    *p = '\0';
    c->file->savedChar = ch;
    return T_IDENTIFIER;
}

/* getnumber - get a number */
static int getnumber(ScanContext *c,int ch)
{
    char *p = c->file->token;
    int real = FALSE;

    /* store the initial digit */
    *p++ = ch;

    /* get the integer part */
    while ((ch = GetCh(c)) != EOF && isdigit(ch))
	*p++ = ch;

    /* check for a decimal point */
    if (ch == '.') {

        /* remember the decimal point */
        real = TRUE;
        *p++ = ch;

        /* get the fractional part */
        while ((ch = GetCh(c)) != EOF && isdigit(ch))
	    *p++ = ch;
    }

    /* check for an exponent */
    if (ch == 'E' || ch == 'e') {
        real = TRUE;
	*p++ = ch;

	/* check for a sign */
	if ((ch = GetCh(c)) == '+' || ch == '-') {
	    *p++ = ch;
            ch = GetCh(c);
        }

	/* get the exponent */
	while (ch != EOF && isdigit(ch)) {
	    *p++ = ch;
            ch = GetCh(c);
        }
    }
    *p = '\0';

    /* delete a final decimal point */
    if (p[-1] == '.')
        p[-1] = '\0';

    /* convert the string to an integer and return successfully */
    if (real) {
        double value = atof(c->t_token);
        SetValue(c,&c->t_value,realType,&value);
    }
    else {
        FixedValue value = {0,0};
        value.value = atol(c->t_token);
        SetValue(c,&c->t_value,fixedType,&value);
    }
    c->file->savedChar = ch;
    return T_NUMBER;
}

/* getxnumber - get a hex number */
static int getxnumber(ScanContext *c)
{
    FixedValue value = {0,0};
    char *p;
    int ch;
    
    /* get the number */
    p = c->file->token;
    while ((ch = GetCh(c)) != EOF && isxdigit(ch)) {
	value.value = (value.value << 4) + chtodigit(ch);
	*p++ = ch;
    }
    SetValue(c,&c->t_value,fixedType,&value);
    c->file->savedChar = ch;
    *p = '\0';
    return T_NUMBER;
}

/* chtodigit - convert a character to a digit */
static int chtodigit(int ch)
{
    if (ch >= '0' && ch <= '9')
	return ch - '0';
    else if (ch >= 'a' && ch <= 'f')
	return ch - 'a' + 10;
    else if (ch >= 'A' && ch <= 'F')
	return ch - 'A' + 10;
    else
	return -1;
}

/* SkipSpaces - skip leading spaces */
int SkipSpaces(ScanContext *c)
{
    int ch;
    while ((ch = GetCh(c)) != '\0' && ch != '\n' && isspace(ch))
	;
    return ch;
}

/* isidchar - is this an identifier character */
static int isidchar(int ch)
{
    return isupper(ch)
        || islower(ch)
        || isdigit(ch)
        || ch == '_'
	|| ch == '.';
}

/* GetCh - get the next character */
int GetCh(ScanContext *c)
{
    int ch;

    /* check for a lookahead character */
    if ((ch = c->file->savedChar) != '\0') {
	c->file->savedChar = '\0';
	return ch;
    }

    /* check for a buffered character */
    while (*c->file->linePtr == '\0') {

	/* check for end of file */
	if (c->file->atEOF)
	    return EOF;

	/* read another line */
	else {

	    /* read the line */
	    c->file->linePtr = c->file->line;
	    while ((ch = getc(c->file->input)) != EOF && ch != '\n')
		*c->file->linePtr++ = ch;
	    *c->file->linePtr++ = '\n'; *c->file->linePtr = '\0';
	    c->file->linePtr = c->file->line;
	    ++c->file->lineNumber;

	    /* check for end of file */
	    if (ch < 0)
	 	c->file->atEOF = TRUE;
	 }
    }

    /* return the current character */
    return *c->file->linePtr++;
}

/* UngetCh - unget a character */
void UngetCh(ScanContext *c,int ch)
{
    c->file->savedChar = ch;
}

/* SetFileInfo - set the file and line number information in the error info structure */
void SetFileInfo(ScanContext *c,char *file,long line,long offset)
{
    strcpy(c->errorInfo->file,file);
    c->errorInfo->line = line;
    c->errorInfo->offset = offset;
}

/* ParseError - report an error in the current line */
void ParseError(ScanContext *c,char *fmt,...)
{
    FileContext *f = c->file;
    va_list ap;

    /* format the error message */
    va_start(ap,fmt);
    vsprintf(c->errorInfo->text,fmt,ap);
    va_end(ap);

    /* handle the case where there is a file open */
    if (f) SetFileInfo(c,f->fileName,f->lineNumber,f->linePtr - f->line);

    /* store the error code */
    c->errorInfo->status = mpsError;

    /* abort the assembly */
    Abort(c,mpsError);
}

/* ParseWarning - report a non-fatal error in the current line */
void ParseWarning(ScanContext *c,char *fmt,...)
{
    FileContext *f = c->file;
    Warning *w;
    va_list ap;

    /* allocate the warning structure */
    if ((w = (Warning *)malloc(sizeof(Warning))) == NULL)
	ParseError(c,"Insufficient memory");

    /* format the error message */
    va_start(ap,fmt);
    vsprintf(w->text,fmt,ap);
    va_end(ap);

    /* handle the case where there is a file open */
    if (f) {
        strcpy(w->file,f->fileName);
        w->line = f->lineNumber;
        w->offset = f->linePtr - f->line;
    }

    /* no file open */
    else
        w->file[0] = '\0';

    /* store the warning */
    *c->pNextWarning = w;
    c->pNextWarning = &w->next;
    w->next = NULL;
}
