/* msymbol.c - symbol table routines */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mparse.h"

/* prototypes */
static void AddNumberedRegister(ScanContext *c,char *baseName,int n);
static void AddNamedRegister(ScanContext *c,char *name);

/* InitSymbols - initialize the symbol table routines */
void InitSymbols(ScanContext *c)
{
    int i;
    
    /* start with an empty symbol table */
    c->symbols = NULL;
    
    /* scalar registers */
    for (i = 0; i < 32; ++i)
	AddNumberedRegister(c,"r",i);

    /* pixel registers (synonyms for vector registers) */
    for (i = 0; i < 8; ++i)
	AddNumberedRegister(c,"p",i);

    /* vector registers */
    for (i = 0; i < 8; ++i)
	AddNumberedRegister(c,"v",i);

    /* program counter and stack pointer */
    AddNamedRegister(c,"pc");
    AddNamedRegister(c,"sp");

    /* x and y index registers */
    AddNamedRegister(c,"rx");
    AddNamedRegister(c,"ry");
    AddNamedRegister(c,"xy");
    AddNamedRegister(c,"xy_width");
    AddNamedRegister(c,"xy_itype");
    AddNamedRegister(c,"xy_base");
    AddNamedRegister(c,"xy_mipmap");
    AddNamedRegister(c,"xy_tile");

    /* u and v index registers */
    AddNamedRegister(c,"ru");
    AddNamedRegister(c,"rv");
    AddNamedRegister(c,"uv");
    AddNamedRegister(c,"uv_width");
    AddNamedRegister(c,"uv_itype");
    AddNamedRegister(c,"uv_base");
    AddNamedRegister(c,"uv_mipmap");
    AddNamedRegister(c,"uv_tile");

    /* dma registers */
    AddNamedRegister(c,"dma_1");
    AddNamedRegister(c,"dma_eaddr");
    AddNamedRegister(c,"dma_flags");
    AddNamedRegister(c,"dma_length");
    AddNamedRegister(c,"dma_iaddr");
    AddNamedRegister(c,"dma_2");
    AddNamedRegister(c,"dma_x");
    AddNamedRegister(c,"dma_y");
    AddNamedRegister(c,"dma_stat");

    /* special registers */
    AddNamedRegister(c,"rc0");
    AddNamedRegister(c,"rc1");
    AddNamedRegister(c,"sv_shift");
    AddNamedRegister(c,"ac_shift");
    AddNamedRegister(c,"rz");
}

/* AddNumberedRegister - add a numbered register to the symbol table */
static void AddNumberedRegister(ScanContext *c,char *baseName,int n)
{
    char name[10];
    sprintf(name,"%s%d",baseName,n);
    AddSymbol(c,name,registerType,name);
}

/* AddNamedRegister - add a named register to the symbol table */
static void AddNamedRegister(ScanContext *c,char *name)
{
    AddSymbol(c,name,registerType,name);
}

/* AddNamedFunction - add a named function to the symbol table */
void AddNamedFunction(ScanContext *c,char *name,FHandler handler)
{
    AddSymbol(c,name,functionType,handler);
}

/* AddSymbol - add a symbol definition to the symbol table */
Symbol *AddSymbol(ScanContext *c,char *name,int type,void *pValue)
{
    Symbol *s;

    /* look for an existing symbol table entry */
    for (s = c->symbols; s != NULL; s = s->next)
	if (strcmp(name,s->name) == 0)
	    goto storeValue;

    /* allocate a new symbol structure */
    if ((s = (Symbol *)malloc(sizeof(Symbol))) == NULL)
	ParseError(c,"Insufficient memory");
    strcpy(s->name,name);
    s->value.type = nilType;
    s->next = c->symbols;
    c->symbols = s;

storeValue:
    /* free the old value and set the new value */
    FreeValue(&s->value);
    SetValue(c,&s->value,type,pValue);

    /* return the symbol */
    return s;
}

/* FindSymbol - find a symbol definition in the symbol table */
Symbol *FindSymbol(ScanContext *c,char *name)
{
    Symbol *s;
    for (s = c->symbols; s != NULL; s = s->next)
	if (strcmp(name,s->name) == 0)
	    return s;
    return NULL;
}

/* FreeSymbols - free all symbols */
void FreeSymbols(ScanContext *c)
{
    Symbol *s,*next;
    for (s = c->symbols; s != NULL; s = next) {
    	next = s->next;
	FreeValue(&s->value);
	free(s);
    }
}

/* AddExportSymbol - add an export symbol definition */
ExportSymbol *AddExportSymbol(ScanContext *c,char *name)
{
    ExportSymbol *s;

    /* look for an existing symbol table entry */
    for (s = c->file->exportSymbols; s != NULL; s = s->next)
	if (strcmp(name,s->name) == 0)
	    return s;

    /* allocate a new symbol structure */
    if ((s = (ExportSymbol *)malloc(sizeof(ExportSymbol))) == NULL)
	ParseError(c,"Insufficient memory");
    strcpy(s->name,name);
    s->next = c->file->exportSymbols;
    c->file->exportSymbols = s;
    return s;
}

/* FindExportSymbol - find an export symbol definition */
ExportSymbol *FindExportSymbol(ScanContext *c,char *name)
{
    ExportSymbol *s;
    for (s = c->file->exportSymbols; s != NULL; s = s->next)
	if (strcmp(name,s->name) == 0)
	    return s;
    return NULL;
}

/* FreeExportSymbols - free all symbols */
void FreeExportSymbols(FileContext *c)
{
    ExportSymbol *s,*next;
    for (s = c->exportSymbols; s != NULL; s = next) {
    	next = s->next;
	free(s);
    }
}

/* SetValue - set the value of a value structure */
void SetValue(ScanContext *c,Value *v,int type,void *pValue)
{
    switch (v->type = type) {
    case identifierType:
    case registerType:
    case stringType:
	if ((v->svalue = malloc(strlen((char *)pValue) + 1)) == NULL)
	    ParseError(c,"Insufficient memory");
	strcpy(v->svalue,(char *)pValue);
	break;
    case labelType:
    case fixedType:
	v->fvalue = ((FixedValue *)pValue)->value;
        v->ffracbits = ((FixedValue *)pValue)->fracbits;
	break;
    case realType:
        v->rvalue = *(double *)pValue;
	break;
    case functionType:
        v->fhandler = (FHandler)pValue;
        break;
    }
}

/* SetFixedValue - set a value to a fixed */
void SetFixedValue(ScanContext *c,Value *v,long value,int fracbits)
{
    v->type = fixedType;
    v->fvalue = value;
    v->ffracbits = fracbits;
}

/* ConvertToReal - convert a fixed to a real */
void ConvertToReal(Value *v)
{
    v->type = realType;
    v->rvalue = (double)v->fvalue / (1 << v->ffracbits);
}

/* ConvertToFixed - convert a real to a fixed */
void ConvertToFixed(Value *v,int fracbits)
{
    v->type = fixedType;
    v->fvalue = (long)(v->rvalue * (1 << fracbits));
    v->ffracbits = fracbits;
}

/* DisplayValue - display a value */
void DisplayValue(Value *v,FILE *fp)
{
    switch (v->type) {
    case identifierType:
	fprintf(fp,"<Identifier: %s>",v->svalue);
	break;
    case registerType:
	fprintf(fp,"<Register: %s>",v->svalue);
	break;
    case stringType:
	fprintf(fp,"<String: '%s'>",v->svalue);
	break;
    case labelType:
	fprintf(fp,"<Label: %ld>",v->fvalue);
	break;
    case fixedType:
	fprintf(fp,"<Fixed: %ld>",v->fvalue);
	break;
    case realType:
	fprintf(fp,"<Real: %g>",v->rvalue);
	break;
    case functionType:
	fprintf(fp,"<Function: %lx>",(long)v->fhandler);
	break;
    default:
	fprintf(fp,"<Unknown: %d>",v->type);
	break;
    }
}

/* FreeValue - free the current value of a value structure */
void FreeValue(Value *v)
{
    switch (v->type) {
    case identifierType:
    case registerType:
    case stringType:
	if (v->svalue)
	    free(v->svalue);
	break;
    }
    v->type = nilType;
}

/* CopyValue - copy a value structure */
void CopyValue(ScanContext *c,Value *dst,Value *src)
{
    *dst = *src;
    switch (src->type) {
    case identifierType:
    case registerType:
    case stringType:
	dst->svalue = malloc(strlen(src->svalue) + 1);
	if (dst->svalue == NULL)
	    ParseError(c,"Insufficient memory");
	strcpy(dst->svalue,src->svalue);
	break;
    }
}
