/* noutput.c - routines to generate lisp output from merlin source */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include "mparse.h"

/* prototypes */
static void OutputWarnings(ScanContext *c,FILE *fp);
static int OutputProlog(ScanContext *c,FILE *fp);
static int OutputSymbols(ScanContext *c,FILE *fp);
static void OutputValue(ScanContext *c,Value *v,FILE *fp);
static void OutputConstantData(ScanContext *c,FILE *fp);
static void OutputInstruction(ScanContext *c,Instruction *i,FILE *fp);
static void OutputInstructionCode(ScanContext *c,Instruction *i,FILE *fp);
static void RangeCheckLabelReference(ScanContext *c,int bits,MScalar offset);
static void OutputPostlog(ScanContext *c,FILE *fp,int level);
static MScalar NextInstructionAddress(Instruction *i);
static int InstructionSize(Instruction *i);
static void OutputString(char *str,FILE *fp);

/* OutputInstructions - output instructions in lisp format */
void OutputInstructions(ScanContext *c,FILE *fp)
{
    Instruction *i;
    int level;

    /* output warning messages */
    OutputWarnings(c,fp);

    /* output the prolog */
    level = OutputProlog(c,fp);

    /* output constant data */
    OutputConstantData(c,fp);

    /* output the postlog */
    OutputPostlog(c,fp,level);
    
    /* loop through the instructions */
    for (i = c->instructions; i != NULL; i = i->next)
	OutputInstruction(c,i,fp);
}

/* OutputWarnings - output the warning messages */
static void OutputWarnings(ScanContext *c,FILE *fp)
{
    Warning *w = c->warnings;
    while (w != NULL) {
        fprintf(fp,"(mparse-warning ");
        OutputString(w->text,fp);
        if (w->file[0]) {
            putc(' ',fp);
            OutputString(w->file,fp);
            putc(' ',fp);
            fprintf(fp," %ld %ld",w->line,w->offset);
        }
        fprintf(fp,")\n");
        w = w->next;
    }
}

/* OutputProlog - output the lisp prolog */
static int OutputProlog(ScanContext *c,FILE *fp)
{
    int level;
    
    /* say what this file is */
    fprintf(fp,"; Output automatically generated by MPARSE");

    /* bind the symbols */
    level = OutputSymbols(c,fp);

    /* store the symbol table */
    fprintf(fp,"\n  (&d 'set-symbols! (the-environment))");
                  
    /* output the start address */
    fprintf(fp,"\n  (&d 'set-start-address! #x%08lx)",GetStartAddress(c));

    /* return the nesting level */
    return level;
}

#define FIRSTENV 2 /* constant from xlisp.h */

/* OutputSymbols - output the symbol table */
static int OutputSymbols(ScanContext *c,FILE *fp)
{
    int count = FIRSTENV + 2;
    int first = FALSE;
    int level = 2;
    Symbol *s;
    fprintf(fp,"\n(let ((&mpe &m)");
    fprintf(fp,"\n      (&debugger &d)");
    for (s = c->symbols; s != NULL; s = s->next) {
        if (!FunctionP(&s->value)) {
            if (++count >= 256) {
                fprintf(fp,")\n%*s(let (",level,"");
                count = FIRSTENV;
                first = TRUE;
                level += 2;
            }
 	    if (first) first = FALSE;
            else fprintf(fp,"\n    %*s",level,"");
            fprintf(fp,"(%s ",s->name);
	    OutputValue(c,&s->value,fp);
	    putc(')',fp);
        }
    }
    fprintf(fp,")");
    return level / 2;
}

/* OutputValue - output a typed value */
static void OutputValue(ScanContext *c,Value *v,FILE *fp)
{
    switch (v->type) {
    case nilType:
        fprintf(fp,"nil");
        break;
    case identifierType:
	fprintf(fp,"'%s",v->svalue);
	break;
    case labelType:
	fprintf(fp,"#x%08lx",v->fvalue);
	break;
    case fixedType:
	fprintf(fp,"%ld",v->fvalue);
	break;
    case realType:
	fprintf(fp,"%g",v->rvalue);
	break;
    case registerType:
    	fprintf(fp,"(&m 'find-register-by-name '%s)",v->svalue);
	break;
    }
}

/* OutputConstantData - output constant data in lisp format */
static void OutputConstantData(ScanContext *c,FILE *fp)
{
    ConstantData *d;
    MScalar value;
    if (d = c->data) {
	fprintf(fp,"\n  (map (lambda (b)");
	fprintf(fp,"\n         (&m 'init-scalar! (car b) (cdr b)))");
	fprintf(fp,"\n       '(");
        for (; d != NULL; d = d->next) {
	    SetFileInfo(c,d->file,d->line,0); /* in case of an error */
	    if (d != c->data)
	        fprintf(fp,"\n         ");
            if (d->value.type == stringType)
                fprintf(fp,"(#x%08lx . \"%s\")",d->addr,d->value.svalue);
            else {
	        value = GetNumericValue(c,&d->value,"Invalid constant data");
                fprintf(fp,"(#x%08lx . %ld)",d->addr,value);
            }
	}
	fprintf(fp,"))");
    }
}

/* OutputPostlog - output the lisp postlog */
static void OutputPostlog(ScanContext *c,FILE *fp,int level)
{
    /* end of let and load function definition */
    while (--level >= 0)
        fprintf(fp,")");
    fprintf(fp,"\n");
}

/* OutputInstruction - output an instruction in lisp format */
static void OutputInstruction(ScanContext *c,Instruction *i,FILE *fp)
{
    /* output the prolog */
    fprintf(fp,"\n(&m 'store-instruction! #x%08lx",i->addr);
    fprintf(fp,"\n  :debugger &d");
    fprintf(fp,"\n  :source '(");
    OutputString(i->file,fp);
    fprintf(fp," %ld %ld)",i->line,i->lineCount);
    
    /* output the instruction code */
    fprintf(fp,"\n  :code ");
    OutputInstructionCode(c,i,fp);
    
    /* output the before/after methods */
    if (i->beforeMethod)
	fprintf(fp,"\n  :before '%s",i->beforeMethod);
    if (i->afterMethod)
	fprintf(fp,"\n  :after '%s",i->afterMethod);
    fprintf(fp,")");
}

/* OutputInstructionCode - output the instruction code */
static void OutputInstructionCode(ScanContext *c,Instruction *i,FILE *fp)
{
    char opname[LSIZE],args[LSIZE],*pat,*dummy;
    Operand *op = i->operands;
    int n = i->nOperands;
    int shift = FALSE;
    int rotate = FALSE;
    UINT32 flags;
    int code;

    /* initialize */
    strcpy(opname,i->instDef->opname);
    args[0] = '\0';

    /* build the opcode */
    for (pat = i->form->pattern; (code = *pat++) != 0; )
        if (code == DUP)
            AppendOperandType(opname,pat[-2],&dummy);
        else
            AppendOperandType(opname,code,&pat);

    /* compute the flags */
    flags = i->flags & (ifStop | ifBreak);

    /* match each pattern element */
    for (pat = i->form->pattern; (code = *pat++) != 0; )
        switch (code) {
        case SR_EQ:     /* scalar register matching another operand */
            ++pat;
	case SR:	/* scalar register */
        case HR:        /* half vector register */
	case VR:	/* vector register */
	case PR:	/* pixel register */
	case RZ:	/* rz register */
	case PC:	/* pc register */
	case RUR:	/* register unit register */
	case RUV:	/* register ru or rv */
	case IO:	/* i/o register */
        case AC_SHIFT:  /* ac_shift register */
        case SV_SHIFT:  /* sv_shift register */
	    sprintf(&args[strlen(args)]," %s",op->value.svalue);
	    ++op;
	    break;
	case RC0:	/* counter registers */
        case RC1:
	    if (op->value.type == nilType)
                strcat(args," nil");
            else
                sprintf(&args[strlen(args)]," %s",op->value.svalue);
	    ++op;
	    break;
        case SKIP:
            ++pat;
            if (op->value.type == nilType)
                sprintf(&args[strlen(args)]," nil");
            else
                sprintf(&args[strlen(args)]," %s",op->value.svalue);
            ++op;
            break;
	case IMM:	/* immediate value */
	case UIMM:	/* unsigned immediate value */
	    ++pat;
	    sprintf(&args[strlen(args)]," %ld",
		    op->value.fvalue);
	    ++op;
	    break;
	case RUIMM:	/* register unit immediate value */
	    ++pat;
	    sprintf(&args[strlen(args)]," %ld",
		    op->value.fvalue << 16);
	    ++op;
	    break;
	case MULIMM:	/* multiplier immediate value */
	    sprintf(&args[strlen(args)]," %ld",
		    op->value.fvalue);
	    ++op;
	    break;
	case CC:	/* condition code */
	    sprintf(&args[strlen(args)]," cc_%s",op->value.svalue);
	    ++op;
	    break;
	case STALLOP:	/* bra stall op */
	    sprintf(&args[strlen(args)]," so_%s",op->value.svalue);
	    ++op;
	    break;
	case T:	/* implicit alw condition code */
	    strcat(args," cc_t");
	    break;
	case DSOFF4:	/* data space offset */
	case DSOFF8:
	case DSOFF16:
	    sprintf(&args[strlen(args)]," %s",op->value.svalue);
	    sprintf(&args[strlen(args)]," %ld",
		    op->value2.fvalue);
	    ++op;
	    break;
	case BILINEAR:	/* bilinear indexed indirect */
	    sprintf(&args[strlen(args)]," %s",op->value.svalue);
	    ++op;
	    break;
	case LABEL:	/* branch target */
	    SetFileInfo(c,i->file,i->line,0); /* in case of an error */
	    {	MScalar addr = GetNumericValue(c,&op->value,"Undefined label");
	    	CheckInstructionAddress(c,addr);
	    	sprintf(&args[strlen(args)]," $%08lx",addr);
		++op;
	    }
	    break;
	case RLABEL:	/* branch target (relative branches) */
	    SetFileInfo(c,i->file,i->line,0); /* in case of an error */
	    {	MScalar addr = GetNumericValue(c,&op->value,"Undefined label");
		MScalar offset = addr - NextInstructionAddress(i);
		int bits = *pat++;
		offset /= bytesPerInstruction; /* here because of a bug in Visual C++ */
		RangeCheckLabelReference(c,bits,offset);
		sprintf(&args[strlen(args)]," %ld",offset);
		++op;
	    }
	    break;
        case DUP:       /* duplicate last (register) operand */
	    sprintf(&args[strlen(args)]," %s",op[-1].value.svalue);
            break;
        case NO_STALLOP:/* implicit nop stallop */
	    sprintf(&args[strlen(args)]," so_next");
            break;
	}
    
    /* output the subinstruction */
    fprintf(fp,"'(#x%08lx \"%s%s\")",flags,opname,args);
}

/* AppendOperandType - append the operand type to an opcode */
void AppendOperandType(char *opname,int code,char **ppat)
{
    switch (code) {
        case SR_EQ:     /* scalar register matching another operand */
            ++(*ppat);
	case SR:	/* scalar register */
        case HR:        /* half vector register */
	    strcat(opname,"_sr");
	    break;
	case VR:	/* vector register */
	    strcat(opname,"_vr");
	    break;
	case PR:	/* pixel register */
	    strcat(opname,"_pr");
	    break;
	case RC0:	/* counter registers */
        case RC1:
	    strcat(opname,"_rc");
	    break;
	case RZ:	/* rz register */
	    strcat(opname,"_rz");
	    break;
	case PC:	/* pc register */
	    strcat(opname,"_pc");
	    break;
	case RUR:	/* register unit register */
	    strcat(opname,"_rur");
	    break;
	case RUV:	/* register ru or rv */
	    strcat(opname,"_ruv");
	    break;
	case IO:	/* i/o register */
	    strcat(opname,"_io");
	    break;
        case AC_SHIFT:  /* ac_shift register */
	    strcat(opname,"_ac_shift");
	    break;
        case SV_SHIFT:  /* sv_shift register */
	    strcat(opname,"_sv_shift");
	    break;
	case IMM:	/* immediate value */
	    strcat(opname,"_imm");
            if (**ppat != 5 && **ppat != 7)
                sprintf(&opname[strlen(opname)],"%d",**ppat);
	    ++(*ppat);
	    break;
	case UIMM:	/* unsigned immediate value */
	    strcat(opname,"_uimm");
            if (**ppat != 5 && **ppat != 7)
                sprintf(&opname[strlen(opname)],"%d",**ppat);
	    ++(*ppat);
	    break;
	case RUIMM:	/* register unit immediate value */
	    strcat(opname,"_ruimm");
            ++(*ppat);
	    break;
	case MULIMM:	/* multiplier immediate value */
	    strcat(opname,"_mulimm");
	    break;
	case SHIFT:	/* shift operator */
	    strcat(opname,"_shift");
	    break;
	case ROTATE:	/* rotate operator */
	    strcat(opname,"_rotate");
	    break;
	case CC:	/* condition code */
	    strcat(opname,"_cc");
	    break;
	case T:		/* implicit alw condition code */
	    strcat(opname,"_cc");
	    break;
	case DSOFF4:	/* data space offset */
	    strcat(opname,"_dsoff4");
	    break;
	case DSOFF8:
	    strcat(opname,"_dsoff8");
	    break;
	case DSOFF16:
	    strcat(opname,"_dsoff16");
	    break;
	case BILINEAR:	/* bilinear indexed indirect */
	    strcat(opname,"_bilinear");
	    break;
	case LABEL:	/* branch target */
	    strcat(opname,"_imm");
	    break;
	case RLABEL:	/* branch target (relative branches) */
	    strcat(opname,"_imm");
	    ++(*ppat);
	    break;
        case IND:
            strcat(opname,"_ind");
            break;
        case STALLOP:
        case NO_STALLOP:
            strcat(opname,"_stallop");
            break;
    }
}

/* RangeCheckLabelReference - range check a label reference */
static void RangeCheckLabelReference(ScanContext *c,int bits,MScalar offset)
{
    MScalar range2 = 1 << (bits - 1);
    MScalar lo = -range2;
    MScalar hi = range2 - 1;
    if (offset < lo || offset > hi)
        ParseError(c,"Branch offset out of range %ld (%ld..%ld)",offset,lo,hi);
}

/* NextInstructionAddress - compute the address of the next instruction */
static MScalar NextInstructionAddress(Instruction *i)
{
    MScalar addr = i->addr;
    for (;;) {
        addr += InstructionSize(i);
        if (i->flags & ifStop)
            break;
        i = i->next;
    }
    return addr;
}

/* InstructionSize - get the size of an instruction */
static int InstructionSize(Instruction *i)
{
    if (i->flags & if48Bits)
        return 6;
    else if (i->flags & if32Bits)
        return 4;
    return 2;
}

/* OutputString - output a quoted string */
static void OutputString(char *str,FILE *fp)
{
    int ch;
    putc('"',fp);
    while ((ch = *str++) != '\0') {
        if (ch == '\\')
	    putc(ch,fp);
	putc(ch,fp);
    }
    putc('"',fp);
}
